@@ -9,6 +9,7 @@
 enum gscon_fsm_event {
 	/* local SCCP stack tells us incoming conn from MSC */
 	GSCON_EV_A_CONN_IND,
+	GSCON_EV_A_INITIAL_USER_DATA,
 	/* RSL side requests CONNECT to MSC */
 	GSCON_EV_MO_COMPL_L3,
 	/* MSC confirms the SCCP connection */
@@ -65,6 +66,7 @@ void gscon_update_id(struct gsm_subscriber_connection *conn);
 void gscon_submit_rsl_dtap(struct gsm_subscriber_connection *conn,
 			   struct msgb *msg, int link_id, int allow_sacch);
 int gscon_sigtran_send(struct gsm_subscriber_connection *conn, struct msgb *msg);
+void gscon_bssmap_clear(struct gsm_subscriber_connection *conn, enum gsm0808_cause cause);
 
 struct osmo_mgcpc_ep *gscon_ensure_mgw_endpoint(struct gsm_subscriber_connection *conn,
 						uint16_t msc_assigned_cic, struct gsm_lchan *for_lchan);
