@@ -56,8 +56,7 @@
 #include <osmocom/abis/abis.h>
 #include <osmocom/gsm/protocol/gsm_12_21.h>
 #include <osmocom/bsc/bss.h>
-
-struct gsm_network *bsc_gsmnet;
+#include <osmocom/bsc/bts.h>
 
 static int net_listen_testnr;
 static int restart;
@@ -92,7 +91,7 @@ static int ipaccess_connect(struct e1inp_line *line, struct sockaddr_in *sa)
 
 	bfd->fd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
 	bfd->cb = ipaccess_fd_cb;
-	bfd->when = BSC_FD_READ | BSC_FD_WRITE;
+	bfd->when = OSMO_FD_READ | OSMO_FD_WRITE;
 	bfd->data = line;
 	bfd->priv_nr = E1INP_SIGN_OML;
 
@@ -143,21 +142,22 @@ static int ia_config_connect(struct gsm_bts *bts, struct sockaddr_in *sin)
 	}
 	line->ops = &ipaccess_e1inp_line_ops;
 
-	/* create E1 timeslots for signalling and TRAU frames */
-	e1inp_ts_config_sign(&line->ts[1-1], line);
-	e1inp_ts_config_sign(&line->ts[2-1], line);
+	sign_ts = e1inp_line_ipa_oml_ts(line);
+	rsl_ts = e1inp_line_ipa_rsl_ts(line, 0);
 
-	/* create signalling links for TS1 */
-	sign_ts = &line->ts[1-1];
-	rsl_ts = &line->ts[2-1];
+	/* create E1 timeslots for signalling and TRAU frames */
+	e1inp_ts_config_sign(sign_ts, line);
+	e1inp_ts_config_sign(rsl_ts, line);
+
+	/* create signalling links for TRX0 */
 	oml_link = e1inp_sign_link_create(sign_ts, E1INP_SIGN_OML,
-					  bts->c0, 0xff, 0);
+					  bts->c0, IPAC_PROTO_OML, 0);
 	rsl_link = e1inp_sign_link_create(rsl_ts, E1INP_SIGN_RSL,
-					  bts->c0, 0, 0);
+					  bts->c0, IPAC_PROTO_RSL, 0);
 
 	/* create back-links from bts/trx */
 	bts->oml_link = oml_link;
-	bts->c0->rsl_link = rsl_link;
+	bts->c0->rsl_link_primary = rsl_link;
 
 	/* default port at BTS for incoming connections is 3006 */
 	if (sin->sin_port == 0)
@@ -268,6 +268,8 @@ static int nwl_sig_cb(unsigned int subsys, unsigned int signal,
 	return 0;
 }
 
+static const struct value_string ipa_nvflag_strs[];
+
 static int print_attr_rep(struct msgb *mb)
 {
 	/* Parse using nanoBTS own formatting for Get Attribute Response */
@@ -281,22 +283,67 @@ static int print_attr_rep(struct msgb *mb)
 	char oml_ip[20] = {0};
 	uint16_t oml_port = 0;
 	char unit_id[40]  = {0};
+	unsigned int indent = 0;
 
 
-	abis_nm_tlv_parse(&tp, bts, foh->data, oh->length-sizeof(*foh));
+	if (abis_nm_tlv_parse(&tp, bts, foh->data, oh->length-sizeof(*foh)) < 0) {
+		LOGPFOH(DNM, LOGL_ERROR, foh, "%s(): tlv_parse failed\n", __func__);
+		return -EINVAL;
+	}
 
 	abis_nm_tlv_attr_primary_oml(&tp, &ia, &oml_port);
 	osmo_strlcpy(oml_ip, inet_ntoa(ia), sizeof(oml_ip));
 
 	abis_nm_tlv_attr_unit_id(&tp, unit_id,  sizeof(unit_id));
 
-	fprintf(stdout, "{ \"primary_oml_ip\": \"%s\", \"primary_oml_port\": %" PRIu16 ", \"unit_id\": \"%s\" }\n",
-		oml_ip, oml_port, unit_id);
+#define ENDL(last) \
+	fprintf(stdout, "%s\n", last ? "" : ",")
+#define print_offset(fmt, args...) \
+	fprintf(stdout, "%*s" fmt, indent * 4, "", ## args)
+#define print_field(field, fmt, args...) \
+	print_offset("\"%s\": \"" fmt "\"", field, ## args)
+
+	print_offset("{\n");
+	indent++;
+
+	print_field("primary_oml_ip", "%s", oml_ip); ENDL(false);
+	print_field("primary_oml_port", "%u", oml_port); ENDL(false);
+	print_field("unit_id", "%s", unit_id); ENDL(false);
+
+	uint16_t Fx = (TLVP_VAL(&tp, NM_ATT_IPACC_NV_FLAGS)[2] << 8)
+		    | (TLVP_VAL(&tp, NM_ATT_IPACC_NV_FLAGS)[0] << 0);
+	uint16_t Mx = (TLVP_VAL(&tp, NM_ATT_IPACC_NV_FLAGS)[3] << 8)
+		    | (TLVP_VAL(&tp, NM_ATT_IPACC_NV_FLAGS)[1] << 0);
+	const struct value_string *nvflag = ipa_nvflag_strs;
+
+	print_offset("\"nv_flags\": {\n");
+	indent++;
+
+	while (nvflag->value && nvflag->str) {
+		const char *val = (Fx & nvflag->value) ? "yes" : "no";
+		if (~Mx & nvflag->value)
+			val = "unknown";
+		print_field(nvflag->str, "%s", val);
+
+		nvflag++;
+
+		if (nvflag->value && nvflag->str)
+			ENDL(false); /* more fields to print */
+		else
+			ENDL(true); /* this was the last field */
+	}
+
+	indent--;
+	print_offset("}\n");
+
+	indent--;
+	print_offset("}\n");
+
 	return 0;
 }
 
 static int nm_state_event(int evt, uint8_t obj_class, void *obj,
-			  struct gsm_nm_state *old_state, struct gsm_nm_state *new_state,
+			  const struct gsm_nm_state *old_state, const struct gsm_nm_state *new_state,
 			  struct abis_om_obj_inst *obj_inst);
 
 static int nm_sig_cb(unsigned int subsys, unsigned int signal,
@@ -323,11 +370,10 @@ static int nm_sig_cb(unsigned int subsys, unsigned int signal,
 			printf("The BTS has nacked the restart. Exiting.\n");
 		exit(0);
 		break;
-	case S_NM_STATECHG_OPER:
-	case S_NM_STATECHG_ADM:
+	case S_NM_STATECHG:
 		nsd = signal_data;
-		nm_state_event(signal, nsd->obj_class, nsd->obj, nsd->old_state,
-				nsd->new_state, nsd->obj_inst);
+		nm_state_event(signal, nsd->obj_class, nsd->obj, &nsd->old_state,
+				&nsd->new_state, nsd->obj_inst);
 		break;
 	case S_NM_GET_ATTR_REP:
 		fprintf(stderr, "Received SIGNAL S_NM_GET_ATTR_REP\n");
@@ -501,7 +547,7 @@ static const struct value_string ipa_nvflag_strs[] = {
 	{ 0x0002, "static-gw" },
 	{ 0x0004, "no-dhcp-vsi" },
 	{ 0x0008, "dhcp-enabled" },
-	{ 0x0040, "led-disabled" },
+	{ 0x0040, "led-enabled" },
 	{ 0x0100, "secondary-oml-enabled" },
 	{ 0x0200, "diag-enabled" },
 	{ 0x0400, "cli-enabled" },
@@ -516,8 +562,10 @@ static int ipa_nvflag_set(uint16_t *flags, uint16_t *mask, const char *name, int
 {
 	int rc;
 	rc = get_string_value(ipa_nvflag_strs, name);
-	if (rc < 0)
+	if (rc < 0) {
+		fprintf(stderr, "Unknown attribute '%s'\n", name);
 		return rc;
+	}
 
 	*mask |= rc;
 	if (en)
@@ -541,6 +589,7 @@ static void bootstrap_om(struct gsm_bts_trx *trx)
 	if (get_attr) {
 		msgb_put_u8(nmsg_get, NM_ATT_IPACC_PRIM_OML_CFG);
 		msgb_put_u8(nmsg_get, NM_ATT_IPACC_UNIT_ID);
+		msgb_put_u8(nmsg_get, NM_ATT_IPACC_NV_FLAGS);
 	}
 	if (unit_id) {
 		len = strlen(unit_id);
@@ -631,7 +680,7 @@ out_err:
 }
 
 static int nm_state_event(int evt, uint8_t obj_class, void *obj,
-			  struct gsm_nm_state *old_state, struct gsm_nm_state *new_state,
+			  const struct gsm_nm_state *old_state, const struct gsm_nm_state *new_state,
 			  struct abis_om_obj_inst *obj_inst)
 {
 	if (obj_class == NM_OC_BASEB_TRANSC) {
@@ -640,9 +689,9 @@ static int nm_state_event(int evt, uint8_t obj_class, void *obj,
 			bootstrap_om(trx);
 			found_trx = 1;
 		}
-	} else if (evt == S_NM_STATECHG_OPER &&
+	} else if (evt == S_NM_STATECHG &&
 	    obj_class == NM_OC_RADIO_CARRIER &&
-	    new_state->availability == 3) {
+	    new_state->availability == NM_AVSTATE_OFF_LINE) {
 		struct gsm_bts_trx *trx = obj;
 
 		if (net_listen_testnr)
@@ -925,7 +974,7 @@ static const struct log_info_cat log_categories[] = {
 		.name = "DNM",
 		.description = "A-bis Network Management / O&M (NM/OML)",
 		.color = "\033[1;36m",
-		.loglevel = LOGL_DEBUG,
+		.loglevel = LOGL_NOTICE,
 		.enabled = 1,
 	},
 };
@@ -1120,7 +1169,7 @@ int main(int argc, char **argv)
 	}
 
 	bts->oml_link->ts->sign.delay = 10;
-	bts->c0->rsl_link->ts->sign.delay = 10;
+	bts->c0->rsl_link_primary->ts->sign.delay = 10;
 	while (1) {
 		rc = osmo_select_main(0);
 		if (rc < 0)
@@ -1129,21 +1178,3 @@ int main(int argc, char **argv)
 
 	exit(0);
 }
-
-/* Stub */
-int osmo_bsc_sigtran_send(struct gsm_subscriber_connection *conn, struct msgb *msg)
-{
-	return 0;
-}
-
-/* Stub */
-int osmo_bsc_sigtran_open_conn(struct gsm_subscriber_connection *conn, struct msgb *msg)
-{
-	return 0;
-}
-
-/* Stub */
-int rsl_chan_ms_power_ctrl(struct gsm_lchan *lchan)
-{
-	return 0;
-}
