@@ -88,11 +88,19 @@ static const struct osmo_tdef_state_timeout assignment_fsm_timeouts[32] = {
 			switch (gsm48_chan_mode_to_non_vamos(conn->assignment.req.ch_mode_rate_list[0].chan_mode)) { \
 			case GSM48_CMODE_SIGN: \
 				rate_ctr_inc(rate_ctr_group_get_ctr(bts->bts_ctrs, BTS_##counter##_SIGN)); \
+				LOG_ASSIGNMENT(conn, LOGL_DEBUG, "incrementing rate counter: bts%u %s %s\n", \
+					       bts->nr, \
+					       bts_ctr_description[BTS_##counter##_SIGN].name, \
+					       bts_ctr_description[BTS_##counter##_SIGN].description); \
 				break; \
 			case GSM48_CMODE_SPEECH_V1: \
 			case GSM48_CMODE_SPEECH_EFR: \
 			case GSM48_CMODE_SPEECH_AMR: \
 				rate_ctr_inc(rate_ctr_group_get_ctr(bts->bts_ctrs, BTS_##counter##_SPEECH)); \
+				LOG_ASSIGNMENT(conn, LOGL_DEBUG, "incrementing rate counter: bts%u %s %s\n", \
+					       bts->nr, \
+					       bts_ctr_description[BTS_##counter##_SPEECH].name, \
+					       bts_ctr_description[BTS_##counter##_SPEECH].description); \
 				break; \
 			default: \
 				break; \
@@ -120,10 +128,13 @@ void assignment_reset(struct gsm_subscriber_connection *conn)
 	}
 
 	if (conn->assignment.created_ci_for_msc) {
-		gscon_forget_mgw_endpoint_ci(conn, conn->assignment.created_ci_for_msc);
+		/* Store ci pointer locally, because gscon_forget_mgw_endpoint_ci() NULLs
+		 * conn->assignment.created_ci_for_msc. */
+		struct osmo_mgcpc_ep_ci *ci = conn->assignment.created_ci_for_msc;
+		gscon_forget_mgw_endpoint_ci(conn, ci);
 		/* If this is the last endpoint released, the mgw_endpoint_fsm will terminate and tell
 		 * the gscon about it. */
-		osmo_mgcpc_ep_ci_dlcx(conn->assignment.created_ci_for_msc);
+		osmo_mgcpc_ep_ci_dlcx(ci);
 	}
 
 	conn->assignment = (struct assignment_fsm_data){
@@ -218,7 +229,7 @@ static void send_assignment_complete(struct gsm_subscriber_connection *conn)
 
 	resp = gsm0808_create_ass_compl2(lchan->abis_ip.ass_compl.rr_cause,
 					 chosen_channel,
-					 lchan->encr.alg_id, perm_spch,
+					 ALG_A5_NR_TO_BSSAP(lchan->encr.alg_a5_n), perm_spch,
 					 addr_local_p, sc_ptr, NULL, lcls_get_status(conn));
 
 	if (!resp) {
@@ -275,17 +286,16 @@ static void assignment_success(struct gsm_subscriber_connection *conn)
 		}
 	}
 
-	if (lchan_changed) {
-		/* Rembered this only for error handling: should assignment fail, assignment_reset() will release
-		 * the MGW endpoint right away. If successful, the conn continues to use the endpoint. */
-		conn->assignment.created_ci_for_msc = NULL;
+	/* Rembered this only for error handling: should assignment fail, assignment_reset() will release
+	 * the MGW endpoint right away. If successful, the conn continues to use the endpoint. */
+	conn->assignment.created_ci_for_msc = NULL;
 
-		/* New RTP information is now accepted */
-		conn->user_plane.msc_assigned_cic = conn->assignment.req.msc_assigned_cic;
-		osmo_strlcpy(conn->user_plane.msc_assigned_rtp_addr, conn->assignment.req.msc_rtp_addr,
-			     sizeof(conn->user_plane.msc_assigned_rtp_addr));
-		conn->user_plane.msc_assigned_rtp_port = conn->assignment.req.msc_rtp_port;
-	}
+	/* New RTP information is now accepted. If there is no RTP stream, this information is zero / empty. Either way
+	 * store the result of this assignment. */
+	conn->user_plane.msc_assigned_cic = conn->assignment.req.msc_assigned_cic;
+	osmo_strlcpy(conn->user_plane.msc_assigned_rtp_addr, conn->assignment.req.msc_rtp_addr,
+		     sizeof(conn->user_plane.msc_assigned_rtp_addr));
+	conn->user_plane.msc_assigned_rtp_port = conn->assignment.req.msc_rtp_port;
 
 	assignment_count_result(CTR_ASSIGNMENT_COMPLETED);
 
