@@ -60,6 +60,8 @@
 
 enum gscon_fsm_states {
 	ST_INIT,
+	/* wait for initial BSSMAP after the MSC opened a new SCCP connection */
+	ST_WAIT_INITIAL_USER_DATA,
 	/* waiting for CC from MSC */
 	ST_WAIT_CC,
 	/* active connection */
@@ -67,11 +69,13 @@ enum gscon_fsm_states {
 	ST_ASSIGNMENT,
 	ST_HANDOVER,
 	/* BSSMAP CLEAR has been received */
-	ST_CLEARING,
+	ST_WAIT_CLEAR_CMD,
+	ST_WAIT_SCCP_RLSD,
 };
 
 static const struct value_string gscon_fsm_event_names[] = {
 	{GSCON_EV_A_CONN_IND, "MT-CONNECT.ind"},
+	{GSCON_EV_A_INITIAL_USER_DATA, "A_INITIAL_USER_DATA"},
 	{GSCON_EV_MO_COMPL_L3, "MO_COMPL_L3"},
 	{GSCON_EV_A_CONN_CFM, "MO-CONNECT.cfm"},
 	{GSCON_EV_A_CLEAR_CMD, "CLEAR_CMD"},
@@ -94,8 +98,10 @@ static const struct value_string gscon_fsm_event_names[] = {
 };
 
 struct osmo_tdef_state_timeout conn_fsm_timeouts[32] = {
+	[ST_WAIT_INITIAL_USER_DATA] = { .T = -25 },
 	[ST_WAIT_CC] = { .T = -3210 },
-	[ST_CLEARING] = { .T = -4 },
+	[ST_WAIT_CLEAR_CMD] = { .T = -4 },
+	[ST_WAIT_SCCP_RLSD] = { .T = -4 },
 };
 
 /* Transition to a state, using the T timer defined in conn_fsm_timeouts.
@@ -138,34 +144,79 @@ int gscon_sigtran_send(struct gsm_subscriber_connection *conn, struct msgb *msg)
 	return rc;
 }
 
-static void gscon_bssmap_clear(struct gsm_subscriber_connection *conn,
-			       enum gsm0808_cause cause)
+void gscon_bssmap_clear(struct gsm_subscriber_connection *conn, enum gsm0808_cause cause)
 {
+	/* already clearing? */
+	switch (conn->fi->state) {
+	case ST_WAIT_CLEAR_CMD:
+	case ST_WAIT_SCCP_RLSD:
+		return;
+	default:
+		break;
+	}
 
+	conn->clear_cause = cause;
+	conn_fsm_state_chg(ST_WAIT_CLEAR_CMD);
+}
+
+static void gscon_fsm_wait_clear_cmd_onenter(struct osmo_fsm_inst *fi, uint32_t prev_state)
+{
 	struct msgb *resp;
 	int rc;
-
-	if (conn->rx_clear_command) {
-		LOGPFSML(conn->fi, LOGL_DEBUG, "Not sending BSSMAP CLEAR REQUEST, already got CLEAR COMMAND from MSC\n");
-		return;
-	}
+	struct gsm_subscriber_connection *conn = fi->priv;
+	enum gsm0808_cause cause = conn->clear_cause;
 
 	if (!conn->sccp.msc) {
-		LOGPFSML(conn->fi, LOGL_ERROR, "Unable to deliver BSSMAP Clear Request message, no MSC for this conn\n");
-		return;
+		LOGPFSML(fi, LOGL_ERROR, "Unable to deliver BSSMAP Clear Request message, no MSC for this conn\n");
+		goto nothing_sent;
 	}
 
-	LOGPFSML(conn->fi, LOGL_DEBUG, "Tx BSSMAP CLEAR REQUEST(%s) to MSC\n", gsm0808_cause_name(cause));
+	LOGPFSML(fi, LOGL_DEBUG, "Tx BSSMAP CLEAR REQUEST(%s) to MSC\n", gsm0808_cause_name(cause));
 	resp = gsm0808_create_clear_rqst(cause);
 	if (!resp) {
-		LOGPFSML(conn->fi, LOGL_ERROR, "Unable to compose BSSMAP Clear Request message\n");
-		return;
+		LOGPFSML(fi, LOGL_ERROR, "Unable to compose BSSMAP Clear Request message\n");
+		goto nothing_sent;
 	}
 
 	rate_ctr_inc(rate_ctr_group_get_ctr(conn->sccp.msc->msc_ctrs, MSC_CTR_BSSMAP_TX_DT1_CLEAR_RQST));
 	rc = osmo_bsc_sigtran_send(conn, resp);
-	if (rc < 0)
+	if (rc < 0) {
 		LOGPFSML(conn->fi, LOGL_ERROR, "Unable to deliver BSSMAP Clear Request message\n");
+		goto nothing_sent;
+	}
+	return;
+
+nothing_sent:
+	/* Normally, we request a CLEAR from the MSC and terminate as soon as the CLEAR COMMAND has been issued by the
+	 * MSC. But if we are trying to clear without being able to send anything to the MSC, we might as well shut down
+	 * the conn right away now. */
+	conn_fsm_state_chg(ST_WAIT_SCCP_RLSD);
+}
+
+void gscon_fsm_wait_sccp_rlsd_onenter(struct osmo_fsm_inst *fi, uint32_t prev_state)
+{
+	struct gsm_subscriber_connection *conn = fi->priv;
+
+	/* According to 3GPP 48.008 3.1.9.1. "The BSS need not wait for the radio channel
+	 * release to be completed or for the guard timer to expire before returning the
+	 * CLEAR COMPLETE message" */
+	if (!gscon_sigtran_send(conn, gsm0808_create_clear_complete()))
+		rate_ctr_inc(rate_ctr_group_get_ctr(conn->sccp.msc->msc_ctrs, MSC_CTR_BSSMAP_TX_DT1_CLEAR_COMPLETE));
+
+	/* Give the handover_fsm a chance to book this as handover success before tearing down everything,
+	 * making it look like a sudden death failure. */
+	if (conn->ho.fi)
+		osmo_fsm_inst_dispatch(conn->ho.fi, HO_EV_CONN_RELEASING, NULL);
+
+	if (conn->lcs.loc_req)
+		osmo_fsm_inst_dispatch(conn->lcs.loc_req->fi, LCS_LOC_REQ_EV_CONN_CLEAR, NULL);
+
+	gscon_release_lchans(conn, true, bsc_gsm48_rr_cause_from_gsm0808_cause(conn->clear_cause));
+	osmo_mgcpc_ep_clear(conn->user_plane.mgw_endpoint);
+
+	/* If there is no SCCP connection at all, then no need to wait for an SCCP RLSD. */
+	if (!conn->sccp.msc || conn->sccp.state != SUBSCR_SCCP_ST_CONNECTED)
+		osmo_fsm_inst_term(fi, OSMO_FSM_TERM_REGULAR, NULL);
 }
 
 /* forward MO DTAP from RSL side to BSSAP side */
@@ -210,89 +261,133 @@ void gscon_release_lchans(struct gsm_subscriber_connection *conn, bool do_rr_rel
 	gscon_release_lchan(conn, conn->lchan, do_rr_release, false, cause_rr);
 }
 
-static void handle_bssap_n_connect(struct osmo_fsm_inst *fi, struct osmo_scu_prim *scu_prim)
+static int validate_initial_user_data(struct osmo_fsm_inst *fi, struct msgb *msg)
 {
-	struct gsm_subscriber_connection *conn = fi->priv;
-	struct msgb *msg = scu_prim->oph.msg;
 	struct bssmap_header *bs;
-	uint8_t bssmap_type;
+	enum BSS_MAP_MSG_TYPE bssmap_type;
 
 	msg->l3h = msgb_l2(msg);
 	if (!msgb_l3(msg)) {
 		LOGPFSML(fi, LOGL_ERROR, "internal error: no l3 in msg\n");
-		goto refuse;
+		return -EINVAL;
 	}
 
 	if (msgb_l3len(msg) < sizeof(*bs)) {
-		LOGPFSML(fi, LOGL_NOTICE, "message too short for BSSMAP header (%u < %zu)\n",
+		LOGPFSML(fi, LOGL_ERROR, "message too short for BSSMAP header (%u < %zu)\n",
 			 msgb_l3len(msg), sizeof(*bs));
-		goto refuse;
+		return -EINVAL;
 	}
 
 	bs = (struct bssmap_header*)msgb_l3(msg);
 	if (msgb_l3len(msg) < (bs->length + sizeof(*bs))) {
-		LOGPFSML(fi, LOGL_NOTICE,
+		LOGPFSML(fi, LOGL_ERROR,
 			 "message too short for length indicated in BSSMAP header (%u < %u)\n",
 			 msgb_l3len(msg), bs->length);
-		goto refuse;
+		return -EINVAL;
 	}
 
 	switch (bs->type) {
 	case BSSAP_MSG_BSS_MANAGEMENT:
 		break;
 	default:
-		LOGPFSML(fi, LOGL_NOTICE,
-			 "message type not allowed for N-CONNECT: %s\n", gsm0808_bssap_name(bs->type));
-		goto refuse;
+		LOGPFSML(fi, LOGL_ERROR,
+			 "message type not allowed for initial BSSMAP: %s\n", gsm0808_bssap_name(bs->type));
+		return -EINVAL;
 	}
 
 	msg->l4h = &msg->l3h[sizeof(*bs)];
+
+	/* Validate initial message type. See also BSC_Tests.TC_outbound_connect. */
+	bssmap_type = msg->l4h[0];
+	switch (bssmap_type) {
+	case BSS_MAP_MSG_HANDOVER_RQST:
+	case BSS_MAP_MSG_PERFORM_LOCATION_RQST:
+		return 0;
+
+	default:
+		LOGPFSML(fi, LOGL_ERROR, "No support for initial BSSMAP: %s: %s\n",
+			 gsm0808_bssap_name(bs->type), gsm0808_bssmap_name(bssmap_type));
+		return -EINVAL;
+	}
+}
+
+static void handle_initial_user_data(struct osmo_fsm_inst *fi, struct msgb *msg)
+{
+	struct gsm_subscriber_connection *conn = fi->priv;
+	struct bssmap_header *bs;
+	enum BSS_MAP_MSG_TYPE bssmap_type;
+
+	/* validate_initial_user_data() must be called before this */
+	OSMO_ASSERT(msgb_l4(msg));
+
+	bs = msgb_l3(msg);
 	bssmap_type = msg->l4h[0];
 
-	LOGPFSML(fi, LOGL_DEBUG, "Rx N-CONNECT: %s: %s\n", gsm0808_bssap_name(bs->type),
+	/* FIXME: Extract optional IMSI and update FSM using osmo_fsm_inst_set_id() (OS#2969) */
+
+	LOGPFSML(fi, LOGL_DEBUG, "Rx initial BSSMAP: %s: %s\n", gsm0808_bssap_name(bs->type),
 		 gsm0808_bssmap_name(bssmap_type));
 
 	switch (bssmap_type) {
 	case BSS_MAP_MSG_HANDOVER_RQST:
-	case BSS_MAP_MSG_PERFORM_LOCATION_RQST:
-		break;
-
-	default:
-		LOGPFSML(fi, LOGL_NOTICE, "No support for N-CONNECT: %s: %s\n",
-			 gsm0808_bssap_name(bs->type), gsm0808_bssmap_name(bssmap_type));
-		goto refuse;
-	}
-
-	/* First off, accept the new conn. */
-	if (osmo_sccp_tx_conn_resp(conn->sccp.msc->a.sccp_user, scu_prim->u.connect.conn_id,
-				   &scu_prim->u.connect.called_addr, NULL, 0)) {
-		LOGPFSML(fi, LOGL_ERROR, "Cannot send SCCP CONN RESP\n");
-		goto refuse;
-	}
-
-	/* Make sure the conn FSM will osmo_sccp_tx_disconn() on term */
-	conn->sccp.state = SUBSCR_SCCP_ST_CONNECTED;
-
-	switch (bssmap_type) {
-	case BSS_MAP_MSG_HANDOVER_RQST:
-		/* Inter-BSC MT Handover Request, another BSS is handovering to us. */
+		rate_ctr_inc(&conn->sccp.msc->msc_ctrs->ctr[MSC_CTR_BSSMAP_RX_DT1_HANDOVER_RQST]);
+		/* Inter-BSC incoming Handover Request, another BSS is handovering to us. */
 		handover_start_inter_bsc_in(conn, msg);
 		return;
 
 	case BSS_MAP_MSG_PERFORM_LOCATION_RQST:
+		rate_ctr_inc(&conn->sccp.msc->msc_ctrs->ctr[MSC_CTR_BSSMAP_RX_DT1_PERFORM_LOCATION_REQUEST]);
 		/* Location Services: MSC asks for location of an IDLE subscriber */
 		conn_fsm_state_chg(ST_ACTIVE);
 		lcs_loc_req_start(conn, msg);
 		return;
 
 	default:
-		OSMO_ASSERT(false);
+		LOGPFSML(fi, LOGL_ERROR, "No support for initial BSSMAP: %s: %s\n",
+			 gsm0808_bssap_name(bs->type), gsm0808_bssmap_name(bssmap_type));
+		osmo_fsm_inst_term(fi, OSMO_FSM_TERM_REGULAR, NULL);
+		return;
+	}
+}
+
+static void handle_sccp_n_connect(struct osmo_fsm_inst *fi, struct osmo_scu_prim *scu_prim)
+{
+	struct gsm_subscriber_connection *conn = fi->priv;
+	struct msgb *msg = scu_prim->oph.msg;
+
+	/* Make sure the conn FSM will osmo_sccp_tx_disconn() on term */
+	conn->sccp.state = SUBSCR_SCCP_ST_CONNECTED;
+
+	msg->l3h = msgb_l2(msg);
+
+	/* If (BSSMAP) user data is included, validate it before accepting the connection */
+	if (msgb_l3(msg) && msgb_l3len(msg)) {
+		if (validate_initial_user_data(fi, msg)) {
+			osmo_fsm_inst_term(fi, OSMO_FSM_TERM_REGULAR, NULL);
+			return;
+		}
 	}
 
-refuse:
-	osmo_sccp_tx_disconn(conn->sccp.msc->a.sccp_user, scu_prim->u.connect.conn_id,
-			     &scu_prim->u.connect.called_addr, 0);
-	osmo_fsm_inst_term(fi, OSMO_FSM_TERM_REGULAR, NULL);
+	/* accept the new conn. */
+	if (osmo_sccp_tx_conn_resp(conn->sccp.msc->a.sccp_user, scu_prim->u.connect.conn_id,
+				   &scu_prim->u.connect.called_addr, NULL, 0)) {
+		LOGPFSML(fi, LOGL_ERROR, "Cannot send SCCP CONN RESP\n");
+		osmo_fsm_inst_term(fi, OSMO_FSM_TERM_REGULAR, NULL);
+		return;
+	}
+
+	/* The initial user data may already be included in this N-Connect, or it may come later in a separate message.
+	 * If it is already included, also go to ST_WAIT_INITIAL_USER_DATA now, so that we don't have to tend to two
+	 * separate code paths doing the same thing (handling of HANDOVER_END). */
+	OSMO_ASSERT(conn_fsm_state_chg(ST_WAIT_INITIAL_USER_DATA) == 0);
+
+	/* It is usually a bad idea to continue using a fi after a state change, because the fi might terminate during
+	 * the state change. In this case it is certain that the fi stays around for the initial user data. */
+	if (msgb_l3(msg) && msgb_l3len(msg)) {
+		handle_initial_user_data(fi, msg);
+	} else {
+		LOGPFSML(fi, LOGL_DEBUG, "N-Connect does not contain user data (no BSSMAP message included)\n");
+	}
 }
 
 static void gscon_fsm_init(struct osmo_fsm_inst *fi, uint32_t event, void *data)
@@ -301,7 +396,6 @@ static void gscon_fsm_init(struct osmo_fsm_inst *fi, uint32_t event, void *data)
 	struct osmo_scu_prim *scu_prim = NULL;
 	struct msgb *msg = NULL;
 	int rc;
-	enum handover_result ho_result;
 
 	switch (event) {
 	case GSCON_EV_MO_COMPL_L3:
@@ -329,11 +423,28 @@ static void gscon_fsm_init(struct osmo_fsm_inst *fi, uint32_t event, void *data)
 			osmo_fsm_inst_term(fi, OSMO_FSM_TERM_REGULAR, NULL);
 			return;
 		}
-		/* FIXME: Extract optional IMSI and update FSM using osmo_fsm_inst_set_id()
-		 * related: OS2969 (same as above) */
-
-		handle_bssap_n_connect(fi, scu_prim);
+		handle_sccp_n_connect(fi, scu_prim);
 		break;
+	default:
+		OSMO_ASSERT(false);
+	}
+}
+
+static void gscon_fsm_wait_initial_user_data(struct osmo_fsm_inst *fi, uint32_t event, void *data)
+{
+	struct gsm_subscriber_connection *conn = fi->priv;
+	struct msgb *msg = data;
+	enum handover_result ho_result;
+
+	switch (event) {
+	case GSCON_EV_A_INITIAL_USER_DATA:
+		if (validate_initial_user_data(fi, msg)) {
+			osmo_fsm_inst_term(fi, OSMO_FSM_TERM_REGULAR, NULL);
+			return;
+		}
+		handle_initial_user_data(fi, msg);
+		break;
+
 	case GSCON_EV_HANDOVER_END:
 		ho_result = HO_RESULT_ERROR;
 		if (data)
@@ -350,12 +461,10 @@ static void gscon_fsm_init(struct osmo_fsm_inst *fi, uint32_t event, void *data)
 				return;
 			}
 			LOG_HO(conn, LOGL_ERROR,
-			       "Conn is in state %s, the only accepted handover kind is inter-BSC MT\n",
+			       "Conn is in state %s, the only accepted handover kind is inter-BSC incoming handover\n",
 			       osmo_fsm_inst_state_name(conn->fi));
 		}
 		gscon_bssmap_clear(conn, GSM0808_CAUSE_EQUIPMENT_FAILURE);
-		if (conn->fi->state != ST_CLEARING)
-			osmo_fsm_inst_state_chg(fi, ST_CLEARING, 60, -4);
 		return;
 	default:
 		OSMO_ASSERT(false);
@@ -375,7 +484,6 @@ static void gscon_fsm_wait_cc(struct osmo_fsm_inst *fi, uint32_t event, void *da
 			   confirmed connection, then instead simply drop the connection */
 			LOGPFSML(fi, LOGL_INFO,
 				 "Connection confirmed but lchan was dropped previously, clearing conn\n");
-			osmo_fsm_inst_state_chg(conn->fi, ST_CLEARING, 60, -4);
 			gscon_bssmap_clear(conn, GSM0808_CAUSE_EQUIPMENT_FAILURE);
 			break;
 		}
@@ -668,15 +776,25 @@ bool gscon_connect_mgw_to_msc(struct gsm_subscriber_connection *conn,
 static const struct osmo_fsm_state gscon_fsm_states[] = {
 	[ST_INIT] = {
 		.name = "INIT",
-		.in_event_mask = S(GSCON_EV_MO_COMPL_L3) | S(GSCON_EV_A_CONN_IND)
-			| S(GSCON_EV_HANDOVER_END),
-		.out_state_mask = S(ST_WAIT_CC) | S(ST_ACTIVE) | S(ST_CLEARING),
+		.in_event_mask = S(GSCON_EV_MO_COMPL_L3) | S(GSCON_EV_A_CONN_IND),
+		.out_state_mask = 0
+			| S(ST_WAIT_INITIAL_USER_DATA)
+			| S(ST_WAIT_CC) | S(ST_ACTIVE) | S(ST_WAIT_CLEAR_CMD) | S(ST_WAIT_SCCP_RLSD),
 		.action = gscon_fsm_init,
+	},
+	[ST_WAIT_INITIAL_USER_DATA] = {
+		.name = "WAIT_INITIAL_USER_DATA",
+		.in_event_mask = 0
+			| S(GSCON_EV_A_INITIAL_USER_DATA)
+			| S(GSCON_EV_HANDOVER_END)
+			,
+		.out_state_mask = S(ST_WAIT_CC) | S(ST_ACTIVE) | S(ST_WAIT_CLEAR_CMD) | S(ST_WAIT_SCCP_RLSD),
+		.action = gscon_fsm_wait_initial_user_data,
 	 },
 	[ST_WAIT_CC] = {
 		.name = "WAIT_CC",
 		.in_event_mask = S(GSCON_EV_A_CONN_CFM),
-		.out_state_mask = S(ST_ACTIVE) | S(ST_CLEARING),
+		.out_state_mask = S(ST_ACTIVE) | S(ST_WAIT_CLEAR_CMD) | S(ST_WAIT_SCCP_RLSD),
 		.action = gscon_fsm_wait_cc,
 	},
 	[ST_ACTIVE] = {
@@ -686,26 +804,32 @@ static const struct osmo_fsm_state gscon_fsm_states[] = {
 				 | S(GSCON_EV_LCS_LOC_REQ_END)
 				 | S(GSCON_EV_MO_COMPL_L3)
 				 ,
-		.out_state_mask = S(ST_CLEARING) | S(ST_ASSIGNMENT) |
+		.out_state_mask = S(ST_WAIT_CLEAR_CMD) | S(ST_WAIT_SCCP_RLSD) | S(ST_ASSIGNMENT) |
 				  S(ST_HANDOVER),
 		.action = gscon_fsm_active,
 	},
 	[ST_ASSIGNMENT] = {
 		.name = "ASSIGNMENT",
 		.in_event_mask = EV_TRANSPARENT_SCCP | S(GSCON_EV_ASSIGNMENT_END),
-		.out_state_mask = S(ST_ACTIVE) | S(ST_CLEARING),
+		.out_state_mask = S(ST_ACTIVE) | S(ST_WAIT_CLEAR_CMD) | S(ST_WAIT_SCCP_RLSD),
 		.action = gscon_fsm_assignment,
 	},
 	[ST_HANDOVER] = {
 		.name = "HANDOVER",
 		.in_event_mask = EV_TRANSPARENT_SCCP | S(GSCON_EV_HANDOVER_END),
-		.out_state_mask = S(ST_ACTIVE) | S(ST_CLEARING),
+		.out_state_mask = S(ST_ACTIVE) | S(ST_WAIT_CLEAR_CMD) | S(ST_WAIT_SCCP_RLSD),
 		.action = gscon_fsm_handover,
 	},
-	[ST_CLEARING] = {
-		.name = "CLEARING",
-		/* dead end state */
-	 },
+	[ST_WAIT_CLEAR_CMD] = {
+		.name = "WAIT_CLEAR_CMD",
+		.onenter = gscon_fsm_wait_clear_cmd_onenter,
+		.out_state_mask = S(ST_WAIT_SCCP_RLSD),
+	},
+	[ST_WAIT_SCCP_RLSD] = {
+		.name = "WAIT_SCCP_RLSD",
+		.onenter = gscon_fsm_wait_sccp_rlsd_onenter,
+		.in_event_mask = S(GSCON_EV_HANDOVER_END),
+	},
 };
 
 void gscon_change_primary_lchan(struct gsm_subscriber_connection *conn, struct gsm_lchan *new_lchan)
@@ -754,19 +878,15 @@ void gscon_lchan_releasing(struct gsm_subscriber_connection *conn, struct gsm_lc
 		lchan_forget_conn(conn->lchan);
 		conn->lchan = NULL;
 	}
-	/* If the conn has no lchan anymore, it was released by the BTS and needs to Clear towards MSC. */
-	if (!conn->lchan) {
+	/* If the conn has no lchan anymore, it was released by the BTS and needs to Clear towards MSC.
+	 * However, if a Location Request is still busy, do not send Clear Request. */
+	if (!conn->lchan && !conn->lcs.loc_req) {
 		switch (conn->fi->state) {
 		case ST_WAIT_CC:
 			/* The SCCP connection was not yet confirmed by a CC, the BSSAP is not fully established
 			   yet so we cannot release it. First wait for the CC, and release in gscon_fsm_wait_cc(). */
 			break;
 		default:
-			/* Ensure that the FSM is in ST_CLEARING. */
-			osmo_fsm_inst_state_chg(conn->fi, ST_CLEARING, 60, -4);
-			/* fall thru, omit an error log if already in ST_CLEARING */
-		case ST_CLEARING:
-			/* Request a Clear Command from the MSC. */
 			gscon_bssmap_clear(conn, GSM0808_CAUSE_EQUIPMENT_FAILURE);
 			break;
 		}
@@ -805,8 +925,7 @@ void gscon_forget_lchan(struct gsm_subscriber_connection *conn, struct gsm_lchan
 				 osmo_fsm_inst_name(conn->fi), detach_label);
 	}
 
-	if ((conn->fi && conn->fi->state != ST_CLEARING)
-	    && !conn->lchan
+	if (!conn->lchan
 	    && !conn->ho.new_lchan
 	    && !conn->assignment.new_lchan
 	    && !conn->lcs.loc_req)
@@ -836,49 +955,27 @@ void gscon_forget_mgw_endpoint_ci(struct gsm_subscriber_connection *conn, struct
 
 	if (conn->user_plane.mgw_endpoint_ci_msc == ci)
 		conn->user_plane.mgw_endpoint_ci_msc = NULL;
+
+	if (conn->assignment.created_ci_for_msc == ci)
+		conn->assignment.created_ci_for_msc = NULL;
 }
 
 static void gscon_fsm_allstate(struct osmo_fsm_inst *fi, uint32_t event, void *data)
 {
 	struct gsm_subscriber_connection *conn = fi->priv;
-	const enum gsm0808_cause *cause_0808;
 	const struct tlv_parsed *tp;
 	struct osmo_mobile_identity mi_imsi;
 
 	/* Regular allstate event processing */
 	switch (event) {
 	case GSCON_EV_A_CLEAR_CMD:
-		conn->rx_clear_command = true;
-
-		/* Give the handover_fsm a chance to book this as handover success before tearing down everything,
-		 * making it look like a sudden death failure. */
-		if (conn->ho.fi)
-			osmo_fsm_inst_dispatch(conn->ho.fi, HO_EV_CONN_RELEASING, NULL);
-
-		if (conn->lcs.loc_req)
-			osmo_fsm_inst_dispatch(conn->lcs.loc_req->fi, LCS_LOC_REQ_EV_CONN_CLEAR, NULL);
-
 		OSMO_ASSERT(data);
-		cause_0808 = data;
-		/* MSC tells us to cleanly shut down */
-		if (conn->fi->state != ST_CLEARING)
-			osmo_fsm_inst_state_chg(fi, ST_CLEARING, 60, -4);
-		LOGPFSML(fi, LOGL_DEBUG, "Releasing all lchans (if any) after BSSMAP Clear Command\n");
-		gscon_release_lchans(conn, true, bsc_gsm48_rr_cause_from_gsm0808_cause(*cause_0808));
-		/* FIXME: Release all terestrial resources in ST_CLEARING */
-		/* According to 3GPP 48.008 3.1.9.1. "The BSS need not wait for the radio channel
-		 * release to be completed or for the guard timer to expire before returning the
-		 * CLEAR COMPLETE message" */
-
-		/* Close MGCP connections */
-		osmo_mgcpc_ep_clear(conn->user_plane.mgw_endpoint);
-
-		rate_ctr_inc(rate_ctr_group_get_ctr(conn->sccp.msc->msc_ctrs, MSC_CTR_BSSMAP_TX_DT1_CLEAR_COMPLETE));
-		gscon_sigtran_send(conn, gsm0808_create_clear_complete());
+		conn->clear_cause = *(const enum gsm0808_cause *)data;
+		conn_fsm_state_chg(ST_WAIT_SCCP_RLSD);
 		break;
 	case GSCON_EV_A_DISC_IND:
-		/* MSC or SIGTRAN network has hard-released SCCP connection,
-		 * terminate the FSM now. */
+		/* MSC or SIGTRAN network has hard-released SCCP connection, terminate the FSM now.
+		 * Cleanup is done in gscon_pre_term() and gscon_cleanup(). */
 		osmo_fsm_inst_term(fi, OSMO_FSM_TERM_REGULAR, data);
 		break;
 	case GSCON_EV_FORGET_MGW_ENDPOINT:
@@ -891,7 +988,9 @@ static void gscon_fsm_allstate(struct osmo_fsm_inst *fi, uint32_t event, void *d
 			conn->lchan->release.rr_cause =
 				bsc_gsm48_rr_cause_from_rsl_cause(conn->lchan->release.rsl_error_cause);
 		}
-		gscon_bssmap_clear(conn, GSM0808_CAUSE_RADIO_INTERFACE_FAILURE);
+		/* Request BSSMAP Clear, but do not abort an ongoing Location Request */
+		if (!conn->lcs.loc_req)
+			gscon_bssmap_clear(conn, GSM0808_CAUSE_RADIO_INTERFACE_FAILURE);
 		break;
 	case GSCON_EV_MGW_MDCX_RESP_MSC:
 		LOGPFSML(fi, LOGL_DEBUG, "Rx MDCX of MSC side (LCLS?)\n");
@@ -968,6 +1067,13 @@ static void gscon_pre_term(struct osmo_fsm_inst *fi, enum osmo_fsm_term_cause ca
 
 	osmo_mgcpc_ep_clear(conn->user_plane.mgw_endpoint);
 	conn->user_plane.mgw_endpoint = NULL;
+	conn->user_plane.mgw_endpoint = NULL;
+
+	if (conn->ho.fi)
+		osmo_fsm_inst_dispatch(conn->ho.fi, HO_EV_CONN_RELEASING, NULL);
+
+	if (conn->lcs.loc_req)
+		osmo_fsm_inst_dispatch(conn->lcs.loc_req->fi, LCS_LOC_REQ_EV_CONN_CLEAR, NULL);
 
 	if (conn->lcls.fi) {
 		/* request termination of LCLS FSM */
@@ -976,9 +1082,7 @@ static void gscon_pre_term(struct osmo_fsm_inst *fi, enum osmo_fsm_term_cause ca
 	}
 
 	LOGPFSML(fi, LOGL_DEBUG, "Releasing all lchans (if any) because this conn is terminating\n");
-	/* when things go smoothly, the lchan should have been released before FSM instance termination. So if this is
-	 * necessary it's probably "abnormal". */
-	gscon_release_lchans(conn, true, GSM48_RR_CAUSE_ABNORMAL_UNSPEC);
+	gscon_release_lchans(conn, true, bsc_gsm48_rr_cause_from_gsm0808_cause(conn->clear_cause));
 
 	/* drop pending messages */
 	gscon_dtap_queue_flush(conn, 0);
@@ -1008,8 +1112,9 @@ static int gscon_timer_cb(struct osmo_fsm_inst *fi)
 	case -4:
 		/* The MSC has sent a BSSMAP Clear Command, we acknowledged that, but the conn was never
 		 * disconnected. */
-		LOGPFSML(fi, LOGL_ERROR, "Long after a BSSMAP Clear Command, the conn is still not"
-			 " released. For sanity, discarding this conn now.\n");
+		LOGPFSML(fi, LOGL_ERROR, "Long after expecting %s, the conn is still not"
+			 " released. For sanity, discarding this conn now.\n",
+			 fi->state == ST_WAIT_CLEAR_CMD ? "BSSMAP Clear Command" : "SCCP RLSD");
 		a_reset_conn_fail(conn->sccp.msc);
 		osmo_fsm_inst_term(fi, OSMO_FSM_TERM_ERROR, NULL);
 		break;
@@ -1057,6 +1162,9 @@ struct gsm_subscriber_connection *bsc_subscr_con_allocate(struct gsm_network *ne
 	INIT_LLIST_HEAD(&conn->hodec2.penalty_timers);
 	conn->sccp.conn_id = -1;
 
+	/* Default clear cause (on RR translates to GSM48_RR_CAUSE_ABNORMAL_UNSPEC) */
+	conn->clear_cause = GSM0808_CAUSE_EQUIPMENT_FAILURE;
+
 	/* don't allocate from 'conn' context, as gscon_cleanup() will call talloc_free(conn) before
 	 * libosmocore will call talloc_free(conn->fi), i.e. avoid use-after-free during cleanup */
 	conn->fi = osmo_fsm_inst_alloc(&gscon_fsm, net, conn, LOGL_DEBUG, NULL);
@@ -1139,19 +1247,23 @@ static void rll_ind_cb(struct gsm_lchan *lchan, uint8_t link_id, void *_data, en
 
 	switch (rllr_ind) {
 	case BSC_RLLR_IND_EST_CONF:
-		rsl_data_request(msg, OBSC_LINKID_CB(msg));
+		rsl_data_request(msg, link_id);
 		break;
 	case BSC_RLLR_IND_REL_IND:
-		bsc_sapi_n_reject(lchan->conn, OBSC_LINKID_CB(msg),
+		bsc_sapi_n_reject(lchan->conn, RSL_LINK_ID2DLCI(link_id),
 				  GSM0808_CAUSE_MS_NOT_EQUIPPED);
 		msgb_free(msg);
 		break;
 	case BSC_RLLR_IND_ERR_IND:
 	case BSC_RLLR_IND_TIMEOUT:
-		bsc_sapi_n_reject(lchan->conn, OBSC_LINKID_CB(msg),
+		bsc_sapi_n_reject(lchan->conn, RSL_LINK_ID2DLCI(link_id),
 				  GSM0808_CAUSE_BSS_NOT_EQUIPPED);
 		msgb_free(msg);
 		break;
+	default:
+		LOGPLCHAN(lchan, DRLL, LOGL_NOTICE, "Received unknown rllr_ind %u\n", rllr_ind);
+		msgb_free(msg);
+		break;
 	}
 }
 
@@ -1188,7 +1300,7 @@ static void gsm0808_send_rsl_dtap(struct gsm_subscriber_connection *conn,
 		rc = rll_establish(msg->lchan, sapi, rll_ind_cb, msg);
 		if (rc) {
 			msgb_free(msg);
-			bsc_sapi_n_reject(conn, link_id, GSM0808_CAUSE_BSS_NOT_EQUIPPED);
+			bsc_sapi_n_reject(conn, RSL_LINK_ID2DLCI(link_id), GSM0808_CAUSE_BSS_NOT_EQUIPPED);
 			goto failed_to_send;
 		}
 		return;
@@ -1203,7 +1315,6 @@ static void gsm0808_send_rsl_dtap(struct gsm_subscriber_connection *conn,
 failed_to_send:
 	LOGPFSML(conn->fi, LOGL_ERROR, "Tx BSSMAP CLEAR REQUEST to MSC\n");
 	gscon_bssmap_clear(conn, GSM0808_CAUSE_EQUIPMENT_FAILURE);
-	osmo_fsm_inst_state_chg(conn->fi, ST_ACTIVE, 0, 0);
 }
 
 void gscon_submit_rsl_dtap(struct gsm_subscriber_connection *conn,
