@@ -1,6 +1,7 @@
 /* ip.access nanoBTS specific code */
 
 /* (C) 2009-2018 by Harald Welte <laforge@gnumonks.org>
+ * (C) 2020 by sysmocom s.f.m.c. GmbH <info@sysmocom.de>
  *
  * All Rights Reserved
  *
@@ -31,11 +32,13 @@
 #include <osmocom/gsm/tlv.h>
 #include <osmocom/core/msgb.h>
 #include <osmocom/core/socket.h>
+#include <osmocom/core/stat_item.h>
 #include <osmocom/core/talloc.h>
 #include <osmocom/core/utils.h>
 #include <osmocom/bsc/gsm_data.h>
 #include <osmocom/bsc/abis_nm.h>
 #include <osmocom/bsc/abis_rsl.h>
+#include <osmocom/bsc/abis_osmo.h>
 #include <osmocom/bsc/debug.h>
 #include <osmocom/abis/subchan_demux.h>
 #include <osmocom/gsm/ipa.h>
@@ -45,10 +48,17 @@
 #include <osmocom/bsc/bts_ipaccess_nanobts_omlattr.h>
 #include <osmocom/bsc/paging.h>
 #include <osmocom/bsc/timeslot_fsm.h>
+#include <osmocom/bsc/bts.h>
+#include <osmocom/bsc/bts_sm.h>
+#include <osmocom/bsc/nm_common_fsm.h>
+#include <osmocom/bsc/bsc_stats.h>
 
 static int bts_model_nanobts_start(struct gsm_network *net);
 static void bts_model_nanobts_e1line_bind_ops(struct e1inp_line *line);
 
+static int power_ctrl_send_def_params(const struct gsm_bts_trx *trx);
+static int power_ctrl_enc_rsl_params(struct msgb *msg, const struct gsm_power_ctrl_params *cp);
+
 static char *get_oml_status(const struct gsm_bts *bts)
 {
 	if (bts->oml_link)
@@ -64,6 +74,11 @@ struct gsm_bts_model bts_model_nanobts = {
 	.oml_rcvmsg = &abis_nm_rcvmsg,
 	.oml_status = &get_oml_status,
 	.e1line_bind_ops = bts_model_nanobts_e1line_bind_ops,
+
+	/* MS/BS Power control specific API */
+	.power_ctrl_send_def_params = &power_ctrl_send_def_params,
+	.power_ctrl_enc_rsl_params = &power_ctrl_enc_rsl_params,
+
 	/* Some nanoBTS firmwares (if not all) don't support SI2ter and cause
 	 * problems on some MS if it is enabled, see OS#3063. Disable it by
 	 * default, can still be enabled through VTY cmd with same name.
@@ -116,6 +131,7 @@ struct gsm_bts_model bts_model_nanobts = {
 			[NM_ATT_IPACC_REVOC_DATE] =	{ TLV_TYPE_TL16V },
 		},
 	},
+	.features_get_reported = true,
 };
 
 
@@ -124,125 +140,56 @@ static int nm_statechg_event(int evt, struct nm_statechg_signal_data *nsd)
 {
 	uint8_t obj_class = nsd->obj_class;
 	void *obj = nsd->obj;
-	struct gsm_nm_state *new_state = nsd->new_state;
 
+	struct gsm_bts_sm *bts_sm;
 	struct gsm_bts *bts;
 	struct gsm_bts_trx *trx;
+	struct gsm_bts_bb_trx *bb_transc;
 	struct gsm_bts_trx_ts *ts;
-	struct gsm_bts_gprs_nsvc *nsvc;
-
-	struct msgb *msgb;
+	struct gsm_gprs_nsvc *nsvc;
+	struct gsm_gprs_nse *nse;
+	struct gsm_gprs_cell *cell;
 
 	if (!is_ipaccess_bts(nsd->bts))
 		return 0;
 
-	/* This event-driven BTS setup is currently only required on nanoBTS */
-
-	/* S_NM_STATECHG_ADM is called after we call chg_adm_state() and would create
-	 * endless loop */
-	if (evt != S_NM_STATECHG_OPER)
-		return 0;
-
 	switch (obj_class) {
 	case NM_OC_SITE_MANAGER:
-		bts = container_of(obj, struct gsm_bts, site_mgr);
-		if ((new_state->operational == NM_OPSTATE_ENABLED &&
-		     new_state->availability == NM_AVSTATE_OK) ||
-		    (new_state->operational == NM_OPSTATE_DISABLED &&
-		     new_state->availability == NM_AVSTATE_OFF_LINE))
-			abis_nm_opstart(bts, obj_class, 0xff, 0xff, 0xff);
+		bts_sm = obj;
+		osmo_fsm_inst_dispatch(bts_sm->mo.fi, NM_EV_STATE_CHG_REP, nsd);
 		break;
 	case NM_OC_BTS:
 		bts = obj;
-		if (new_state->availability == NM_AVSTATE_DEPENDENCY) {
-			msgb = nanobts_attr_bts_get(bts);
-			abis_nm_set_bts_attr(bts, msgb->data, msgb->len);
-			msgb_free(msgb);
-			abis_nm_chg_adm_state(bts, obj_class,
-					      bts->bts_nr, 0xff, 0xff,
-					      NM_STATE_UNLOCKED);
-			abis_nm_opstart(bts, obj_class,
-					bts->bts_nr, 0xff, 0xff);
-		}
+		osmo_fsm_inst_dispatch(bts->mo.fi, NM_EV_STATE_CHG_REP, nsd);
+		break;
+	case NM_OC_BASEB_TRANSC:
+		bb_transc = obj;
+		osmo_fsm_inst_dispatch(bb_transc->mo.fi, NM_EV_STATE_CHG_REP, nsd);
 		break;
 	case NM_OC_CHANNEL:
 		ts = obj;
 		trx = ts->trx;
-		if (new_state->operational == NM_OPSTATE_DISABLED &&
-		    new_state->availability == NM_AVSTATE_DEPENDENCY) {
-			enum abis_nm_chan_comb ccomb =
-						abis_nm_chcomb4pchan(ts->pchan_from_config);
-			if (abis_nm_set_channel_attr(ts, ccomb) == -EINVAL) {
-				ipaccess_drop_oml_deferred(trx->bts);
-				return -1;
-			}
-			abis_nm_chg_adm_state(trx->bts, obj_class,
-					      trx->bts->bts_nr, trx->nr, ts->nr,
-					      NM_STATE_UNLOCKED);
-			abis_nm_opstart(trx->bts, obj_class,
-					trx->bts->bts_nr, trx->nr, ts->nr);
-		}
+		osmo_fsm_inst_dispatch(ts->mo.fi, NM_EV_STATE_CHG_REP, nsd);
 		break;
 	case NM_OC_RADIO_CARRIER:
 		trx = obj;
-		if (new_state->operational == NM_OPSTATE_DISABLED &&
-		    new_state->availability == NM_AVSTATE_OK)
-			abis_nm_opstart(trx->bts, obj_class, trx->bts->bts_nr,
-					trx->nr, 0xff);
+		osmo_fsm_inst_dispatch(trx->mo.fi, NM_EV_STATE_CHG_REP, nsd);
 		break;
 	case NM_OC_GPRS_NSE:
-		bts = container_of(obj, struct gsm_bts, gprs.nse);
-		if (bts->gprs.mode == BTS_GPRS_NONE)
-			break;
-		if (new_state->availability == NM_AVSTATE_DEPENDENCY) {
-			msgb = nanobts_attr_nse_get(bts);
-			abis_nm_ipaccess_set_attr(bts, obj_class, bts->bts_nr,
-						  0xff, 0xff, msgb->data,
-						  msgb->len);
-			msgb_free(msgb);
-			abis_nm_opstart(bts, obj_class, bts->bts_nr,
-					0xff, 0xff);
-		}
+		nse = obj;
+		osmo_fsm_inst_dispatch(nse->mo.fi, NM_EV_STATE_CHG_REP, nsd);
 		break;
 	case NM_OC_GPRS_CELL:
-		bts = container_of(obj, struct gsm_bts, gprs.cell);
-		if (bts->gprs.mode == BTS_GPRS_NONE)
-			break;
-		if (new_state->availability == NM_AVSTATE_DEPENDENCY) {
-			msgb = nanobts_attr_cell_get(bts);
-			abis_nm_ipaccess_set_attr(bts, obj_class, bts->bts_nr,
-						  0, 0xff, msgb->data,
-						  msgb->len);
-			msgb_free(msgb);
-			abis_nm_opstart(bts, obj_class, bts->bts_nr,
-					0, 0xff);
-			abis_nm_chg_adm_state(bts, obj_class, bts->bts_nr,
-					      0, 0xff, NM_STATE_UNLOCKED);
-			abis_nm_chg_adm_state(bts, NM_OC_GPRS_NSE, bts->bts_nr,
-					      0xff, 0xff, NM_STATE_UNLOCKED);
-		}
+		cell = obj;
+		osmo_fsm_inst_dispatch(cell->mo.fi, NM_EV_STATE_CHG_REP, nsd);
 		break;
 	case NM_OC_GPRS_NSVC:
 		nsvc = obj;
-		bts = nsvc->bts;
-		if (bts->gprs.mode == BTS_GPRS_NONE)
-			break;
 		/* We skip NSVC1 since we only use NSVC0 */
 		if (nsvc->id == 1)
 			break;
-		if ((new_state->availability == NM_AVSTATE_OFF_LINE) ||
-		    (new_state->availability == NM_AVSTATE_DEPENDENCY)) {
-			msgb = nanobts_attr_nscv_get(bts);
-			abis_nm_ipaccess_set_attr(bts, obj_class, bts->bts_nr,
-						  nsvc->id, 0xff,
-						  msgb->data, msgb->len);
-			msgb_free(msgb);
-			abis_nm_opstart(bts, obj_class, bts->bts_nr,
-					nsvc->id, 0xff);
-			abis_nm_chg_adm_state(bts, obj_class, bts->bts_nr,
-					      nsvc->id, 0xff,
-					      NM_STATE_UNLOCKED);
-		}
+		osmo_fsm_inst_dispatch(nsvc->mo.fi, NM_EV_STATE_CHG_REP, nsd);
+		break;
 	default:
 		break;
 	}
@@ -255,47 +202,46 @@ static int sw_activ_rep(struct msgb *mb)
 	struct abis_om_fom_hdr *foh = msgb_l3(mb);
 	struct e1inp_sign_link *sign_link = mb->dst;
 	struct gsm_bts *bts = sign_link->trx->bts;
-	struct gsm_bts_trx *trx = gsm_bts_trx_num(bts, foh->obj_inst.trx_nr);
+	struct gsm_bts_trx *trx;
+	struct gsm_gprs_nsvc *nsvc;
+	struct gsm_bts_trx_ts *ts;
 
-	if (!trx)
-		return -EINVAL;
-
-	if (!is_ipaccess_bts(trx->bts))
+	if (!is_ipaccess_bts(bts))
 		return 0;
 
 	switch (foh->obj_class) {
+	case NM_OC_SITE_MANAGER:
+		osmo_fsm_inst_dispatch(bts->site_mgr->mo.fi, NM_EV_SW_ACT_REP, NULL);
+		break;
+	case NM_OC_BTS:
+		osmo_fsm_inst_dispatch(bts->mo.fi, NM_EV_SW_ACT_REP, NULL);
+		break;
 	case NM_OC_BASEB_TRANSC:
-		abis_nm_chg_adm_state(trx->bts, foh->obj_class,
-				      trx->bts->bts_nr, trx->nr, 0xff,
-				      NM_STATE_UNLOCKED);
-		abis_nm_opstart(trx->bts, foh->obj_class,
-				trx->bts->bts_nr, trx->nr, 0xff);
-		/* TRX software is active, tell it to initiate RSL Link */
-		abis_nm_ipaccess_rsl_connect(trx, trx->bts->ip_access.rsl_ip,
-					     3003, trx->rsl_tei);
+		if (!(trx = gsm_bts_trx_num(bts, foh->obj_inst.trx_nr)))
+			return -EINVAL;
+		osmo_fsm_inst_dispatch(trx->bb_transc.mo.fi, NM_EV_SW_ACT_REP, NULL);
 		break;
-	case NM_OC_RADIO_CARRIER: {
-		/*
-		 * Locking the radio carrier will make it go
-		 * offline again and we would come here. The
-		 * framework should determine that there was
-		 * no change and avoid recursion.
-		 *
-		 * This code is here to make sure that on start
-		 * a TRX remains locked.
-		 */
-		int rc_state = trx->mo.nm_state.administrative;
-		/* Patch ARFCN into radio attribute */
-		struct msgb *msgb = nanobts_attr_radio_get(trx->bts, trx);
-		abis_nm_set_radio_attr(trx, msgb->data, msgb->len);
-		msgb_free(msgb);
-		abis_nm_chg_adm_state(trx->bts, foh->obj_class,
-				      trx->bts->bts_nr, trx->nr, 0xff,
-				      rc_state);
-		abis_nm_opstart(trx->bts, foh->obj_class, trx->bts->bts_nr,
-				trx->nr, 0xff);
+	case NM_OC_RADIO_CARRIER:
+		if (!(trx = gsm_bts_trx_num(bts, foh->obj_inst.trx_nr)))
+			return -EINVAL;
+		osmo_fsm_inst_dispatch(trx->mo.fi, NM_EV_SW_ACT_REP, NULL);
+		break;
+	case NM_OC_CHANNEL:
+		if (!(ts = abis_nm_get_ts(mb)))
+			return -EINVAL;
+		osmo_fsm_inst_dispatch(ts->mo.fi, NM_EV_SW_ACT_REP, NULL);
+		break;
+	case NM_OC_GPRS_NSE:
+		osmo_fsm_inst_dispatch(bts->site_mgr->gprs.nse.mo.fi, NM_EV_SW_ACT_REP, NULL);
+		break;
+	case NM_OC_GPRS_CELL:
+		osmo_fsm_inst_dispatch(bts->gprs.cell.mo.fi, NM_EV_SW_ACT_REP, NULL);
+		break;
+	case NM_OC_GPRS_NSVC:
+		if (!(nsvc = gsm_bts_sm_nsvc_num(bts->site_mgr, foh->obj_inst.trx_nr)))
+			return -EINVAL;
+		osmo_fsm_inst_dispatch(nsvc->mo.fi, NM_EV_SW_ACT_REP, NULL);
 		break;
-		}
 	}
 	return 0;
 }
@@ -311,22 +257,196 @@ static void nm_rx_opstart_ack_chan(struct msgb *oml_msg)
 		LOG_TS(ts, LOGL_ERROR, "Channel OPSTART ACK for uninitialized TS\n");
 		return;
 	}
-
+	osmo_fsm_inst_dispatch(ts->mo.fi, NM_EV_OPSTART_ACK, NULL);
 	osmo_fsm_inst_dispatch(ts->fi, TS_EV_OML_READY, NULL);
 }
 
 static void nm_rx_opstart_ack(struct msgb *oml_msg)
 {
 	struct abis_om_fom_hdr *foh = msgb_l3(oml_msg);
+	struct e1inp_sign_link *sign_link = oml_msg->dst;
+	struct gsm_bts *bts = sign_link->trx->bts;
+	struct gsm_bts_trx *trx;
+	struct gsm_gprs_nsvc *nsvc;
+
 	switch (foh->obj_class) {
+	case NM_OC_SITE_MANAGER:
+		osmo_fsm_inst_dispatch(bts->site_mgr->mo.fi, NM_EV_OPSTART_ACK, NULL);
+		break;
+	case NM_OC_BTS:
+		osmo_fsm_inst_dispatch(bts->mo.fi, NM_EV_OPSTART_ACK, NULL);
+		break;
+	case NM_OC_RADIO_CARRIER:
+		if (!(trx = gsm_bts_trx_num(bts, foh->obj_inst.trx_nr)))
+			return;
+		osmo_fsm_inst_dispatch(trx->mo.fi, NM_EV_OPSTART_ACK, NULL);
+		break;
+	case NM_OC_BASEB_TRANSC:
+		if (!(trx = gsm_bts_trx_num(bts, foh->obj_inst.trx_nr)))
+			return;
+		osmo_fsm_inst_dispatch(trx->bb_transc.mo.fi, NM_EV_OPSTART_ACK, NULL);
+		break;
 	case NM_OC_CHANNEL:
 		nm_rx_opstart_ack_chan(oml_msg);
 		break;
+	case NM_OC_GPRS_NSE:
+		osmo_fsm_inst_dispatch(bts->site_mgr->gprs.nse.mo.fi, NM_EV_OPSTART_ACK, NULL);
+		break;
+	case NM_OC_GPRS_CELL:
+		osmo_fsm_inst_dispatch(bts->gprs.cell.mo.fi, NM_EV_OPSTART_ACK, NULL);
+		break;
+	case NM_OC_GPRS_NSVC:
+		if (!(nsvc = gsm_bts_sm_nsvc_num(bts->site_mgr, foh->obj_inst.trx_nr)))
+			return;
+		osmo_fsm_inst_dispatch(nsvc->mo.fi, NM_EV_OPSTART_ACK, NULL);
+		break;
 	default:
 		break;
 	}
 }
 
+static void nm_rx_opstart_nack(struct msgb *oml_msg)
+{
+	struct abis_om_fom_hdr *foh = msgb_l3(oml_msg);
+	struct e1inp_sign_link *sign_link = oml_msg->dst;
+	struct gsm_bts *bts = sign_link->trx->bts;
+	struct gsm_bts_trx *trx;
+	struct gsm_bts_trx_ts *ts;
+	struct gsm_gprs_nsvc *nsvc;
+
+	switch (foh->obj_class) {
+	case NM_OC_SITE_MANAGER:
+		osmo_fsm_inst_dispatch(bts->site_mgr->mo.fi, NM_EV_OPSTART_NACK, NULL);
+		break;
+	case NM_OC_BTS:
+		osmo_fsm_inst_dispatch(bts->mo.fi, NM_EV_OPSTART_ACK, NULL);
+		break;
+	case NM_OC_RADIO_CARRIER:
+		if (!(trx = gsm_bts_trx_num(bts, foh->obj_inst.trx_nr)))
+			return;
+		osmo_fsm_inst_dispatch(trx->mo.fi, NM_EV_OPSTART_NACK, NULL);
+		break;
+	case NM_OC_BASEB_TRANSC:
+		if (!(trx = gsm_bts_trx_num(bts, foh->obj_inst.trx_nr)))
+			return;
+		osmo_fsm_inst_dispatch(trx->bb_transc.mo.fi, NM_EV_OPSTART_NACK, NULL);
+		break;
+	case NM_OC_CHANNEL:
+		if (!(ts = abis_nm_get_ts(oml_msg)))
+			return;
+		osmo_fsm_inst_dispatch(ts->mo.fi, NM_EV_OPSTART_NACK, NULL);
+		break;
+	case NM_OC_GPRS_NSE:
+		osmo_fsm_inst_dispatch(bts->site_mgr->gprs.nse.mo.fi, NM_EV_OPSTART_NACK, NULL);
+		break;
+	case NM_OC_GPRS_CELL:
+		osmo_fsm_inst_dispatch(bts->gprs.cell.mo.fi, NM_EV_OPSTART_NACK, NULL);
+		break;
+	case NM_OC_GPRS_NSVC:
+		if (!(nsvc = gsm_bts_sm_nsvc_num(bts->site_mgr, foh->obj_inst.trx_nr)))
+			return;
+		osmo_fsm_inst_dispatch(nsvc->mo.fi, NM_EV_OPSTART_NACK, NULL);
+		break;
+	default:
+		break;
+	}
+}
+
+static void nm_rx_get_attr_rep(struct msgb *oml_msg)
+{
+	struct abis_om_fom_hdr *foh = msgb_l3(oml_msg);
+	struct e1inp_sign_link *sign_link = oml_msg->dst;
+	struct gsm_bts *bts = sign_link->trx->bts;
+	struct gsm_bts_trx *trx;
+
+	switch (foh->obj_class) {
+	case NM_OC_BTS:
+		osmo_fsm_inst_dispatch(bts->mo.fi, NM_EV_GET_ATTR_REP, NULL);
+		break;
+	case NM_OC_BASEB_TRANSC:
+		if (!(trx = gsm_bts_trx_num(bts, foh->obj_inst.trx_nr)))
+			return;
+		osmo_fsm_inst_dispatch(trx->bb_transc.mo.fi, NM_EV_GET_ATTR_REP, NULL);
+		break;
+	default:
+		LOGPFOH(DNM, LOGL_ERROR, foh, "Get Attributes Response received on incorrect object class %d!\n", foh->obj_class);
+	}
+}
+
+static void nm_rx_set_bts_attr_ack(struct msgb *oml_msg)
+{
+	struct abis_om_fom_hdr *foh = msgb_l3(oml_msg);
+	struct e1inp_sign_link *sign_link = oml_msg->dst;
+	struct gsm_bts *bts = sign_link->trx->bts;
+
+	if (foh->obj_class != NM_OC_BTS) {
+		LOG_BTS(bts, DNM, LOGL_ERROR, "Set BTS Attr Ack received on non BTS object!\n");
+		return;
+	}
+	osmo_fsm_inst_dispatch(bts->mo.fi, NM_EV_SET_ATTR_ACK, NULL);
+}
+
+
+static void nm_rx_set_radio_attr_ack(struct msgb *oml_msg)
+{
+	struct abis_om_fom_hdr *foh = msgb_l3(oml_msg);
+	struct e1inp_sign_link *sign_link = oml_msg->dst;
+	struct gsm_bts *bts = sign_link->trx->bts;
+	struct gsm_bts_trx *trx = gsm_bts_trx_num(bts, foh->obj_inst.trx_nr);
+
+	if (!trx || foh->obj_class != NM_OC_RADIO_CARRIER) {
+		LOGPFOH(DNM, LOGL_ERROR, foh, "Set Radio Carrier Attr Ack received on non Radio Carrier object!\n");
+		return;
+	}
+	osmo_fsm_inst_dispatch(trx->mo.fi, NM_EV_SET_ATTR_ACK, NULL);
+}
+
+static void nm_rx_set_chan_attr_ack(struct msgb *oml_msg)
+{
+	struct abis_om_fom_hdr *foh = msgb_l3(oml_msg);
+	struct gsm_bts_trx_ts *ts = abis_nm_get_ts(oml_msg);
+
+	if (!ts || foh->obj_class != NM_OC_CHANNEL) {
+		LOGPFOH(DNM, LOGL_ERROR, foh, "Set Channel Attr Ack received on non Radio Channel object!\n");
+		return;
+	}
+	osmo_fsm_inst_dispatch(ts->mo.fi, NM_EV_SET_ATTR_ACK, NULL);
+}
+
+static void nm_rx_ipacc_set_attr_ack(struct msgb *oml_msg)
+{
+	struct e1inp_sign_link *sign_link = oml_msg->dst;
+	struct gsm_bts *bts = sign_link->trx->bts;
+	struct abis_om_hdr *oh = msgb_l2(oml_msg);
+	uint8_t idstrlen = oh->data[0];
+	struct abis_om_fom_hdr *foh;
+	void *obj;
+	struct gsm_gprs_nse *nse;
+	struct gsm_gprs_cell *cell;
+	struct gsm_gprs_nsvc *nsvc;
+
+	foh = (struct abis_om_fom_hdr *) (oh->data + 1 + idstrlen);
+	obj = gsm_objclass2obj(bts, foh->obj_class, &foh->obj_inst);
+
+	switch (foh->obj_class) {
+	case NM_OC_GPRS_NSE:
+		nse = obj;
+		osmo_fsm_inst_dispatch(nse->mo.fi, NM_EV_SET_ATTR_ACK, NULL);
+		break;
+	case NM_OC_GPRS_CELL:
+		cell = obj;
+		osmo_fsm_inst_dispatch(cell->mo.fi, NM_EV_SET_ATTR_ACK, NULL);
+		break;
+	case NM_OC_GPRS_NSVC:
+		if (!(nsvc = gsm_bts_sm_nsvc_num(bts->site_mgr, foh->obj_inst.trx_nr)))
+			return;
+		osmo_fsm_inst_dispatch(nsvc->mo.fi, NM_EV_SET_ATTR_ACK, NULL);
+		break;
+	default:
+		LOGPFOH(DNM, LOGL_ERROR, foh, "IPACC Set Attr Ack received on incorrect object class %d!\n", foh->obj_class);
+	}
+}
+
 /* Callback function to be called every time we receive a signal from NM */
 static int bts_ipa_nm_sig_cb(unsigned int subsys, unsigned int signal,
 		     void *handler_data, void *signal_data)
@@ -337,12 +457,29 @@ static int bts_ipa_nm_sig_cb(unsigned int subsys, unsigned int signal,
 	switch (signal) {
 	case S_NM_SW_ACTIV_REP:
 		return sw_activ_rep(signal_data);
-	case S_NM_STATECHG_OPER:
-	case S_NM_STATECHG_ADM:
+	case S_NM_STATECHG:
 		return nm_statechg_event(signal, signal_data);
 	case S_NM_OPSTART_ACK:
 		nm_rx_opstart_ack(signal_data);
 		return 0;
+	case S_NM_OPSTART_NACK:
+		nm_rx_opstart_nack(signal_data);
+		return 0;
+	case S_NM_GET_ATTR_REP:
+		nm_rx_get_attr_rep(signal_data);
+		return 0;
+	case S_NM_SET_BTS_ATTR_ACK:
+		nm_rx_set_bts_attr_ack(signal_data);
+		return 0;
+	case S_NM_SET_RADIO_ATTR_ACK:
+		nm_rx_set_radio_attr_ack(signal_data);
+		return 0;
+	case S_NM_SET_CHAN_ATTR_ACK:
+		nm_rx_set_chan_attr_ack(signal_data);
+		return 0;
+	case S_NM_IPACC_SET_ATTR_ACK:
+		nm_rx_ipacc_set_attr_ack(signal_data);
+		return 0;
 	default:
 		break;
 	}
@@ -391,12 +528,13 @@ find_bts_by_unitid(struct gsm_network *net, uint16_t site_id, uint16_t bts_id)
 /* These are exported because they are used by the VTY interface. */
 void ipaccess_drop_rsl(struct gsm_bts_trx *trx, const char *reason)
 {
-	if (!trx->rsl_link)
+	if (!trx->rsl_link_primary)
 		return;
 
 	LOG_TRX(trx, DLINP, LOGL_NOTICE, "Dropping RSL link: %s\n", reason);
-	e1inp_sign_link_destroy(trx->rsl_link);
-	trx->rsl_link = NULL;
+	e1inp_sign_link_destroy(trx->rsl_link_primary);
+	trx->rsl_link_primary = NULL;
+	osmo_stat_item_dec(osmo_stat_item_group_get_item(trx->bts->bts_statg, BTS_STAT_RSL_CONNECTED), 1);
 
 	if (trx->bts->c0 == trx)
 		paging_flush_bts(trx->bts, NULL);
@@ -406,6 +544,9 @@ void ipaccess_drop_oml(struct gsm_bts *bts, const char *reason)
 {
 	struct gsm_bts *rdep_bts;
 	struct gsm_bts_trx *trx;
+	struct gsm_bts_trx_ts *ts ;
+	uint8_t tn;
+	uint8_t i;
 
 	/* First of all, remove deferred drop if enabled */
 	osmo_timer_del(&bts->oml_drop_link_timer);
@@ -417,10 +558,31 @@ void ipaccess_drop_oml(struct gsm_bts *bts, const char *reason)
 	e1inp_sign_link_destroy(bts->oml_link);
 	bts->oml_link = NULL;
 	bts->uptime = 0;
+	osmo_stat_item_dec(osmo_stat_item_group_get_item(bts->bts_statg, BTS_STAT_OML_CONNECTED), 1);
+	gsm_bts_stats_reset(bts);
+
+	/* Also drop the associated OSMO link */
+	e1inp_sign_link_destroy(bts->osmo_link);
+	bts->osmo_link = NULL;
 
 	/* we have issues reconnecting RSL, drop everything. */
-	llist_for_each_entry(trx, &bts->trx_list, list)
+	llist_for_each_entry(trx, &bts->trx_list, list) {
 		ipaccess_drop_rsl(trx, "OML link drop");
+		osmo_fsm_inst_dispatch(trx->bb_transc.mo.fi, NM_EV_OML_DOWN, NULL);
+		osmo_fsm_inst_dispatch(trx->mo.fi, NM_EV_OML_DOWN, NULL);
+		for (tn = 0; tn < TRX_NR_TS; tn++) {
+			ts = &trx->ts[tn];
+			osmo_fsm_inst_dispatch(ts->mo.fi, NM_EV_OML_DOWN, NULL);
+		}
+	}
+
+	osmo_fsm_inst_dispatch(bts->mo.fi, NM_EV_OML_DOWN, NULL);
+	osmo_fsm_inst_dispatch(bts->gprs.cell.mo.fi, NM_EV_OML_DOWN, NULL);
+
+	osmo_fsm_inst_dispatch(bts->site_mgr->mo.fi, NM_EV_OML_DOWN, NULL);
+	osmo_fsm_inst_dispatch(bts->site_mgr->gprs.nse.mo.fi, NM_EV_OML_DOWN, NULL);
+	for (i = 0; i < ARRAY_SIZE(bts->site_mgr->gprs.nsvc); i++)
+		osmo_fsm_inst_dispatch(bts->site_mgr->gprs.nsvc[i].mo.fi, NM_EV_OML_DOWN, NULL);
 
 	gsm_bts_all_ts_dispatch(bts, TS_EV_OML_DOWN, NULL);
 
@@ -478,7 +640,7 @@ static void ipaccess_sign_link_reject(const struct ipaccess_unit *dev, const str
 	/* Write to log and increase counter */
 	LOGP(DLINP, LOGL_ERROR, "Unable to find BTS configuration for %u/%u/%u, disconnecting\n", site_id, bts_id,
 		trx_id);
-	rate_ctr_inc(&bsc_gsmnet->bsc_ctrs->ctr[BSC_CTR_UNKNOWN_UNIT_ID]);
+	rate_ctr_inc(rate_ctr_group_get_ctr(bsc_gsmnet->bsc_ctrs, BSC_CTR_UNKNOWN_UNIT_ID));
 
 	/* Get remote IP */
 	if (osmo_sock_get_remote_ip(ts->driver.ipaccess.fd.fd, ip, sizeof(ip)))
@@ -525,15 +687,24 @@ ipaccess_sign_link_up(void *unit_data, struct e1inp_line *line,
 	struct e1inp_sign_link *sign_link = NULL;
 	struct timespec tp;
 	int rc;
+	struct e1inp_ts *sign_ts = e1inp_line_ipa_oml_ts(line);
 
 	bts = find_bts_by_unitid(bsc_gsmnet, dev->site_id, dev->bts_id);
 	if (!bts) {
-		ipaccess_sign_link_reject(dev, &line->ts[E1INP_SIGN_OML - 1]);
+		ipaccess_sign_link_reject(dev, sign_ts);
 		return NULL;
 	}
-	DEBUGP(DLINP, "Identified BTS %u/%u/%u\n",
+	DEBUGP(DLINP, "%s: Identified BTS %u/%u/%u\n", e1inp_signtype_name(type),
 			dev->site_id, dev->bts_id, dev->trx_id);
 
+	/* Check if this BTS has a valid configuration. If not we will drop it
+	 * immediately. */
+	if (gsm_bts_check_cfg(bts) != 0) {
+		LOGP(DLINP, LOGL_NOTICE, "(bts=%u) BTS config invalid, dropping BTS!\n", bts->nr);
+		ipaccess_drop_oml_deferred(bts);
+		return NULL;
+	}
+
 	switch(type) {
 	case E1INP_SIGN_OML:
 		/* remove old OML signal link for this BTS. */
@@ -548,16 +719,21 @@ ipaccess_sign_link_up(void *unit_data, struct e1inp_line *line,
 
 		/* create new OML link. */
 		sign_link = bts->oml_link =
-			e1inp_sign_link_create(&line->ts[E1INP_SIGN_OML - 1],
+			e1inp_sign_link_create(sign_ts,
 						E1INP_SIGN_OML, bts->c0,
 						bts->oml_tei, 0);
-		rc = clock_gettime(CLOCK_MONOTONIC, &tp);
+		rc = osmo_clock_gettime(CLOCK_MONOTONIC, &tp);
 		bts->uptime = (rc < 0) ? 0 : tp.tv_sec; /* we don't need sub-second precision for uptime */
 		if (!(sign_link->trx->bts->ip_access.flags & OML_UP)) {
 			e1inp_event(sign_link->ts, S_L_INP_TEI_UP,
 					sign_link->tei, sign_link->sapi);
 			sign_link->trx->bts->ip_access.flags |= OML_UP;
 		}
+		osmo_stat_item_inc(osmo_stat_item_group_get_item(bts->bts_statg, BTS_STAT_OML_CONNECTED), 1);
+
+		/* Create link for E1INP_SIGN_OSMO */
+		//SAPI must be 0, no IPAC_PROTO_EXT_PCU, see ipaccess_bts_read_cb
+		bts->osmo_link = e1inp_sign_link_create(sign_ts, E1INP_SIGN_OSMO, bts->c0, IPAC_PROTO_OSMO, 0);
 		break;
 	case E1INP_SIGN_RSL: {
 		struct e1inp_ts *ts;
@@ -572,12 +748,12 @@ ipaccess_sign_link_up(void *unit_data, struct e1inp_line *line,
 
 		/* set new RSL link for this TRX. */
 		line = bts->oml_link->ts->line;
-		ts = &line->ts[E1INP_SIGN_RSL + dev->trx_id - 1];
+		ts = e1inp_line_ipa_rsl_ts(line, dev->trx_id);
 		e1inp_ts_config_sign(ts, line);
-		sign_link = trx->rsl_link =
+		sign_link = trx->rsl_link_primary =
 				e1inp_sign_link_create(ts, E1INP_SIGN_RSL,
-						       trx, trx->rsl_tei, 0);
-		trx->rsl_link->ts->sign.delay = 0;
+						       trx, trx->rsl_tei_primary, 0);
+		trx->rsl_link_primary->ts->sign.delay = 0;
 		if (!(sign_link->trx->bts->ip_access.flags &
 					(RSL_UP << sign_link->trx->nr))) {
 			e1inp_event(sign_link->ts, S_L_INP_TEI_UP,
@@ -585,6 +761,7 @@ ipaccess_sign_link_up(void *unit_data, struct e1inp_line *line,
 			sign_link->trx->bts->ip_access.flags |=
 					(RSL_UP << sign_link->trx->nr);
 		}
+		osmo_stat_item_inc(osmo_stat_item_group_get_item(bts->bts_statg, BTS_STAT_RSL_CONNECTED), 1);
 		break;
 	}
 	default:
@@ -596,10 +773,12 @@ ipaccess_sign_link_up(void *unit_data, struct e1inp_line *line,
 static void ipaccess_sign_link_down(struct e1inp_line *line)
 {
 	/* No matter what link went down, we close both signal links. */
-	struct e1inp_ts *ts = &line->ts[E1INP_SIGN_OML-1];
+	struct e1inp_ts *ts = e1inp_line_ipa_oml_ts(line);
 	struct gsm_bts *bts = NULL;
 	struct e1inp_sign_link *link;
 
+	LOGPIL(line, DLINP, LOGL_NOTICE, "Signalling link down\n");
+
 	llist_for_each_entry(link, &ts->sign.sign_links, list) {
 		/* Get bts pointer from the first element of the list. */
 		if (bts == NULL)
@@ -610,6 +789,8 @@ static void ipaccess_sign_link_down(struct e1inp_line *line)
 	}
 	if (bts != NULL)
 		ipaccess_drop_oml(bts, "link down");
+	else
+		LOGPIL(line, DLINP, LOGL_NOTICE, "Signalling link down for unknown BTS\n");
 }
 
 /* This function is called if we receive one OML/RSL message. */
@@ -625,6 +806,9 @@ static int ipaccess_sign_link(struct msgb *msg)
 	case E1INP_SIGN_OML:
 	        ret = abis_nm_rcvmsg(msg);
 	        break;
+	case E1INP_SIGN_OSMO:
+	        ret = abis_osmo_rcvmsg(msg);
+	        break;
 	default:
 		LOGP(DLINP, LOGL_ERROR, "Unknown signal link type %d\n",
 			link->type);
@@ -651,3 +835,153 @@ static void bts_model_nanobts_e1line_bind_ops(struct e1inp_line *line)
 {
         e1inp_line_bind_ops(line, &ipaccess_e1inp_line_ops);
 }
+
+static void enc_meas_proc_params(struct msgb *msg, uint8_t ptype,
+				 const struct gsm_power_ctrl_meas_params *mp)
+{
+	struct ipac_preproc_ave_cfg *ave_cfg;
+	uint8_t *ie_len;
+
+	/* No averaging => no Measurement Averaging parameters */
+	if (mp->algo == GSM_PWR_CTRL_MEAS_AVG_ALGO_NONE)
+		return;
+
+	/* (TLV) Measurement Averaging parameters for RxLev/RxQual */
+	ie_len = msgb_tl_put(msg, RSL_IPAC_EIE_MEAS_AVG_CFG);
+
+	ave_cfg = (struct ipac_preproc_ave_cfg *) msgb_put(msg, sizeof(*ave_cfg));
+	ave_cfg->param_id = ptype & 0x03;
+
+	/* H_REQAVE and H_REQT */
+	ave_cfg->h_reqave = mp->h_reqave & 0x1f;
+	ave_cfg->h_reqt = mp->h_reqt & 0x1f;
+
+	/* Averaging method and parameters */
+	ave_cfg->ave_method = (mp->algo - 1) & 0x07;
+	switch (mp->algo) {
+	case GSM_PWR_CTRL_MEAS_AVG_ALGO_OSMO_EWMA:
+		msgb_v_put(msg, mp->ewma.alpha);
+		break;
+	case GSM_PWR_CTRL_MEAS_AVG_ALGO_WEIGHTED:
+	case GSM_PWR_CTRL_MEAS_AVG_ALGO_MOD_MEDIAN:
+		/* FIXME: unknown format */
+		break;
+	case GSM_PWR_CTRL_MEAS_AVG_ALGO_UNWEIGHTED:
+	case GSM_PWR_CTRL_MEAS_AVG_ALGO_NONE:
+		/* No parameters here */
+		break;
+	}
+
+	/* Update length part of the containing IE */
+	*ie_len = msg->tail - (ie_len + 1);
+}
+
+static void enc_power_params(struct msgb *msg, const struct gsm_power_ctrl_params *cp)
+{
+	struct ipac_preproc_pc_comp *thresh_comp;
+	struct ipac_preproc_pc_thresh *thresh;
+
+	/* These parameters are valid for dynamic mode only */
+	OSMO_ASSERT(cp->mode == GSM_PWR_CTRL_MODE_DYN_BTS);
+
+	/* (TLV) Measurement Averaging Configure */
+	enc_meas_proc_params(msg, IPAC_RXQUAL_AVE, &cp->rxqual_meas);
+	enc_meas_proc_params(msg, IPAC_RXLEV_AVE, &cp->rxlev_meas);
+
+	/* (TV) Thresholds: {L,U}_RXLEV_XX_P and {L,U}_RXQUAL_XX_P */
+	if (cp->dir == GSM_PWR_CTRL_DIR_UL)
+		msgb_v_put(msg, RSL_IPAC_EIE_MS_PWR_CTL);
+	else
+		msgb_v_put(msg, RSL_IPAC_EIE_BS_PWR_CTL);
+
+	thresh = (struct ipac_preproc_pc_thresh *) msgb_put(msg, sizeof(*thresh));
+
+	/* {L,U}_RXLEV_XX_P (see 3GPP TS 45.008, A.3.2.1, a & b) */
+	thresh->l_rxlev = cp->rxlev_meas.lower_thresh & 0x3f;
+	thresh->u_rxlev = cp->rxlev_meas.upper_thresh & 0x3f;
+
+	/* {L,U}_RXQUAL_XX_P (see 3GPP TS 45.008, A.3.2.1, c & d) */
+	thresh->l_rxqual = cp->rxqual_meas.lower_thresh & 0x07;
+	thresh->u_rxqual = cp->rxqual_meas.upper_thresh & 0x07;
+
+	/* (TV) PC Threshold Comparators */
+	msgb_v_put(msg, RSL_IPAC_EIE_PC_THRESH_COMP);
+
+	thresh_comp = (struct ipac_preproc_pc_comp *) msgb_put(msg, sizeof(*thresh_comp));
+
+	/* RxLev: P1, N1, P2, N2 (see 3GPP TS 45.008, A.3.2.1, a & b) */
+	thresh_comp->p1 = cp->rxlev_meas.lower_cmp_p & 0x1f;
+	thresh_comp->n1 = cp->rxlev_meas.lower_cmp_n & 0x1f;
+	thresh_comp->p2 = cp->rxlev_meas.upper_cmp_p & 0x1f;
+	thresh_comp->n2 = cp->rxlev_meas.upper_cmp_n & 0x1f;
+
+	/* RxQual: P3, N3, P4, N4 (see 3GPP TS 45.008, A.3.2.1, c & d) */
+	thresh_comp->p3 = cp->rxqual_meas.lower_cmp_p & 0x1f;
+	thresh_comp->n3 = cp->rxqual_meas.lower_cmp_n & 0x1f;
+	thresh_comp->p4 = cp->rxqual_meas.upper_cmp_p & 0x1f;
+	thresh_comp->n4 = cp->rxqual_meas.upper_cmp_n & 0x1f;
+
+	/* Minimum interval between power level changes (P_CON_INTERVAL) */
+	thresh_comp->pc_interval = cp->ctrl_interval;
+
+	/* Change step limitations: POWER_{INC,RED}_STEP_SIZE */
+	thresh_comp->inc_step_size = cp->inc_step_size_db & 0x0f;
+	thresh_comp->red_step_size = cp->red_step_size_db & 0x0f;
+}
+
+void osmobts_enc_power_params_osmo_ext(struct msgb *msg, const struct gsm_power_ctrl_params *cp);
+static void add_power_params_ie(struct msgb *msg, enum abis_rsl_ie iei,
+				const struct gsm_bts_trx *trx,
+				const struct gsm_power_ctrl_params *cp)
+{
+	uint8_t *ie_len = msgb_tl_put(msg, iei);
+	uint8_t msg_len = msgb_length(msg);
+
+	enc_power_params(msg, cp);
+	if (iei == RSL_IE_MS_POWER_PARAM && is_osmobts(trx->bts))
+		osmobts_enc_power_params_osmo_ext(msg, cp);
+
+	*ie_len = msgb_length(msg) - msg_len;
+}
+
+static int power_ctrl_send_def_params(const struct gsm_bts_trx *trx)
+{
+	const struct gsm_power_ctrl_params *ms_power_ctrl = &trx->bts->ms_power_ctrl;
+	const struct gsm_power_ctrl_params *bs_power_ctrl = &trx->bts->bs_power_ctrl;
+	struct abis_rsl_common_hdr *ch;
+	struct msgb *msg;
+
+	/* Sending this message does not make sense if neither MS Power control
+	 * nor BS Power control is to be performed by the BTS itself ('dyn-bts'). */
+	if (ms_power_ctrl->mode != GSM_PWR_CTRL_MODE_DYN_BTS &&
+	    bs_power_ctrl->mode != GSM_PWR_CTRL_MODE_DYN_BTS)
+		return 0;
+
+	msg = rsl_msgb_alloc();
+	if (msg == NULL)
+		return -ENOMEM;
+
+	ch = (struct abis_rsl_common_hdr *) msgb_put(msg, sizeof(*ch));
+	ch->msg_discr = ABIS_RSL_MDISC_TRX;
+	ch->msg_type = RSL_MT_IPAC_MEAS_PREPROC_DFT;
+
+	/* BS/MS Power IEs (to be re-defined in channel specific messages) */
+	msgb_tv_put(msg, RSL_IE_MS_POWER, 0); /* dummy value */
+	msgb_tv_put(msg, RSL_IE_BS_POWER, 0); /* dummy value */
+
+	/* MS/BS Power Parameters IEs */
+	if (ms_power_ctrl->mode == GSM_PWR_CTRL_MODE_DYN_BTS)
+		add_power_params_ie(msg, RSL_IE_MS_POWER_PARAM, trx, ms_power_ctrl);
+	if (bs_power_ctrl->mode == GSM_PWR_CTRL_MODE_DYN_BTS)
+		add_power_params_ie(msg, RSL_IE_BS_POWER_PARAM, trx, bs_power_ctrl);
+
+	msg->dst = trx->rsl_link_primary;
+
+	return abis_rsl_sendmsg(msg);
+}
+
+static int power_ctrl_enc_rsl_params(struct msgb *msg, const struct gsm_power_ctrl_params *cp)
+{
+	/* We send everything in "Measurement Pre-processing Defaults" */
+	return 0;
+}
