@@ -36,16 +36,22 @@
 #include <osmocom/abis/e1_input.h>
 #include <osmocom/bsc/signal.h>
 #include <osmocom/bsc/timeslot_fsm.h>
+#include <osmocom/bsc/system_information.h>
+#include <osmocom/bsc/bts.h>
 
 #include <osmocom/core/timer.h>
 
 #include <osmocom/abis/lapd.h>
 
+enum reset_timer_state {
+	RESET_T_NONE = 0,
+	RESET_T_STOP_LAPD = 1,		/* first timer expiration: stop LAPD SAP */
+	RESET_T_RESTART_LAPD = 2,	/* second timer expiration: restart LAPD SAP */
+};
+
 /* TODO: put in a separate file ? */
 
 extern int abis_nm_sendmsg(struct gsm_bts *bts, struct msgb *msg);
-/* was static in system_information.c */
-extern int generate_cell_chan_list(uint8_t * chan_list, struct gsm_bts *bts);
 
 static void nokia_abis_nm_queue_send_next(struct gsm_bts *bts);
 static void reset_timer_cb(void *_bts);
@@ -54,7 +60,7 @@ static int dump_elements(uint8_t * data, int len) __attribute__((unused));
 
 static void bootstrap_om_bts(struct gsm_bts *bts)
 {
-	LOGP(DNM, LOGL_NOTICE, "bootstrapping OML for BTS %u\n", bts->nr);
+	LOG_BTS(bts, DNM, LOGL_NOTICE, "bootstrapping OML\n");
 
 	if (!bts->nokia.skip_reset) {
 		if (!bts->nokia.did_reset)
@@ -67,14 +73,14 @@ static void bootstrap_om_bts(struct gsm_bts *bts)
 
 static void bootstrap_om_trx(struct gsm_bts_trx *trx)
 {
-	LOGP(DNM, LOGL_NOTICE, "bootstrapping OML for TRX %u/%u\n",
-	     trx->bts->nr, trx->nr);
+	LOG_TRX(trx, DNM, LOGL_NOTICE, "bootstrapping OML\n");
 
 	gsm_trx_all_ts_dispatch(trx, TS_EV_OML_READY, NULL);
 }
 
 static int shutdown_om(struct gsm_bts *bts)
 {
+	gsm_bts_stats_reset(bts);
 	/* TODO !? */
 	return 0;
 }
@@ -198,8 +204,7 @@ static int nm_sig_cb(unsigned int subsys, unsigned int signal,
 		return 0;
 
 	switch (signal) {
-	case S_NM_STATECHG_OPER:
-	case S_NM_STATECHG_ADM:
+	case S_NM_STATECHG:
 		nm_statechg_evt(signal, signal_data);
 		break;
 	default:
@@ -757,7 +762,7 @@ static int make_fu_config(struct gsm_bts_trx *trx, uint8_t id,
 	/* set CA */
 
 	if (generate_cell_chan_list(&fu_config[38], trx->bts) != 0) {
-		fprintf(stderr, "generate_cell_chan_list failed\n");
+		LOG_TRX(trx, DNM, LOGL_ERROR, "generate_cell_chan_list failed\n");
 		return 0;
 	}
 
@@ -805,8 +810,7 @@ static int make_fu_config(struct gsm_bts_trx *trx, uint8_t id,
 			chan_config = 11;
 			break;
 		default:
-			fprintf(stderr,
-				"unsupported channel config %s for timeslot %d\n",
+			LOG_TRX(trx, DNM, LOGL_ERROR, "unsupported channel config %s for timeslot %d\n",
 				gsm_pchan_name(ts->pchan_from_config), i);
 			return 0;
 		}
@@ -995,17 +999,17 @@ void set_real_time(uint8_t * real_time)
   build the configuration data
 */
 
-static int make_bts_config(uint8_t bts_type, int n_trx, uint8_t * fu_config,
+static int make_bts_config(struct gsm_bts *bts, uint8_t bts_type, int n_trx, uint8_t * fu_config,
 			   int need_hopping)
 {
 	/* is it an InSite BTS ? */
 	if (bts_type == 0x0E || bts_type == 0x0F || bts_type == 0x10) {	/* TODO */
 		if (n_trx != 1) {
-			fprintf(stderr, "InSite has only one TRX\n");
+			LOG_BTS(bts, DNM, LOGL_ERROR, "InSite has only one TRX\n");
 			return 0;
 		}
 		if (need_hopping != 0) {
-			fprintf(stderr, "InSite does not support hopping\n");
+			LOG_BTS(bts, DNM, LOGL_ERROR, "InSite does not support hopping\n");
 			return 0;
 		}
 		memcpy(fu_config, bts_config_insite, sizeof(bts_config_insite));
@@ -1085,7 +1089,7 @@ static int abis_nm_send(struct gsm_bts *bts, uint8_t msg_type, uint16_t ref,
 	noh->reference = htons(ref);
 	memcpy(noh->data, data, len_data);
 
-	DEBUGPC(DNM, "Sending %s\n", get_msg_type_name_string(msg_type));
+	LOG_BTS(bts, DNM, LOGL_DEBUG, "Sending %s\n", get_msg_type_name_string(msg_type));
 
 	return abis_nm_sendmsg(bts, msg);
 }
@@ -1160,7 +1164,7 @@ static int abis_nm_reset(struct gsm_bts *bts, uint16_t ref)
 {
 	uint8_t *data = reset;
 	int len_data = sizeof(reset);
-	LOGP(DLINP, LOGL_INFO, "Nokia BTS reset timer: %d\n", bts->nokia.bts_reset_timer_cnf);
+	LOG_BTS(bts, DLINP, LOGL_INFO, "Nokia BTS reset timer: %d\n", bts->nokia.bts_reset_timer_cnf);
 	return abis_nm_send(bts, NOKIA_MSG_RESET_REQ, ref, data, len_data);
 }
 
@@ -1247,7 +1251,7 @@ static int abis_nm_send_multi_segments(struct gsm_bts *bts, uint8_t msg_type,
 			memcpy(oh->data, data, len_to_send);
 		}
 
-		DEBUGPC(DNM, "Sending multi-segment %d\n", seq);
+		LOG_BTS(bts, DNM, LOGL_DEBUG, "Sending multi-segment %d\n", seq);
 
 		ret = abis_nm_sendmsg(bts, msg);
 		if (ret < 0)
@@ -1301,7 +1305,7 @@ static int abis_nm_send_config(struct gsm_bts *bts, uint8_t bts_type)
 		idx++;
 	}
 
-	ret = make_bts_config(bts_type, idx, config + len, need_hopping);
+	ret = make_bts_config(bts, bts_type, idx, config + len, need_hopping);
 	len += ret;
 
 #if 0				/* debugging */
@@ -1430,6 +1434,39 @@ static int dump_elements(uint8_t * data, int len)
 	return 0;
 }
 
+static void mo_ok(struct gsm_abis_mo *mo)
+{
+	mo->nm_state.operational    = NM_OPSTATE_ENABLED;
+	mo->nm_state.administrative = NM_STATE_UNLOCKED;
+	mo->nm_state.availability   = NM_AVSTATE_OK;
+}
+
+static void nokia_abis_nm_fake_1221_ok(struct gsm_bts *bts)
+{
+	/*
+	 * The Nokia BTS don't follow the 12.21 model and we don't have OM objects
+	 * for the various elements. However some of the BSC code depends on seeing
+	 * those object "up & running", so when the Nokia init is done, we fake
+	 * a "good" state
+	 */
+	struct gsm_bts_trx *trx;
+
+	mo_ok(&bts->mo);
+	mo_ok(&bts->site_mgr->mo);
+
+	llist_for_each_entry(trx, &bts->trx_list, list) {
+		int i;
+
+		mo_ok(&trx->mo);
+		mo_ok(&trx->bb_transc.mo);
+
+		for (i = 0; i < ARRAY_SIZE(trx->ts); i++) {
+			struct gsm_bts_trx_ts *ts = &trx->ts[i];
+			mo_ok(&ts->mo);
+		}
+	}
+}
+
 /* TODO: put in a separate file ? */
 
 /* taken from abis_nm.c */
@@ -1461,18 +1498,28 @@ static void reset_timer_cb(void *_bts)
 	struct gsm_e1_subslot *e1_link = &bts->oml_e1_link;
 	struct e1inp_line *line;
 
-	bts->nokia.wait_reset = 0;
-
 	/* OML link */
 	line = e1inp_line_find(e1_link->e1_nr);
 	if (!line) {
-		LOGP(DLINP, LOGL_ERROR, "BTS %u OML link referring to "
-		     "non-existing E1 line %u\n", bts->nr, e1_link->e1_nr);
+		LOG_BTS(bts, DLINP, LOGL_ERROR, "BTS OML link referring to non-existing E1 line %u\n",
+			e1_link->e1_nr);
 		return;
 	}
 
-	start_sabm_in_line(line, 0, -1);	/* stop all first */
-	start_sabm_in_line(line, 1, SAPI_OML);	/* start only OML */
+	switch (bts->nokia.wait_reset) {
+	case RESET_T_NONE:				/* shouldn't happen */
+		break;
+	case RESET_T_STOP_LAPD:
+		start_sabm_in_line(line, 0, -1);	/* stop all first */
+		bts->nokia.wait_reset = RESET_T_RESTART_LAPD;
+		osmo_timer_schedule(&bts->nokia.reset_timer, bts->nokia.bts_reset_timer_cnf, 0);
+		break;
+	case RESET_T_RESTART_LAPD:
+		bts->nokia.wait_reset = 0;
+		start_sabm_in_line(line, 0, -1);	/* stop all first */
+		start_sabm_in_line(line, 1, SAPI_OML);	/* start only OML */
+		break;
+	}
 }
 
 /* TODO: put in a separate file ? */
@@ -1487,7 +1534,7 @@ static void reset_timer_cb(void *_bts)
   - receive ACK, start RSL link(s)
   ACK some other messages received from the BTS.
 
-  Probably its also possible to configure the BTS without a reset, this 
+  Probably its also possible to configure the BTS without a reset, this
   has not been tested yet.
 */
 
@@ -1507,18 +1554,17 @@ static int abis_nm_rcvmsg_fom(struct msgb *mb)
 	int len_data;
 
 	if (bts->nokia.wait_reset) {
-		LOGP(DNM, LOGL_INFO,
-		     "Ignore message while waiting for reset\n");
+		LOG_BTS(bts, DNM, LOGL_INFO, "Ignoring message while waiting for reset: %s\n", msgb_hexdump(mb));
 		return ret;
 	}
 
 	if (oh->length < sizeof(struct abis_om_nokia_hdr)) {
-		LOGP(DNM, LOGL_ERROR, "Message too short\n");
+		LOG_BTS(bts, DNM, LOGL_ERROR, "Message too short: %s\n", msgb_hexdump(mb));
 		return -EINVAL;
 	}
 
 	len_data = oh->length - sizeof(struct abis_om_nokia_hdr);
-	LOGP(DNM, LOGL_INFO, "(0x%02X) %s\n", mt, get_msg_type_name_string(mt));
+	LOG_BTS(bts, DNM, LOGL_INFO, "Rx (0x%02X) %s\n", mt, get_msg_type_name_string(mt));
 #if 0				/* debugging */
 	dump_elements(noh->data, len_data);
 #endif
@@ -1528,11 +1574,10 @@ static int abis_nm_rcvmsg_fom(struct msgb *mb)
 		if (find_element(noh->data, len_data,
 				 NOKIA_EI_BTS_TYPE, &bts->nokia.bts_type,
 				 sizeof(uint8_t)) == sizeof(uint8_t))
-			LOGP(DNM, LOGL_INFO, "BTS type = %d (%s)\n",
-			     bts->nokia.bts_type,
-			     get_bts_type_string(bts->nokia.bts_type));
+			LOG_BTS(bts, DNM, LOGL_INFO, "Rx BTS type = %d (%s)\n", bts->nokia.bts_type,
+				get_bts_type_string(bts->nokia.bts_type));
 		else
-			LOGP(DNM, LOGL_ERROR, "BTS type not found\n");
+			LOG_BTS(bts, DNM, LOGL_ERROR, "BTS type not found in NOKIA_MSG_OMU_STARTED\n");
 		/* send START_DOWNLOAD_REQ */
 		abis_nm_download_req(bts, ref);
 		break;
@@ -1550,14 +1595,13 @@ static int abis_nm_rcvmsg_fom(struct msgb *mb)
 		if (find_element
 		    (noh->data, len_data, NOKIA_EI_ACK, &ack,
 		     sizeof(uint8_t)) == sizeof(uint8_t)) {
-			LOGP(DNM, LOGL_INFO, "ACK = %d\n", ack);
+			LOG_BTS(bts, DNM, LOGL_INFO, "Rx ACK = %u\n", ack);
 			if (ack != 1) {
-				LOGP(DNM, LOGL_ERROR, "No ACK received (%d)\n",
-				     ack);
+				LOG_BTS(bts, DNM, LOGL_ERROR, "Rx No ACK (%u): don't know how to proceed\n", ack);
 				/* TODO: properly handle failures (NACK) */
 			}
 		} else
-			LOGP(DNM, LOGL_ERROR, "ACK not found\n");
+			LOG_BTS(bts, DNM, LOGL_ERROR, "Rx MSG_ACK but no EI_ACK found: %s\n", msgb_hexdump(mb));
 
 		/* TODO: the assumption for the following is that no NACK was received */
 
@@ -1565,8 +1609,8 @@ static int abis_nm_rcvmsg_fom(struct msgb *mb)
 		if (!bts->nokia.did_reset) {
 			bts->nokia.did_reset = 1;
 
-			/* 
-			   TODO: For the InSite processing the received data is 
+			/*
+			   TODO: For the InSite processing the received data is
 			   blocked in the driver during reset.
 			   Otherwise the LAPD module might assert because the InSite
 			   sends garbage on the E1 line during reset.
@@ -1574,25 +1618,16 @@ static int abis_nm_rcvmsg_fom(struct msgb *mb)
 			   (function handle_ts1_read()) and ignoring the received data.
 			   It seems to be necessary for the MetroSite too.
 			 */
-			bts->nokia.wait_reset = 1;
 
-			osmo_timer_setup(&bts->nokia.reset_timer,
-					 reset_timer_cb, bts);
-			osmo_timer_schedule(&bts->nokia.reset_timer, bts->nokia.bts_reset_timer_cnf, 0);
-
-			struct gsm_e1_subslot *e1_link = &bts->oml_e1_link;
-			struct e1inp_line *line;
-			/* OML link */
-			line = e1inp_line_find(e1_link->e1_nr);
-			if (!line) {
-				LOGP(DLINP, LOGL_ERROR,
-				     "BTS %u OML link referring to "
-				     "non-existing E1 line %u\n", bts->nr,
-				     e1_link->e1_nr);
-				return -ENOMEM;
-			}
-
-			start_sabm_in_line(line, 0, -1);	/* stop all first */
+			/* we cannot delete / stop the OML LAPD SAP right here, as we are in
+			 * the middle of processing an LAPD I frame and are subsequently returning
+			 * back to the LAPD I frame processing code that assumes the SAP is still
+			 * active. So we first schedule the timer at 0ms in the future, where we
+			 * kill all LAPD SAP and re-arm the timer for the reset duration, after which
+			 * we re-create them */
+			bts->nokia.wait_reset = RESET_T_STOP_LAPD;
+			osmo_timer_setup(&bts->nokia.reset_timer, reset_timer_cb, bts);
+			osmo_timer_schedule(&bts->nokia.reset_timer, 0, 0);
 		}
 
 		/* ACK for CONF DATA message ? */
@@ -1608,11 +1643,8 @@ static int abis_nm_rcvmsg_fom(struct msgb *mb)
 			/* RSL Link */
 			line = e1inp_line_find(e1_link->e1_nr);
 			if (!line) {
-				LOGP(DLINP, LOGL_ERROR,
-				     "TRX (%u/%u) RSL link referring "
-				     "to non-existing E1 line %u\n",
-				     sign_link->trx->bts->nr, sign_link->trx->nr,
-				     e1_link->e1_nr);
+				LOG_BTS(bts, DLINP, LOGL_ERROR, "RSL link referring to "
+					"non-existing E1 line %u\n", e1_link->e1_nr);
 				return -ENOMEM;
 			}
 			/* start TRX */
@@ -1626,6 +1658,8 @@ static int abis_nm_rcvmsg_fom(struct msgb *mb)
 	case NOKIA_MSG_CONF_COMPLETE:
 		/* send ACK */
 		abis_nm_ack(bts, ref);
+		/* fake 12.21 OM */
+		nokia_abis_nm_fake_1221_ok(bts);
 		break;
 	case NOKIA_MSG_BLOCK_CTRL_REQ:	/* seems to be send when something goes wrong !? */
 		/* send ACK (do we have to send an ACK ?) */
@@ -1640,21 +1674,17 @@ static int abis_nm_rcvmsg_fom(struct msgb *mb)
 				 sizeof(info));
 		if (str_len > 0) {
 			info[str_len] = 0;
-			LOGP(DNM, LOGL_INFO, "ALARM Severity %s (%d) : %s\n",
-			     get_severity_string(severity), severity, info);
+			LOG_BTS(bts, DNM, LOGL_NOTICE, "Rx ALARM Severity %s (%d) : %s\n",
+				get_severity_string(severity), severity, info);
 		} else {	/* nothing found, try details */
 			str_len =
-			    find_element(noh->data, len_data,
-					 NOKIA_EI_ALARM_DETAIL, info,
-					 sizeof(info));
+			    find_element(noh->data, len_data, NOKIA_EI_ALARM_DETAIL, info, sizeof(info));
 			if (str_len > 0) {
 				uint16_t code;
 				info[str_len] = 0;
 				code = (info[0] << 8) + info[1];
-				LOGP(DNM, LOGL_INFO,
-				     "ALARM Severity %s (%d), code 0x%X : %s\n",
-				     get_severity_string(severity), severity,
-				     code, info + 2);
+				LOG_BTS(bts, DNM, LOGL_NOTICE, "Rx ALARM Severity %s (%d), code 0x%X : %s\n",
+					get_severity_string(severity), severity, code, info + 2);
 			}
 		}
 		/* send ACK */
@@ -1671,41 +1701,37 @@ static int abis_nm_rcvmsg_fom(struct msgb *mb)
 
 int abis_nokia_rcvmsg(struct msgb *msg)
 {
+	struct e1inp_sign_link *sign_link = (struct e1inp_sign_link *)msg->dst;
+	struct gsm_bts *bts = sign_link->trx->bts;
 	struct abis_om_hdr *oh = msgb_l2(msg);
 	int rc = 0;
 
 	/* Various consistency checks */
 	if (oh->placement != ABIS_OM_PLACEMENT_ONLY) {
-		LOGP(DNM, LOGL_ERROR, "ABIS OML placement 0x%x not supported\n",
-		     oh->placement);
+		LOG_BTS(bts, DNM, LOGL_ERROR, "Rx ABIS OML placement 0x%x not supported\n", oh->placement);
 		if (oh->placement != ABIS_OM_PLACEMENT_FIRST)
 			return -EINVAL;
 	}
 	if (oh->sequence != 0) {
-		LOGP(DNM, LOGL_ERROR, "ABIS OML sequence 0x%x != 0x00\n",
-		     oh->sequence);
+		LOG_BTS(bts, DNM, LOGL_ERROR, "Rx ABIS OML sequence 0x%x != 0x00\n", oh->sequence);
 		return -EINVAL;
 	}
 	msg->l3h = (unsigned char *)oh + sizeof(*oh);
 
 	switch (oh->mdisc) {
 	case ABIS_OM_MDISC_FOM:
-		LOGP(DNM, LOGL_INFO, "ABIS_OM_MDISC_FOM\n");
+		LOG_BTS(bts, DNM, LOGL_INFO, "Rx ABIS_OM_MDISC_FOM\n");
 		rc = abis_nm_rcvmsg_fom(msg);
 		break;
 	case ABIS_OM_MDISC_MANUF:
-		LOGP(DNM, LOGL_INFO, "ABIS_OM_MDISC_MANUF\n");
+		LOG_BTS(bts, DNM, LOGL_INFO, "Rx ABIS_OM_MDISC_MANUF: ignoring\n");
 		break;
 	case ABIS_OM_MDISC_MMI:
 	case ABIS_OM_MDISC_TRAU:
-		LOGP(DNM, LOGL_ERROR,
-		     "unimplemented ABIS OML message discriminator 0x%x\n",
-		     oh->mdisc);
+		LOG_BTS(bts, DNM, LOGL_ERROR, "Rx unimplemented ABIS OML message discriminator 0x%x\n", oh->mdisc);
 		break;
 	default:
-		LOGP(DNM, LOGL_ERROR,
-		     "unknown ABIS OML message discriminator 0x%x\n",
-		     oh->mdisc);
+		LOG_BTS(bts, DNM, LOGL_ERROR, "Rx unknown ABIS OML message discriminator 0x%x\n", oh->mdisc);
 		return -EINVAL;
 	}
 
@@ -1732,14 +1758,6 @@ static struct gsm_network *my_net;
 
 static int bts_model_nokia_site_start(struct gsm_network *net)
 {
-	model_nokia_site.features.data = &model_nokia_site._features_data[0];
-	model_nokia_site.features.data_len =
-	    sizeof(model_nokia_site._features_data);
-
-	osmo_bts_set_feature(&model_nokia_site.features, BTS_FEAT_HOPPING);
-	osmo_bts_set_feature(&model_nokia_site.features, BTS_FEAT_HSCSD);
-        osmo_bts_set_feature(&model_nokia_site.features, BTS_FEAT_MULTI_TSC);
-
 	osmo_signal_register_handler(SS_L_INPUT, inp_sig_cb, NULL);
 	osmo_signal_register_handler(SS_L_GLOBAL, gbl_sig_cb, NULL);
 	osmo_signal_register_handler(SS_NM, nm_sig_cb, NULL);
@@ -1751,5 +1769,12 @@ static int bts_model_nokia_site_start(struct gsm_network *net)
 
 int bts_model_nokia_site_init(void)
 {
+	model_nokia_site.features.data = &model_nokia_site._features_data[0];
+	model_nokia_site.features.data_len = sizeof(model_nokia_site._features_data);
+
+	osmo_bts_set_feature(&model_nokia_site.features, BTS_FEAT_HOPPING);
+	osmo_bts_set_feature(&model_nokia_site.features, BTS_FEAT_HSCSD);
+	osmo_bts_set_feature(&model_nokia_site.features, BTS_FEAT_MULTI_TSC);
+
 	return gsm_bts_model_register(&model_nokia_site);
 }
