@@ -30,6 +30,7 @@
 #include <osmocom/abis/e1_input.h>
 #include <osmocom/bsc/signal.h>
 #include <osmocom/bsc/timeslot_fsm.h>
+#include <osmocom/bsc/bts.h>
 
 static int bts_model_bs11_start(struct gsm_network *net);
 
@@ -61,7 +62,7 @@ static struct gsm_bts_model model_bs11 = {
 			[NM_ATT_BS11_L1_PROT_TYPE] =	{ TLV_TYPE_TV },
 			[NM_ATT_BS11_BIT_ERR_THESH] =	{ TLV_TYPE_FIXED, 2 },
 			[NM_ATT_BS11_DIVERSITY] =	{ TLV_TYPE_TLV },
-			[NM_ATT_BS11_LMT_LOGON_SESSION]={ TLV_TYPE_TLV },	
+			[NM_ATT_BS11_LMT_LOGON_SESSION]={ TLV_TYPE_TLV },
 			[NM_ATT_BS11_LMT_LOGIN_TIME] =	{ TLV_TYPE_TLV },
 			[NM_ATT_BS11_LMT_USER_ACC_LEV] ={ TLV_TYPE_TLV },
 			[NM_ATT_BS11_LMT_USER_NAME] =	{ TLV_TYPE_TLV },
@@ -403,9 +404,16 @@ static void nm_reconfig_ts(struct gsm_bts_trx_ts *ts)
 	if (is_ipaccess_bts(ts->trx->bts))
 		return;
 
-	if (ts_is_tch(ts))
-		abis_nm_conn_terr_traf(ts, e1l->e1_nr, e1l->e1_ts,
-					e1l->e1_ts_ss);
+	switch (ts->pchan_from_config) {
+	case GSM_PCHAN_TCH_F:
+		abis_nm_conn_terr_traf(ts, e1l->e1_nr, e1l->e1_ts, e1l->e1_ts_ss);
+		break;
+	case GSM_PCHAN_TCH_H:
+		LOG_TS(ts, LOGL_ERROR, "No support for half-rate over E1 yet!\n");
+		break;
+	default:
+		break;
+	}
 }
 
 static void nm_reconfig_trx(struct gsm_bts_trx *trx)
@@ -426,7 +434,7 @@ static void nm_reconfig_trx(struct gsm_bts_trx *trx)
 		abis_nm_conn_terr_sign(trx, e1l->e1_nr, e1l->e1_ts,
 					e1l->e1_ts_ss);
 		abis_nm_establish_tei(trx->bts, trx->nr, e1l->e1_nr,
-				      e1l->e1_ts, e1l->e1_ts_ss, trx->rsl_tei);
+				      e1l->e1_ts, e1l->e1_ts_ss, trx->rsl_tei_primary);
 
 		/* Set Radio Attributes */
 		if (trx == trx->bts->c0)
@@ -523,6 +531,11 @@ static void bootstrap_om_bs11(struct gsm_bts *bts)
 
 	/* restart sending event reports */
 	abis_nm_event_reports(bts, 1);
+
+	/* make the timeslot FSM happy. Siemens doesn't send us
+	 * OML state changes for individual timeslots, so we
+	 * bring all of them up here */
+	gsm_bts_all_ts_dispatch(bts, TS_EV_OML_READY, NULL);
 }
 
 static int shutdown_om(struct gsm_bts *bts)
@@ -539,6 +552,7 @@ static int shutdown_om(struct gsm_bts *bts)
 	/* Reset BTS Site manager resource */
 	abis_nm_bs11_reset_resource(bts);
 
+	gsm_bts_stats_reset(bts);
 	gsm_bts_all_ts_dispatch(bts, TS_EV_OML_DOWN, NULL);
 
 	return 0;
@@ -590,13 +604,6 @@ static int inp_sig_cb(unsigned int subsys, unsigned int signal,
 
 static int bts_model_bs11_start(struct gsm_network *net)
 {
-	model_bs11.features.data = &model_bs11._features_data[0];
-	model_bs11.features.data_len = sizeof(model_bs11._features_data);
-
-	osmo_bts_set_feature(&model_bs11.features, BTS_FEAT_HOPPING);
-	osmo_bts_set_feature(&model_bs11.features, BTS_FEAT_HSCSD);
-	osmo_bts_set_feature(&model_bs11.features, BTS_FEAT_MULTI_TSC);
-
 	osmo_signal_register_handler(SS_L_INPUT, inp_sig_cb, NULL);
 	osmo_signal_register_handler(SS_L_GLOBAL, gbl_sig_cb, NULL);
 
@@ -605,5 +612,12 @@ static int bts_model_bs11_start(struct gsm_network *net)
 
 int bts_model_bs11_init(void)
 {
+	model_bs11.features.data = &model_bs11._features_data[0];
+	model_bs11.features.data_len = sizeof(model_bs11._features_data);
+
+	osmo_bts_set_feature(&model_bs11.features, BTS_FEAT_HOPPING);
+	osmo_bts_set_feature(&model_bs11.features, BTS_FEAT_HSCSD);
+	osmo_bts_set_feature(&model_bs11.features, BTS_FEAT_MULTI_TSC);
+
 	return gsm_bts_model_register(&model_bs11);
 }
