@@ -62,16 +62,13 @@ void set_ts_e1link(struct gsm_bts_trx_ts *ts, uint8_t e1_nr,
 struct gsm_bts *gsm_bts_by_lac(struct gsm_network *net, unsigned int lac,
 				struct gsm_bts *start_bts)
 {
-	int i;
 	struct gsm_bts *bts;
 	int skip = 0;
 
 	if (start_bts)
 		skip = 1;
 
-	for (i = 0; i < net->num_bts; i++) {
-		bts = gsm_bts_num(net, i);
-
+	llist_for_each_entry(bts, &net->bts_list, list) {
 		if (skip) {
 			if (start_bts == bts)
 				skip = 0;
@@ -124,7 +121,7 @@ struct gsm_bts *gsm_bts_alloc_register(struct gsm_network *net, enum gsm_bts_typ
 	net->num_bts++;
 
 	bts->type = type;
-	bts->model = model;
+	gsm_set_bts_model(bts, model);
 	bts->bsic = bsic;
 
 	llist_add_tail(&bts->list, &net->bts_list);
@@ -345,6 +342,21 @@ void lchan_update_name(struct gsm_lchan *lchan)
 				      lchan->nr - (lchan->vamos.is_secondary ? ts->max_primary_lchans : 0));
 }
 
+/* If the lchan is currently active, return the duration since activation in milliseconds.
+ * Otherwise return 0. */
+uint64_t gsm_lchan_active_duration_ms(const struct gsm_lchan *lchan)
+{
+	struct timespec now, elapsed;
+
+	if (lchan->active_start.tv_sec == 0 && lchan->active_start.tv_nsec == 0)
+		return 0;
+
+	osmo_clock_gettime(CLOCK_MONOTONIC, &now);
+	timespecsub(&now, &lchan->active_start, &elapsed);
+
+	return elapsed.tv_sec * 1000 + elapsed.tv_nsec / 1000000;
+}
+
 /* obtain the MO structure for a given object instance */
 static inline struct gsm_abis_mo *
 gsm_objclass2mo(struct gsm_bts *bts, uint8_t obj_class,
@@ -588,7 +600,7 @@ int gsm_lchan_and_pchan2chan_nr(const struct gsm_lchan *lchan, enum gsm_phys_cha
 	if (rc < 0)
 		LOG_LCHAN(lchan, LOGL_ERROR,
 			  "Error encoding Channel Number: pchan %s ts %u ss %u%s\n",
-			  gsm_pchan_name(lchan->ts->pchan_from_config), lchan->ts->nr, lchan->nr,
+			  gsm_pchan_name(lchan->ts->pchan_from_config), lchan->ts->nr, lchan_nr,
 			  lchan->vamos.is_secondary ? " (VAMOS shadow)" : "");
 	return rc;
 }
@@ -754,7 +766,7 @@ uint8_t gsm_ts_tsc(const struct gsm_bts_trx_ts *ts)
 bool nm_is_running(const struct gsm_nm_state *s) {
 	if (s->operational != NM_OPSTATE_ENABLED)
 		return false;
-	if ((s->availability != NM_AVSTATE_OK) && (s->availability != 0xff))
+	if (s->availability != NM_AVSTATE_OK)
 		return false;
 	if (s->administrative != NM_STATE_UNLOCKED)
 		return false;
@@ -948,10 +960,8 @@ bool ts_is_capable_of_lchant(struct gsm_bts_trx_ts *ts, enum gsm_chan_t type)
 
 bool ts_is_usable(const struct gsm_bts_trx_ts *ts)
 {
-	if (!trx_is_usable(ts->trx)) {
-		LOGP(DRLL, LOGL_DEBUG, "%s not usable\n", gsm_trx_name(ts->trx));
+	if (!trx_is_usable(ts->trx))
 		return false;
-	}
 
 	if (!ts->fi)
 		return false;
@@ -1129,220 +1139,6 @@ const struct gsm_interf_meas_params interf_meas_params_def = {
 	},
 };
 
-/* Default MS/BS Power Control parameters (see 3GPP TS 45.008, table A.1) */
-const struct gsm_power_ctrl_params power_ctrl_params_def = {
-	/* Static Power Control is the safe default */
-	.mode = GSM_PWR_CTRL_MODE_STATIC,
-
-	/* BS Power reduction value / maximum (in dB) */
-	.bs_power_val_db = 0,  /* no attenuation in static mode */
-	.bs_power_max_db = 12, /* up to 12 dB in dynamic mode */
-
-	/* Power increasing/reducing step size */
-	.inc_step_size_db = 4, /* 2, 4, or 6 dB */
-	.red_step_size_db = 2, /* 2 or 4 dB */
-
-	.ctrl_interval = 1, /* Trigger loop every second SACCH block. TS 45.008 sec 4.7.1 */
-
-	/* RxLev measurement parameters */
-	.rxlev_meas = {
-		.enabled = true,
-		/* Thresholds for RxLev (see 3GPP TS 45.008, A.3.2.1) */
-		.lower_thresh = 32, /* L_RXLEV_XX_P (-78 dBm) */
-		.upper_thresh = 38, /* U_RXLEV_XX_P (-72 dBm) */
-
-		/* Increase {UL,DL}_TXPWR if at least LOWER_CMP_P averages
-		 * out of LOWER_CMP_N averages are lower than L_RXLEV_XX_P */
-		.lower_cmp_p = 10, /* P1 as in 3GPP TS 45.008, A.3.2.1 (case a) */
-		.lower_cmp_n = 12, /* N1 as in 3GPP TS 45.008, A.3.2.1 (case a) */
-		/* Decrease {UL,DL}_TXPWR if at least UPPER_CMP_P averages
-		 * out of UPPER_CMP_N averages are greater than L_RXLEV_XX_P */
-		.upper_cmp_p = 19, /* P2 as in 3GPP TS 45.008, A.3.2.1 (case b) */
-		.upper_cmp_n = 20, /* N2 as in 3GPP TS 45.008, A.3.2.1 (case b) */
-
-		/* No averaging (filtering) by default */
-		.algo = GSM_PWR_CTRL_MEAS_AVG_ALGO_NONE,
-
-		/* Hreqave: the period over which an average is produced */
-		.h_reqave = 4, /* TODO: investigate a reasonable default value */
-		/* Hreqt: the number of averaged results maintained */
-		.h_reqt = 6, /* TODO: investigate a reasonable default value */
-	},
-
-	/* RxQual measurement parameters */
-	.rxqual_meas = {
-		.enabled = true,
-		/* Thresholds for RxQual (see 3GPP TS 45.008, A.3.2.1) */
-		.lower_thresh = 3, /* L_RXQUAL_XX_P (0.8% <= BER < 1.6%) */
-		.upper_thresh = 0, /* U_RXQUAL_XX_P (BER < 0.2%) */
-
-		/* Increase {UL,DL}_TXPWR if at least LOWER_CMP_P averages
-		 * out of LOWER_CMP_N averages are lower than L_RXQUAL_XX_P */
-		.lower_cmp_p = 5, /* P3 as in 3GPP TS 45.008, A.3.2.1 (case c) */
-		.lower_cmp_n = 7, /* N3 as in 3GPP TS 45.008, A.3.2.1 (case c) */
-		/* Decrease {UL,DL}_TXPWR if at least UPPER_CMP_P averages
-		 * out of UPPER_CMP_N averages are greater than L_RXQUAL_XX_P */
-		.upper_cmp_p = 15, /* P4 as in 3GPP TS 45.008, A.3.2.1 (case d) */
-		.upper_cmp_n = 18, /* N4 as in 3GPP TS 45.008, A.3.2.1 (case d) */
-
-		/* No averaging (filtering) by default */
-		.algo = GSM_PWR_CTRL_MEAS_AVG_ALGO_NONE,
-
-		/* Hreqave: the period over which an average is produced */
-		.h_reqave = 4, /* TODO: investigate a reasonable default value */
-		/* Hreqt: the number of averaged results maintained */
-		.h_reqt = 6, /* TODO: investigate a reasonable default value */
-	},
-
-	/* C/I measurement parameters.
-	 * Target C/I retrieved from "GSM/EDGE: Evolution and Performance" Table 10.3.
-	 * Set lower and upper so that (lower + upper) / 2 is equal or slightly
-	 * above the target.
-	 */
-	.ci_fr_meas = { /* FR: Target C/I = 15 dB, Soft blocking threshold = 10 dB */
-		.enabled = false,
-		.lower_thresh = 13,
-		.upper_thresh = 17,
-
-		/* Increase {UL,DL}_TXPWR if at least LOWER_CMP_P averages
-		 * out of LOWER_CMP_N averages are lower than L_CI_FR_XX_P */
-		.lower_cmp_p = 5, /* P3 as in 3GPP TS 45.008, A.3.2.1 (case c) */
-		.lower_cmp_n = 7, /* N3 as in 3GPP TS 45.008, A.3.2.1 (case c) */
-		/* Decrease {UL,DL}_TXPWR if at least UPPER_CMP_P averages
-		 * out of UPPER_CMP_N averages are greater than L_CI_FR_XX_P */
-		.upper_cmp_p = 15, /* P4 as in 3GPP TS 45.008, A.3.2.1 (case d) */
-		.upper_cmp_n = 18, /* N4 as in 3GPP TS 45.008, A.3.2.1 (case d) */
-
-		/* No averaging (filtering) by default */
-		.algo = GSM_PWR_CTRL_MEAS_AVG_ALGO_NONE,
-
-		/* Hreqave: the period over which an average is produced */
-		.h_reqave = 4, /* TODO: investigate a reasonable default value */
-		/* Hreqt: the number of averaged results maintained */
-		.h_reqt = 6, /* TODO: investigate a reasonable default value */
-	},
-	.ci_hr_meas = { /* HR: Target C/I = 18 dB, Soft blocking threshold = 13 dB */
-		.enabled = false,
-		.lower_thresh = 16,
-		.upper_thresh = 21,
-
-		/* Increase {UL,DL}_TXPWR if at least LOWER_CMP_P averages
-		 * out of LOWER_CMP_N averages are lower than L_CI_HR_XX_P */
-		.lower_cmp_p = 5, /* P3 as in 3GPP TS 45.008, A.3.2.1 (case c) */
-		.lower_cmp_n = 7, /* N3 as in 3GPP TS 45.008, A.3.2.1 (case c) */
-		/* Decrease {UL,DL}_TXPWR if at least UPPER_CMP_P averages
-		 * out of UPPER_CMP_N averages are greater than L_CI_HR_XX_P */
-		.upper_cmp_p = 15, /* P4 as in 3GPP TS 45.008, A.3.2.1 (case d) */
-		.upper_cmp_n = 18, /* N4 as in 3GPP TS 45.008, A.3.2.1 (case d) */
-
-		/* No averaging (filtering) by default */
-		.algo = GSM_PWR_CTRL_MEAS_AVG_ALGO_NONE,
-
-		/* Hreqave: the period over which an average is produced */
-		.h_reqave = 4, /* TODO: investigate a reasonable default value */
-		/* Hreqt: the number of averaged results maintained */
-		.h_reqt = 6, /* TODO: investigate a reasonable default value */
-	},
-	.ci_amr_fr_meas = { /* AMR-FR: Target C/I = 9 dB, Soft blocking threshold = 4 dB */
-		.enabled = false,
-		.lower_thresh = 7,
-		.upper_thresh = 11,
-
-		/* Increase {UL,DL}_TXPWR if at least LOWER_CMP_P averages
-		 * out of LOWER_CMP_N averages are lower than L_CI_AMR_FR_XX_P */
-		.lower_cmp_p = 5, /* P3 as in 3GPP TS 45.008, A.3.2.1 (case c) */
-		.lower_cmp_n = 7, /* N3 as in 3GPP TS 45.008, A.3.2.1 (case c) */
-		/* Decrease {UL,DL}_TXPWR if at least UPPER_CMP_P averages
-		 * out of UPPER_CMP_N averages are greater than L_CI_AMR_FR_XX_P */
-		.upper_cmp_p = 15, /* P4 as in 3GPP TS 45.008, A.3.2.1 (case d) */
-		.upper_cmp_n = 18, /* N4 as in 3GPP TS 45.008, A.3.2.1 (case d) */
-
-		/* No averaging (filtering) by default */
-		.algo = GSM_PWR_CTRL_MEAS_AVG_ALGO_NONE,
-
-		/* Hreqave: the period over which an average is produced */
-		.h_reqave = 4, /* TODO: investigate a reasonable default value */
-		/* Hreqt: the number of averaged results maintained */
-		.h_reqt = 6, /* TODO: investigate a reasonable default value */
-	},
-	.ci_amr_hr_meas = { /* AMR-HR: Target C/I = 15 dB, Soft blocking threshold = 10 dB */
-		.enabled = false,
-		.lower_thresh = 13,
-		.upper_thresh = 17,
-
-		/* Increase {UL,DL}_TXPWR if at least LOWER_CMP_P averages
-		 * out of LOWER_CMP_N averages are lower than L_CI_AMR_HR_XX_P */
-		.lower_cmp_p = 5, /* P3 as in 3GPP TS 45.008, A.3.2.1 (case c) */
-		.lower_cmp_n = 7, /* N3 as in 3GPP TS 45.008, A.3.2.1 (case c) */
-		/* Decrease {UL,DL}_TXPWR if at least UPPER_CMP_P averages
-		 * out of UPPER_CMP_N averages are greater than L_CI_AMR_HR_XX_P */
-		.upper_cmp_p = 15, /* P4 as in 3GPP TS 45.008, A.3.2.1 (case d) */
-		.upper_cmp_n = 18, /* N4 as in 3GPP TS 45.008, A.3.2.1 (case d) */
-
-		/* No averaging (filtering) by default */
-		.algo = GSM_PWR_CTRL_MEAS_AVG_ALGO_NONE,
-
-		/* Hreqave: the period over which an average is produced */
-		.h_reqave = 4, /* TODO: investigate a reasonable default value */
-		/* Hreqt: the number of averaged results maintained */
-		.h_reqt = 6, /* TODO: investigate a reasonable default value */
-	},
-	.ci_sdcch_meas = { /* SDCCH: Target C/I = 14 dB, Soft blocking threshold = 9 dB */
-		.enabled = false,
-		.lower_thresh = 12,
-		.upper_thresh = 16,
-
-		/* Increase {UL,DL}_TXPWR if at least LOWER_CMP_P averages
-		 * out of LOWER_CMP_N averages are lower than L_CI_SDCCH_XX_P */
-		.lower_cmp_p = 5, /* P3 as in 3GPP TS 45.008, A.3.2.1 (case c) */
-		.lower_cmp_n = 7, /* N3 as in 3GPP TS 45.008, A.3.2.1 (case c) */
-		/* Decrease {UL,DL}_TXPWR if at least UPPER_CMP_P averages
-		 * out of UPPER_CMP_N averages are greater than L_CI_SDCCH_XX_P */
-		.upper_cmp_p = 15, /* P4 as in 3GPP TS 45.008, A.3.2.1 (case d) */
-		.upper_cmp_n = 18, /* N4 as in 3GPP TS 45.008, A.3.2.1 (case d) */
-
-		/* No averaging (filtering) by default */
-		.algo = GSM_PWR_CTRL_MEAS_AVG_ALGO_NONE,
-
-		/* Hreqave: the period over which an average is produced */
-		.h_reqave = 4, /* TODO: investigate a reasonable default value */
-		/* Hreqt: the number of averaged results maintained */
-		.h_reqt = 6, /* TODO: investigate a reasonable default value */
-	},
-	.ci_gprs_meas = { /* GPRS: Target C/I = 20 dB, Soft blocking threshold = 15 dB */
-		.enabled = false,
-		.lower_thresh = 18,
-		.upper_thresh = 24,
-
-		/* Increase {UL,DL}_TXPWR if at least LOWER_CMP_P averages
-		 * out of LOWER_CMP_N averages are lower than L_CI_GPRS_XX_P */
-		.lower_cmp_p = 5, /* P3 as in 3GPP TS 45.008, A.3.2.1 (case c) */
-		.lower_cmp_n = 7, /* N3 as in 3GPP TS 45.008, A.3.2.1 (case c) */
-		/* Decrease {UL,DL}_TXPWR if at least UPPER_CMP_P averages
-		 * out of UPPER_CMP_N averages are greater than L_CI_GPRS_XX_P */
-		.upper_cmp_p = 15, /* P4 as in 3GPP TS 45.008, A.3.2.1 (case d) */
-		.upper_cmp_n = 18, /* N4 as in 3GPP TS 45.008, A.3.2.1 (case d) */
-
-		/* No averaging (filtering) by default */
-		.algo = GSM_PWR_CTRL_MEAS_AVG_ALGO_NONE,
-
-		/* Hreqave: the period over which an average is produced */
-		.h_reqave = 4, /* TODO: investigate a reasonable default value */
-		/* Hreqt: the number of averaged results maintained */
-		.h_reqt = 6, /* TODO: investigate a reasonable default value */
-	},
-};
-
-void power_ctrl_params_def_reset(struct gsm_power_ctrl_params *params,
-				 enum gsm_power_ctrl_dir dir)
-{
-	*params = power_ctrl_params_def;
-	params->dir = dir;
-	if (dir == GSM_PWR_CTRL_DIR_UL)
-		/* Trigger loop every fourth SACCH block (1.92s). TS 45.008 sec 4.7.1: */
-		params->ctrl_interval = 2;
-}
-
 enum rsl_cmod_spd chan_mode_to_rsl_cmod_spd(enum gsm48_chan_mode chan_mode)
 {
 	switch (gsm48_chan_mode_to_non_vamos(chan_mode)) {
