@@ -490,6 +490,7 @@ static bool parse_ho_request(struct gsm_subscriber_connection *conn, const struc
 		LOG_HO(conn, LOGL_ERROR, "Failed to parse Encryption Information IE\n");
 		return false;
 	}
+	req->ei_as_bitmask = *e->val;
 
 	if ((e = TLVP_GET(tp, GSM0808_IE_KC_128))) {
 		if (e->len != 16) {
@@ -603,6 +604,23 @@ static bool parse_ho_request(struct gsm_subscriber_connection *conn, const struc
 		parse_old2new_bss_info(conn, e->val, e->len, req);
 	}
 
+	/* Decode "Codec List (MSC Preferred)". First set len = 0 to empty the list. (For inter-BSC incoming handover,
+	 * there can't possibly be a list here already, because the conn has just now been created; just do ensure
+	 * sanity.) */
+	conn->codec_list = (struct gsm0808_speech_codec_list){};
+	if ((e = TLVP_GET(tp, GSM0808_IE_SPEECH_CODEC_LIST))) {
+		if (gsm0808_dec_speech_codec_list(&conn->codec_list, e->val, e->len) < 0) {
+			LOG_HO(conn, LOGL_ERROR, "incoming inter-BSC Handover: HO Request:"
+			       " Unable to decode Codec List (MSC Preferred)\n");
+			return false;
+		}
+	}
+	if (aoip && !conn->codec_list.len) {
+		LOG_HO(conn, LOGL_ERROR, "incoming inter-BSC Handover: HO Request:"
+		       " Invalid or empty Codec List (MSC Preferred)\n");
+		return false;
+	}
+
 	/* A lot of IEs remain ignored... */
 
 	return true;
@@ -630,6 +648,7 @@ void handover_start_inter_bsc_in(struct gsm_subscriber_connection *conn,
 	int match_idx;
 	struct osmo_fsm_inst *fi;
 	struct channel_mode_and_rate ch_mode_rate = {};
+	int chosen_a5_n;
 
 	handover_fsm_alloc(conn);
 
@@ -645,7 +664,7 @@ void handover_start_inter_bsc_in(struct gsm_subscriber_connection *conn,
 	ho_fsm_update_id(fi, "interBSCin");
 
 	if (!parse_ho_request(conn, ho_request_msg, req)) {
-		ho_fail(HO_RESULT_ERROR, "Invalid Handover Request message from MSC\n");
+		ho_fail(HO_RESULT_ERROR, "Invalid Handover Request message from MSC");
 		return;
 	}
 
@@ -717,18 +736,30 @@ void handover_start_inter_bsc_in(struct gsm_subscriber_connection *conn,
 		.msc_assigned_cic = req->msc_assigned_cic,
 	};
 
-	if (req->chosen_encr_alg) {
-		info.encr.alg_id = req->chosen_encr_alg;
-		if (info.encr.alg_id > 1 && !req->ei.key_len) {
-			ho_fail(HO_RESULT_ERROR, "Chosen Encryption Algorithm (Serving) reflects A5/%u"
-				" but there is no key (Encryption Information)", info.encr.alg_id - 1);
+	/* Figure out the encryption algorithm */
+	chosen_a5_n = select_best_cipher(req->ei_as_bitmask, bsc_gsmnet->a5_encryption_mask);
+	if (chosen_a5_n < 0) {
+		ho_fail(HO_RESULT_FAIL_RR_HO_FAIL,
+			"There is no A5 encryption mode that both BSC and MSC permit: MSC 0x%x & BSC 0x%x = 0",
+			req->ei_as_bitmask, bsc_gsmnet->a5_encryption_mask);
+		return;
+	}
+	if (chosen_a5_n > 0 && !req->ei.key_len) {
+		/* There is no key. Is A5/0 permitted? */
+		if ((req->ei_as_bitmask & bsc_gsmnet->a5_encryption_mask & 0x1) == 0x1) {
+			chosen_a5_n = 0;
+		} else {
+			ho_fail(HO_RESULT_ERROR,
+				"Encryption is required, but there is no key (Encryption Information)");
 			return;
 		}
 	}
 
-	if (req->ei.key_len) {
+	/* Put encryption info in the chan activation info */
+	info.encr.alg_a5_n = chosen_a5_n;
+	if (chosen_a5_n > 0) {
 		if (req->ei.key_len > sizeof(info.encr.key)) {
-			ho_fail(HO_RESULT_ERROR, "Encryption Information IE key length is too large: %u\n",
+			ho_fail(HO_RESULT_ERROR, "Encryption Information IE key length is too large: %u",
 				req->ei.key_len);
 			return;
 		}
@@ -902,7 +933,7 @@ static void send_handover_performed(struct gsm_subscriber_connection *conn)
 	ho_perf_params.chosen_channel_present = true;
 
 	/* Chosen Encryption Algorithm 3.2.2.44 */
-	ho_perf_params.chosen_encr_alg = lchan->encr.alg_id;
+	ho_perf_params.chosen_encr_alg = ALG_A5_NR_TO_BSSAP(lchan->encr.alg_a5_n);
 	ho_perf_params.chosen_encr_alg_present = true;
 
 	if (ho->new_lchan->activate.info.requires_voice_stream) {
@@ -965,7 +996,7 @@ void handover_end(struct gsm_subscriber_connection *conn, enum handover_result r
 				result = bsc_tx_bssmap_ho_complete(conn, ho->new_lchan);
 		}
 		/* Not 'else': above checks may still result in HO_RESULT_ERROR. */
-		if (result == HO_RESULT_ERROR) {
+		if (result != HO_RESULT_OK) {
 			/* Return a BSSMAP Handover Failure, as described in 3GPP TS 48.008 3.1.5.2.2
 			 * "Handover Resource Allocation Failure" */
 			bsc_tx_bssmap_ho_failure(conn);
@@ -1185,6 +1216,7 @@ static void ho_fsm_wait_rr_ho_detect_onenter(struct osmo_fsm_inst *fi, uint32_t
 	struct handover *ho = &conn->ho;
 
 	struct msgb *rr_ho_cmd = gsm48_make_ho_cmd(ho->new_lchan,
+						   ho->scope, ho->async,
 						   ho->new_lchan->ms_power,
 						   ho->ho_ref);
 	if (!rr_ho_cmd) {
