@@ -204,6 +204,23 @@ static void lchan_on_fully_established(struct gsm_lchan *lchan)
 		return;
 	lchan->activate.concluded = true;
 
+	/* Set active state timekeeping markers. */
+	osmo_clock_gettime(CLOCK_MONOTONIC, &lchan->active_start);
+	lchan->active_stored = lchan->active_start;
+
+	/* Increment rate counters tracking fully established lchans. */
+	switch (lchan->type) {
+	case GSM_LCHAN_TCH_H:
+	case GSM_LCHAN_TCH_F:
+		rate_ctr_inc(rate_ctr_group_get_ctr(lchan->ts->trx->bts->bts_ctrs, BTS_CTR_CHAN_TCH_FULLY_ESTABLISHED));
+		break;
+	case GSM_LCHAN_SDCCH:
+		rate_ctr_inc(rate_ctr_group_get_ctr(lchan->ts->trx->bts->bts_ctrs, BTS_CTR_CHAN_SDCCH_FULLY_ESTABLISHED));
+		break;
+	default:
+		break;
+	}
+
 	switch (lchan->activate.info.activ_for) {
 	case ACTIVATE_FOR_MS_CHANNEL_REQUEST:
 		/* No signalling to do here, MS is free to use the channel, and should go on to connect
@@ -478,8 +495,11 @@ static void lchan_reset(struct gsm_lchan *lchan)
 {
 	LOG_LCHAN(lchan, LOGL_DEBUG, "Clearing lchan state\n");
 
-	if (lchan->conn)
-		gscon_forget_lchan(lchan->conn, lchan);
+	if (lchan->conn) {
+		struct gsm_subscriber_connection *conn = lchan->conn;
+		lchan_forget_conn(lchan);
+		gscon_forget_lchan(conn, lchan);
+	}
 
 	if (lchan->rqd_ref) {
 		talloc_free(lchan->rqd_ref);
@@ -516,15 +536,20 @@ static void lchan_fsm_unused_onenter(struct osmo_fsm_inst *fi, uint32_t prev_sta
 	struct gsm_lchan *lchan = lchan_fi_lchan(fi);
 	struct gsm_bts *bts = lchan->ts->trx->bts;
 	lchan_reset(lchan);
+	chan_counts_ts_update(lchan->ts);
 	osmo_fsm_inst_dispatch(lchan->ts->fi, TS_EV_LCHAN_UNUSED, lchan);
 
-	bsc_update_time_cc_all_allocated(bts->network);
-
 	/* Poll the channel request queue, so that waiting calls can make use of the lchan that just
 	 * has become unused now. */
 	abis_rsl_chan_rqd_queue_poll(bts);
 }
 
+static void lchan_fsm_cbch_onenter(struct osmo_fsm_inst *fi, uint32_t prev_state)
+{
+	struct gsm_lchan *lchan = lchan_fi_lchan(fi);
+	chan_counts_ts_update(lchan->ts);
+}
+
 static void lchan_fsm_wait_after_error_onenter(struct osmo_fsm_inst *fi, uint32_t prev_state)
 {
 	struct gsm_lchan *lchan = lchan_fi_lchan(fi);
@@ -700,7 +725,7 @@ static void lchan_fsm_wait_ts_ready_onenter(struct osmo_fsm_inst *fi, uint32_t p
 		return;
 	}
 
-	bsc_update_time_cc_all_allocated(bts->network);
+	chan_counts_ts_update(lchan->ts);
 
 	lchan->conn = info->for_conn;
 
@@ -747,7 +772,7 @@ static void lchan_fsm_wait_ts_ready_onenter(struct osmo_fsm_inst *fi, uint32_t p
 			: "none",
 		  gsm_lchant_name(lchan->type),
 		  gsm48_chan_mode_name(lchan->activate.ch_mode_rate.chan_mode),
-		  (lchan->activate.info.encr.alg_id ? : 1)-1,
+		  lchan->activate.info.encr.alg_a5_n,
 		  lchan->activate.info.encr.key_len ? osmo_hexdump_nospc(lchan->activate.info.encr.key,
 									 lchan->activate.info.encr.key_len) : "none");
 
@@ -1195,11 +1220,17 @@ static void handle_rll_rel_ind_or_conf(struct osmo_fsm_inst *fi, uint32_t event,
 	lchan->sapis[sapi] = LCHAN_SAPI_UNUSED;
 	rll_indication(lchan, link_id, BSC_RLLR_IND_REL_IND);
 
-	/* Releasing SAPI 0 means the conn becomes invalid; but not if the link_id contains a TCH flag.
-	 * (TODO: is this the correct interpretation?) */
+	/* Releasing SAPI 0 means the conn becomes invalid; but not if the link_id contains a SACCH flag. */
 	if (lchan->conn && sapi == 0 && !(link_id & 0xc0)) {
 		LOG_LCHAN(lchan, LOGL_DEBUG, "lchan is releasing\n");
 		gscon_lchan_releasing(lchan->conn, lchan);
+
+		/* if SAPI=0 is gone, it makes no sense if other SAPIs are still around,
+		 * this is not a valid configuration and we should forget about them.
+		 * This is particularly relevant in case of Ericsson RBS6000, which doesn't
+		 * seem to send a RLL_REL_IND for SAPI=3 if there was already one for SAPI=0 */
+		for_each_active_sapi(sapi, 1, lchan)
+			lchan->sapis[sapi] = LCHAN_SAPI_UNUSED;
 	}
 
 	/* The caller shall check whether all SAPIs are released and cause a state chg */
@@ -1392,8 +1423,9 @@ static void lchan_fsm_wait_rf_release_ack_onenter(struct osmo_fsm_inst *fi, uint
 	 * lchan_reset(), we make sure it does. But in case of releases from error handling, the
 	 * conn might as well notice now already that its lchan is becoming unusable. */
 	if (lchan->conn) {
-		gscon_forget_lchan(lchan->conn, lchan);
+		struct gsm_subscriber_connection *conn = lchan->conn;
 		lchan_forget_conn(lchan);
+		gscon_forget_lchan(conn, lchan);
 	}
 
 	rc = rsl_tx_rf_chan_release(lchan);
@@ -1463,6 +1495,7 @@ static void lchan_fsm_borken_onenter(struct osmo_fsm_inst *fi, uint32_t prev_sta
 
 	/* The actual action besides all the beancounting above */
 	lchan_reset(lchan);
+	chan_counts_ts_update(lchan->ts);
 }
 
 static void lchan_fsm_borken(struct osmo_fsm_inst *fi, uint32_t event, void *data)
@@ -1536,6 +1569,7 @@ static const struct osmo_fsm_state lchan_fsm_states[] = {
 	},
 	[LCHAN_ST_CBCH] = {
 		.name = "CBCH",
+		.onenter = lchan_fsm_cbch_onenter,
 		.out_state_mask = 0
 			| S(LCHAN_ST_UNUSED)
 			,
