@@ -40,6 +40,18 @@
 #define nm_bb_transc_fsm_state_chg(fi, NEXT_STATE) \
 	osmo_fsm_inst_state_chg(fi, NEXT_STATE, 0, 0)
 
+static inline void nm_bb_transc_fsm_becomes_enabled(struct gsm_bts_bb_trx *bb_transc)
+{
+	struct gsm_bts_trx *trx = gsm_bts_bb_trx_get_trx(bb_transc);
+	nm_obj_fsm_becomes_enabled_disabled(trx->bts, bb_transc, NM_OC_BASEB_TRANSC, true);
+}
+
+static inline void nm_bb_transc_fsm_becomes_disabled(struct gsm_bts_bb_trx *bb_transc)
+{
+	struct gsm_bts_trx *trx = gsm_bts_bb_trx_get_trx(bb_transc);
+	nm_obj_fsm_becomes_enabled_disabled(trx->bts, bb_transc, NM_OC_BASEB_TRANSC, false);
+}
+
 //////////////////////////
 // FSM STATE ACTIONS
 //////////////////////////
@@ -57,14 +69,14 @@ static void st_op_disabled_notinstalled_on_enter(struct osmo_fsm_inst *fi, uint3
 static void st_op_disabled_notinstalled(struct osmo_fsm_inst *fi, uint32_t event, void *data)
 {
 	struct nm_statechg_signal_data *nsd;
-	struct gsm_nm_state *new_state;
+	const struct gsm_nm_state *new_state;
 
 	switch (event) {
 	case NM_EV_SW_ACT_REP:
 		break;
 	case NM_EV_STATE_CHG_REP:
 		nsd = (struct nm_statechg_signal_data *)data;
-		new_state = nsd->new_state;
+		new_state = &nsd->new_state;
 		if (new_state->operational == NM_OPSTATE_ENABLED) {
 			/*should not happen... */
 			nm_bb_transc_fsm_state_chg(fi, NM_BB_TRANSC_ST_OP_ENABLED);
@@ -86,7 +98,7 @@ static void st_op_disabled_notinstalled(struct osmo_fsm_inst *fi, uint32_t event
 	}
 }
 
-static void configure_loop(struct gsm_bts_bb_trx *bb_transc, struct gsm_nm_state *state, bool allow_opstart)
+static void configure_loop(struct gsm_bts_bb_trx *bb_transc, const struct gsm_nm_state *state, bool allow_opstart)
 {
 	struct gsm_bts_trx *trx = gsm_bts_bb_trx_get_trx(bb_transc);
 
@@ -139,7 +151,7 @@ static void st_op_disabled_dependency(struct osmo_fsm_inst *fi, uint32_t event,
 {
 	struct gsm_bts_bb_trx *bb_transc = (struct gsm_bts_bb_trx *)fi->priv;
 	struct nm_statechg_signal_data *nsd;
-	struct gsm_nm_state *new_state;
+	const struct gsm_nm_state *new_state;
 
 	switch (event) {
 	case NM_EV_GET_ATTR_REP:
@@ -149,7 +161,7 @@ static void st_op_disabled_dependency(struct osmo_fsm_inst *fi, uint32_t event,
 		return;
 	case NM_EV_STATE_CHG_REP:
 		nsd = (struct nm_statechg_signal_data *)data;
-		new_state = nsd->new_state;
+		new_state = &nsd->new_state;
 		if (new_state->operational == NM_OPSTATE_ENABLED) {
 			/* should not happen... */
 			nm_bb_transc_fsm_state_chg(fi, NM_BB_TRANSC_ST_OP_ENABLED);
@@ -179,8 +191,6 @@ static void st_op_disabled_offline_on_enter(struct osmo_fsm_inst *fi, uint32_t p
 {
 	struct gsm_bts_bb_trx *bb_transc = (struct gsm_bts_bb_trx *)fi->priv;
 
-	/* Warning: In here we may be acessing an state older than new_state
-	   from prev (syncrhonous) FSM state */
 	configure_loop(bb_transc, &bb_transc->mo.nm_state, true);
 }
 
@@ -189,7 +199,7 @@ static void st_op_disabled_offline(struct osmo_fsm_inst *fi, uint32_t event, voi
 	struct gsm_bts_bb_trx *bb_transc = (struct gsm_bts_bb_trx *)fi->priv;
 	struct gsm_bts_trx *trx = gsm_bts_bb_trx_get_trx(bb_transc);
 	struct nm_statechg_signal_data *nsd;
-	struct gsm_nm_state *new_state;
+	const struct gsm_nm_state *new_state;
 
 	switch (event) {
 	case NM_EV_GET_ATTR_REP:
@@ -199,7 +209,7 @@ static void st_op_disabled_offline(struct osmo_fsm_inst *fi, uint32_t event, voi
 		return;
 	case NM_EV_STATE_CHG_REP:
 		nsd = (struct nm_statechg_signal_data *)data;
-		new_state = nsd->new_state;
+		new_state = &nsd->new_state;
 		if (new_state->operational == NM_OPSTATE_ENABLED) {
 			nm_bb_transc_fsm_state_chg(fi, NM_BB_TRANSC_ST_OP_ENABLED);
 			return;
@@ -243,29 +253,35 @@ static void st_op_enabled_on_enter(struct osmo_fsm_inst *fi, uint32_t prev_state
 	bb_transc->mo.get_attr_rep_received = false;
 	bb_transc->mo.opstart_sent = false;
 	bb_transc->mo.adm_unlock_sent = false;
+
+	nm_bb_transc_fsm_becomes_enabled(bb_transc);
 }
 
 static void st_op_enabled(struct osmo_fsm_inst *fi, uint32_t event, void *data)
 {
+	struct gsm_bts_bb_trx *bb_transc = (struct gsm_bts_bb_trx *)fi->priv;
 	struct nm_statechg_signal_data *nsd;
-	struct gsm_nm_state *new_state;
+	const struct gsm_nm_state *new_state;
 
 	switch (event) {
 	case NM_EV_STATE_CHG_REP:
 		nsd = (struct nm_statechg_signal_data *)data;
-		new_state = nsd->new_state;
+		new_state = &nsd->new_state;
 		if (new_state->operational == NM_OPSTATE_ENABLED)
 			return;
 		switch (new_state->availability) { /* operational = DISABLED */
 		case NM_AVSTATE_NOT_INSTALLED:
 		case NM_AVSTATE_POWER_OFF:
+			nm_bb_transc_fsm_becomes_disabled(bb_transc);
 			nm_bb_transc_fsm_state_chg(fi, NM_BB_TRANSC_ST_OP_DISABLED_NOTINSTALLED);
 			return;
 		case NM_AVSTATE_DEPENDENCY:
+			nm_bb_transc_fsm_becomes_disabled(bb_transc);
 			nm_bb_transc_fsm_state_chg(fi, NM_BB_TRANSC_ST_OP_DISABLED_DEPENDENCY);
 			return;
 		case NM_AVSTATE_OFF_LINE:
 		case NM_AVSTATE_OK:
+			nm_bb_transc_fsm_becomes_disabled(bb_transc);
 			nm_bb_transc_fsm_state_chg(fi, NM_BB_TRANSC_ST_OP_DISABLED_OFFLINE);
 			return;
 		default:
@@ -287,8 +303,11 @@ static void st_op_allstate(struct osmo_fsm_inst *fi, uint32_t event, void *data)
 		bb_transc->mo.opstart_sent = false;
 		break;
 	case NM_EV_OML_DOWN:
-		if (fi->state != NM_BB_TRANSC_ST_OP_DISABLED_NOTINSTALLED)
+		if (fi->state != NM_BB_TRANSC_ST_OP_DISABLED_NOTINSTALLED) {
+			if (fi->state == NM_BB_TRANSC_ST_OP_ENABLED)
+				nm_bb_transc_fsm_becomes_disabled(bb_transc);
 			nm_bb_transc_fsm_state_chg(fi, NM_BB_TRANSC_ST_OP_DISABLED_NOTINSTALLED);
+		}
 		break;
 	default:
 		OSMO_ASSERT(0);
