@@ -63,7 +63,7 @@ static void update_msc_osmux_support(struct bsc_msc_data *msc,
 	int rc;
 	bool old_value = msc->remote_supports_osmux;
 
-	rc = tlv_parse(&tp, gsm0808_att_tlvdef(), msg->l4h + 1, length - 1, 0, 0);
+	rc = osmo_bssap_tlv_parse(&tp, msg->l4h + 1, length - 1);
 	if (rc < 0)
 		LOGP(DMSC, LOGL_NOTICE, "Failed parsing TLV looking for Osmux support\n");
 
@@ -114,8 +114,7 @@ static int bssmap_handle_reset(struct bsc_msc_data *msc,
 /* Page a subscriber based on TMSI and LAC via the specified BTS.
  * The msc parameter is the MSC which issued the corresponding paging request.
  * Log an error if paging failed. */
-static void
-page_subscriber(const struct bsc_paging_params *params, struct gsm_bts *bts, uint32_t lac)
+static void page_subscriber(const struct bsc_paging_params *params, struct gsm_bts *bts, uint32_t lac)
 {
 	int ret;
 
@@ -132,16 +131,14 @@ page_subscriber(const struct bsc_paging_params *params, struct gsm_bts *bts, uin
 			       "Paging request failed, or repeated paging on LAC %u\n", lac);
 }
 
-static void
-page_all_bts(const struct bsc_paging_params *params)
+static void page_all_bts(const struct bsc_paging_params *params)
 {
 	struct gsm_bts *bts;
 	llist_for_each_entry(bts, &bsc_gsmnet->bts_list, list)
 		page_subscriber(params, bts, GSM_LAC_RESERVED_ALL_BTS);
 }
 
-static void
-page_cgi(const struct bsc_paging_params *params)
+static void page_cgi(const struct bsc_paging_params *params)
 {
 	int i;
 	for (i = 0; i < params->cil.id_list_len; i++) {
@@ -170,8 +167,7 @@ page_cgi(const struct bsc_paging_params *params)
 	}
 }
 
-static void
-page_lac_and_ci(const struct bsc_paging_params *params)
+static void page_lac_and_ci(const struct bsc_paging_params *params)
 {
 	int i;
 
@@ -193,8 +189,7 @@ page_lac_and_ci(const struct bsc_paging_params *params)
 	}
 }
 
-static void
-page_ci(const struct bsc_paging_params *params)
+static void page_ci(const struct bsc_paging_params *params)
 {
 	int i;
 
@@ -214,8 +209,7 @@ page_ci(const struct bsc_paging_params *params)
 	}
 }
 
-static void
-page_lai_and_lac(const struct bsc_paging_params *params)
+static void page_lai_and_lac(const struct bsc_paging_params *params)
 {
 	int i;
 
@@ -242,8 +236,7 @@ page_lai_and_lac(const struct bsc_paging_params *params)
 	}
 }
 
-static void
-page_lac(const struct bsc_paging_params *params)
+static void page_lac(const struct bsc_paging_params *params)
 {
 	int i;
 
@@ -277,7 +270,10 @@ static int bssmap_handle_paging(struct bsc_msc_data *msc,
 		.tmsi = GSM_RESERVED_TMSI,
 	};
 
-	tlv_parse(&tp, gsm0808_att_tlvdef(), msg->l4h + 1, payload_length - 1, 0, 0);
+	if (osmo_bssap_tlv_parse(&tp, msg->l4h + 1, payload_length - 1) < 0) {
+		LOGP(DMSC, LOGL_ERROR, "%s(): tlv_parse() failed\n", __func__);
+		return -1;
+	}
 	remain = payload_length - 1;
 
 	if (!TLVP_PRESENT(&tp, GSM0808_IE_IMSI)) {
@@ -400,17 +396,18 @@ int bsc_paging_start(struct bsc_paging_params *params)
 	return 0;
 }
 
-/* select the best cipher permitted by the intersection of both masks */
-static int select_best_cipher(uint8_t msc_mask, uint8_t bsc_mask)
+/* Select the best cipher permitted by the intersection of both masks. Return as the n in A5/n, or -1 if the
+ * intersection is empty. */
+int select_best_cipher(uint8_t msc_mask, uint8_t bsc_mask)
 {
 	/* A5/7 ... A5/3: We assume higher is better,
 	 * but: A5/1 is better than A5/2, which is better than A5/0 */
-	const uint8_t codec_strength[8] = { 7, 6, 5, 4, 3, 1, 2, 0 };
+	const uint8_t codec_by_strength[8] = { 7, 6, 5, 4, 3, 1, 2, 0 };
 	uint8_t intersection = msc_mask & bsc_mask;
 	int i;
 
-	for (i = 0; i < ARRAY_SIZE(codec_strength); i++) {
-		uint8_t codec = codec_strength[i];
+	for (i = 0; i < ARRAY_SIZE(codec_by_strength); i++) {
+		uint8_t codec = codec_by_strength[i];
 		if (intersection & (1 << codec))
 			return codec;
 	}
@@ -423,7 +420,10 @@ static int bssmap_handle_clear_cmd(struct gsm_subscriber_connection *conn,
 	struct tlv_parsed tp;
 	enum gsm0808_cause cause_0808;
 
-	tlv_parse(&tp, gsm0808_att_tlvdef(), msg->l4h + 1, length - 1, 0, 0);
+	if (osmo_bssap_tlv_parse(&tp, msg->l4h + 1, length - 1) < 0) {
+		LOGPFSML(conn->fi, LOGL_ERROR, "%s(): tlv_parse() failed\n", __func__);
+		return -1;
+	}
 
 	cause_0808 = gsm0808_get_cause(&tp);
 	if (cause_0808 < 0) {
@@ -482,7 +482,11 @@ static int bssmap_handle_cipher_mode(struct gsm_subscriber_connection *conn,
 
 	conn->ciphering_handled = 1;
 
-	tlv_parse(&tp, gsm0808_att_tlvdef(), msg->l4h + 1, payload_length - 1, 0, 0);
+	if (osmo_bssap_tlv_parse(&tp, msg->l4h + 1, payload_length - 1) < 0) {
+		LOGPFSML(conn->fi, LOGL_ERROR, "%s(): tlv_parse() failed\n", __func__);
+		return -1;
+	}
+
 	if (!TLVP_PRESENT(&tp, GSM0808_IE_ENCRYPTION_INFORMATION)) {
 		LOGP(DMSC, LOGL_ERROR, "IE Encryption Information missing.\n");
 		reject_cause = GSM0808_CAUSE_INFORMATION_ELEMENT_OR_FIELD_MISSING;
@@ -516,9 +520,6 @@ static int bssmap_handle_cipher_mode(struct gsm_subscriber_connection *conn,
 	 * a5_encryption == 2 --> 0x04 ... */
 	enc_bits_msc = data[0];
 
-	/* The bit-mask of permitted ciphers from the MSC (sent in ASSIGNMENT COMMAND) is intersected
-	 * with the vty-configured mask a the BSC.  Finally, the best (highest) possible cipher is
-	 * chosen. */
 	chosen_cipher = select_best_cipher(enc_bits_msc, bsc_gsmnet->a5_encryption_mask);
 	if (chosen_cipher < 0) {
 		LOGP(DMSC, LOGL_ERROR, "Reject: no overlapping A5 ciphers between BSC (0x%02x) "
@@ -541,7 +542,7 @@ static int bssmap_handle_cipher_mode(struct gsm_subscriber_connection *conn,
 		goto reject;
 	}
 
-	conn->lchan->encr.alg_id = RSL_ENC_ALG_A5(chosen_cipher);
+	conn->lchan->encr.alg_a5_n = chosen_cipher;
 	if (enc_key_len) {
 		conn->lchan->encr.key_len = enc_key_len;
 		memcpy(conn->lchan->encr.key, enc_key, enc_key_len);
@@ -633,16 +634,14 @@ static int bssmap_handle_lcls_connect_ctrl(struct gsm_subscriber_connection *con
 	struct msgb *resp;
 	struct tlv_parsed tp;
 	const uint8_t *config, *control;
-	int rc;
 
 	OSMO_ASSERT(conn);
 
-	rc = tlv_parse(&tp, gsm0808_att_tlvdef(), msg->l4h + 1, length - 1, 0, 0);
-	if (rc < 0) {
-		LOGPFSML(conn->fi, LOGL_ERROR, "Error parsing TLVs of LCLS CONNT CTRL: %s\n",
-			 msgb_hexdump(msg));
-		return rc;
+	if (osmo_bssap_tlv_parse(&tp, msg->l4h + 1, length - 1) < 0) {
+		LOGPFSML(conn->fi, LOGL_ERROR, "%s(): tlv_parse() failed\n", __func__);
+		return -1;
 	}
+
 	config = TLVP_VAL_MINLEN(&tp, GSM0808_IE_LCLS_CONFIG, 1);
 	control = TLVP_VAL_MINLEN(&tp, GSM0808_IE_LCLS_CONN_STATUS_CTRL, 1);
 
@@ -825,7 +824,10 @@ static int bssmap_handle_assignm_req(struct gsm_subscriber_connection *conn,
 
 	aoip = gscon_is_aoip(conn);
 
-	tlv_parse(&tp, gsm0808_att_tlvdef(), msg->l4h + 1, length - 1, 0, 0);
+	if (osmo_bssap_tlv_parse(&tp, msg->l4h + 1, length - 1) < 0) {
+		LOGPFSML(conn->fi, LOGL_ERROR, "%s(): tlv_parse() failed\n", __func__);
+		return -1;
+	}
 
 	/* Check for channel type element, if its missing, immediately reject */
 	if (!TLVP_PRESENT(&tp, GSM0808_IE_CHANNEL_TYPE)) {
@@ -990,6 +992,24 @@ reject:
 	return -1;
 }
 
+/* Handle Handover Request message, part of inter-BSC handover:
+ * The MSC opened a new SCCP connection and is asking this BSS to accept an inter-BSC incoming handover.
+ * If we accept, we'll send a Handover Request Acknowledge.
+ * This function is only called when the Handover Request is *not* included in the initial SCCP N-Connect message, but
+ * follows an "empty" N-Connect in a separate DT1 message.
+ */
+static int bssmap_handle_handover_request(struct gsm_subscriber_connection *conn, struct msgb *msg)
+{
+	if (osmo_fsm_inst_dispatch(conn->fi, GSCON_EV_A_INITIAL_USER_DATA, msg)) {
+		/* A Handover Request message should come in on a newly opened SCCP conn. Apparently the MSC has sent a
+		 * Handover Request on an already busy SCCP conn, and naturally we cannot accept another subscriber
+		 * here. This is unlikely to ever happen in practice. Respond in the only possible way: */
+		bsc_tx_bssmap_ho_failure(conn);
+		return -EINVAL;
+	}
+	return 0;
+}
+
 /* Handle Handover Command message, part of inter-BSC handover:
  * This BSS sent a Handover Required message.
  * The MSC contacts the remote BSS and receives from it an RR Handover Command; this BSSMAP Handover
@@ -1010,7 +1030,10 @@ static int bssmap_handle_handover_cmd(struct gsm_subscriber_connection *conn,
 		return -EINVAL;
 	}
 
-	tlv_parse(&tp, gsm0808_att_tlvdef(), msg->l4h + 1, length - 1, 0, 0);
+	if (osmo_bssap_tlv_parse(&tp, msg->l4h + 1, length - 1) < 0) {
+		LOGPFSML(conn->fi, LOGL_ERROR, "%s(): tlv_parse() failed\n", __func__);
+		return -1;
+	}
 
 	/* Check for channel type element, if its missing, immediately reject */
 	if (!TLVP_PRESENT(&tp, GSM0808_IE_LAYER_3_INFORMATION)) {
@@ -1048,7 +1071,10 @@ static int bssmap_handle_confusion(struct gsm_subscriber_connection *conn,
 	enum gsm0808_cause_class cause_class;
 	struct gsm0808_diagnostics *diag;
 
-	osmo_bssap_tlv_parse(&tp, msg->l4h + 1, length - 1);
+	if (osmo_bssap_tlv_parse(&tp, msg->l4h + 1, length - 1) < 0) {
+		LOGPFSML(conn->fi, LOGL_ERROR, "%s(): tlv_parse() failed\n", __func__);
+		return -1;
+	}
 
 	/* Check for the Cause and Diagnostic mandatory elements */
 	if (!TLVP_PRESENT(&tp, GSM0808_IE_CAUSE) || !TLVP_PRESENT(&tp, GSM0808_IE_DIAGNOSTIC)) {
@@ -1091,7 +1117,10 @@ static int bssmap_handle_common_id(struct gsm_subscriber_connection *conn,
 {
 	struct tlv_parsed tp;
 
-	osmo_bssap_tlv_parse(&tp, msg->l4h + 1, length - 1);
+	if (osmo_bssap_tlv_parse(&tp, msg->l4h + 1, length - 1) < 0) {
+		LOGPFSML(conn->fi, LOGL_ERROR, "%s(): tlv_parse() failed\n", __func__);
+		return -1;
+	}
 
 	/* Check for the mandatory elements */
 	if (!TLVP_PRESENT(&tp, GSM0808_IE_IMSI)) {
@@ -1166,14 +1195,18 @@ static int bssmap_rcvmsg_dt1(struct gsm_subscriber_connection *conn,
 		rate_ctr_inc(&ctrs[MSC_CTR_BSSMAP_RX_DT1_CIPHER_MODE_CMD]);
 		ret = bssmap_handle_cipher_mode(conn, msg, length);
 		break;
-	case BSS_MAP_MSG_ASSIGMENT_RQST:
-		rate_ctr_inc(&ctrs[MSC_CTR_BSSMAP_RX_DT1_ASSIGMENT_RQST]);
+	case BSS_MAP_MSG_ASSIGNMENT_RQST:
+		rate_ctr_inc(&ctrs[MSC_CTR_BSSMAP_RX_DT1_ASSIGNMENT_RQST]);
 		ret = bssmap_handle_assignm_req(conn, msg, length);
 		break;
 	case BSS_MAP_MSG_LCLS_CONNECT_CTRL:
 		rate_ctr_inc(&ctrs[MSC_CTR_BSSMAP_RX_DT1_LCLS_CONNECT_CTRL]);
 		ret = bssmap_handle_lcls_connect_ctrl(conn, msg, length);
 		break;
+	case BSS_MAP_MSG_HANDOVER_RQST:
+		rate_ctr_inc(&ctrs[MSC_CTR_BSSMAP_RX_DT1_HANDOVER_RQST]);
+		ret = bssmap_handle_handover_request(conn, msg);
+		break;
 	case BSS_MAP_MSG_HANDOVER_CMD:
 		rate_ctr_inc(&ctrs[MSC_CTR_BSSMAP_RX_DT1_HANDOVER_CMD]);
 		ret = bssmap_handle_handover_cmd(conn, msg, length);
@@ -1215,16 +1248,6 @@ static int bssmap_rcvmsg_dt1(struct gsm_subscriber_connection *conn,
 	return ret;
 }
 
-/* RSL Link Identifier is defined in 3GPP TS 3GPP TS 48.058, section 9.3.2.
- * .... .SSS - SAPI value used on the radio link;
- * ...P P... - priority for SAPI0 messages;
- * CC.. .... - control channel identification:
- *   00.. .... - main signalling channel (FACCH or SDCCH),
- *   01.. .... - SACCH,
- *   other values are reserved. */
-#define DLCI2RSL_LINK_ID(dlci) \
-	((dlci & 0xc0) == 0xc0 ? 0x40 : 0x00) | (dlci & 0x07)
-
 static int dtap_rcvmsg(struct gsm_subscriber_connection *conn,
 		       struct msgb *msg, unsigned int length)
 {
@@ -1433,14 +1456,16 @@ int bsc_tx_bssmap_ho_request_ack(struct gsm_subscriber_connection *conn, struct
 		.l3_info_len = rr_ho_command->len,
 		.chosen_channel_present = true,
 		.chosen_channel = gsm0808_chosen_channel(new_lchan->type, new_lchan->current_ch_mode_rate.chan_mode),
-		.chosen_encr_alg = new_lchan->encr.alg_id,
+		.chosen_encr_alg = ALG_A5_NR_TO_BSSAP(new_lchan->encr.alg_a5_n),
 		.chosen_speech_version = gsm0808_permitted_speech(new_lchan->type,
 								  new_lchan->current_ch_mode_rate.chan_mode),
 	};
 
-	if (gscon_is_aoip(conn)) {
+	if (gscon_is_aoip(conn) && new_lchan->activate.info.requires_voice_stream) {
 		struct osmo_sockaddr_str to_msc_rtp;
 		const struct mgcp_conn_peer *rtp_info = osmo_mgcpc_ep_ci_get_rtp_info(conn->user_plane.mgw_endpoint_ci_msc);
+		int rc;
+		int perm_spch;
 		if (!rtp_info) {
 			LOG_HO(conn, LOGL_ERROR,
 			       "Handover Request Acknowledge: no RTP address known to send as"
@@ -1456,6 +1481,19 @@ int bsc_tx_bssmap_ho_request_ack(struct gsm_subscriber_connection *conn, struct
 			return -EINVAL;
 		}
 		params.aoip_transport_layer = &ss;
+
+		/* speech_codec_chosen */
+		perm_spch = gsm0808_permitted_speech(new_lchan->type, new_lchan->current_ch_mode_rate.chan_mode);
+		params.speech_codec_chosen_present = true;
+		rc = gsm0808_speech_codec_from_chan_type(&params.speech_codec_chosen, perm_spch);
+		if (rc) {
+			LOG_HO(conn, LOGL_ERROR, "Unable to compose Speech Codec (Chosen)\n");
+			return -EINVAL;
+		}
+
+		/* Codec list (BSS Supported) */
+		params.more_items = true;
+		gen_bss_supported_codec_list(&params.codec_list_bss_supported, conn->sccp.msc, new_lchan->ts->trx->bts);
 	}
 
 	rate_ctr_inc(rate_ctr_group_get_ctr(conn->sccp.msc->msc_ctrs, MSC_CTR_BSSMAP_TX_DT1_HANDOVER_RQST_ACKNOWLEDGE));
@@ -1488,7 +1526,7 @@ enum handover_result bsc_tx_bssmap_ho_complete(struct gsm_subscriber_connection
 
 	struct gsm0808_handover_complete params = {
 		.chosen_encr_alg_present = true,
-		.chosen_encr_alg = lchan->encr.alg_id,
+		.chosen_encr_alg = ALG_A5_NR_TO_BSSAP(lchan->encr.alg_a5_n),
 
 		.chosen_channel_present = true,
 		.chosen_channel = gsm0808_chosen_channel(lchan->type, lchan->current_ch_mode_rate.chan_mode),
