@@ -26,7 +26,9 @@
 #include <osmocom/bsc/osmo_bsc_rf.h>
 #include <osmocom/bsc/bsc_msc_data.h>
 #include <osmocom/bsc/signal.h>
-#include <osmocom/bsc/gsm_04_80.h>
+#include <osmocom/bsc/a_reset.h>
+#include <osmocom/bsc/bts.h>
+#include <osmocom/bsc/handover_ctrl.h>
 
 #include <osmocom/core/linuxlist.h>
 #include <osmocom/core/signal.h>
@@ -180,21 +182,12 @@ CTRL_CMD_DEFINE_RO(msc_connection_status, "connection_status");
 static int get_msc_connection_status(struct ctrl_cmd *cmd, void *data)
 {
 	struct bsc_msc_data *msc = (struct bsc_msc_data *)cmd->node;
-	struct osmo_ss7_as *as;
-	const char *as_state_name;
 
 	if (msc == NULL) {
 		cmd->reply = "msc not found";
 		return CTRL_CMD_ERROR;
 	}
-	as = msc_get_ss7_as(msc);
-	if (!as) {
-		cmd->reply = "AS not found for MSC";
-		return CTRL_CMD_ERROR;
-	}
-
-	as_state_name = osmo_fsm_inst_state_name(as->fi);
-	if (!strcmp(as_state_name, "AS_ACTIVE"))
+	if (a_reset_conn_ready(msc))
 		cmd->reply = "connected";
 	else
 		cmd->reply = "disconnected";
@@ -203,7 +196,6 @@ static int get_msc_connection_status(struct ctrl_cmd *cmd, void *data)
 
 /* Backwards compat. */
 CTRL_CMD_DEFINE_RO(msc0_connection_status, "msc_connection_status");
-static int msc_connection_status = 0; /* XXX unused */
 
 static int get_msc0_connection_status(struct ctrl_cmd *cmd, void *data)
 {
@@ -222,13 +214,12 @@ static int msc_connection_status_trap_cb(unsigned int subsys, unsigned int signa
 {
 	struct ctrl_cmd *cmd;
 	struct gsm_network *gsmnet = (struct gsm_network *)handler_data;
+	struct bsc_msc_data *msc = (struct bsc_msc_data *)signal_data;
 
-	if (signal == S_MSC_LOST && msc_connection_status == 1) {
+	if (signal == S_MSC_LOST) {
 		LOGP(DCTRL, LOGL_DEBUG, "MSC connection lost, sending TRAP.\n");
-		msc_connection_status = 0;
-	} else if (signal == S_MSC_CONNECTED && msc_connection_status == 0) {
+	} else if (signal == S_MSC_CONNECTED) {
 		LOGP(DCTRL, LOGL_DEBUG, "MSC connection (re)established, sending TRAP.\n");
-		msc_connection_status = 1;
 	} else {
 		return 0;
 	}
@@ -240,12 +231,19 @@ static int msc_connection_status_trap_cb(unsigned int subsys, unsigned int signa
 	}
 
 	cmd->id = "0";
-	cmd->variable = "msc_connection_status";
+	cmd->variable = talloc_asprintf(cmd, "msc.%d.connection_status", msc->nr);
+	cmd->node = msc;
 
-	get_msc0_connection_status(cmd, NULL);
+	get_msc_connection_status(cmd, NULL);
 
 	ctrl_cmd_send_to_all(gsmnet->ctrl, cmd);
 
+	if (msc->nr == 0) {
+		/* Backwards compat. */
+		cmd->variable = "msc_connection_status";
+		ctrl_cmd_send_to_all(gsmnet->ctrl, cmd);
+	}
+
 	talloc_free(cmd);
 
 	return 0;
@@ -349,7 +347,7 @@ void bsc_gen_location_state_trap(struct gsm_bts *bts)
 {
 	struct bsc_msc_data *msc;
 
-	llist_for_each_entry(msc, &bts->network->bsc_data->mscs, entry)
+	llist_for_each_entry(msc, &bts->network->mscs, entry)
 		generate_location_state_trap(bts, msc);
 }
 
@@ -430,6 +428,20 @@ static int set_bts_loc(struct ctrl_cmd *cmd, void *data)
 	if (!tmp)
 		goto oom;
 
+	tstamp = strtok_r(tmp, ",", &saveptr);
+	valid = strtok_r(NULL, ",", &saveptr);
+	lat = strtok_r(NULL, ",", &saveptr);
+	lon = strtok_r(NULL, ",", &saveptr);
+	height = strtok_r(NULL, "\0", &saveptr);
+
+	/* Check if one of the strtok results was NULL. This will probably never occur since we will only see verified
+	 * input in this code path */
+	if ((tstamp == NULL) || (valid == NULL) || (lat == NULL) || (lon == NULL) || (height == NULL)) {
+		talloc_free(tmp);
+		cmd->reply = "parse error";
+		return CTRL_CMD_ERROR;
+	}
+
 	curloc = talloc_zero(tall_bsc_ctx, struct bts_location);
 	if (!curloc) {
 		talloc_free(tmp);
@@ -437,13 +449,6 @@ static int set_bts_loc(struct ctrl_cmd *cmd, void *data)
 	}
 	INIT_LLIST_HEAD(&curloc->list);
 
-
-	tstamp = strtok_r(tmp, ",", &saveptr);
-	valid = strtok_r(NULL, ",", &saveptr);
-	lat = strtok_r(NULL, ",", &saveptr);
-	lon = strtok_r(NULL, ",", &saveptr);
-	height = strtok_r(NULL, "\0", &saveptr);
-
 	curloc->tstamp = atol(tstamp);
 	curloc->valid = get_string_value(bts_loc_fix_names, valid);
 	curloc->lat = atof(lat);
@@ -652,7 +657,7 @@ static int set_net_inform_msc(struct ctrl_cmd *cmd, void *data)
 	struct bsc_msc_data *msc;
 
 	net = cmd->node;
-	llist_for_each_entry(msc, &net->bsc_data->mscs, entry) {
+	llist_for_each_entry(msc, &net->mscs, entry) {
 		struct ctrl_cmd *trap;
 
 		trap = ctrl_cmd_create(tall_bsc_ctx, CTRL_TYPE_TRAP);
@@ -672,70 +677,6 @@ static int set_net_inform_msc(struct ctrl_cmd *cmd, void *data)
 	return CTRL_CMD_HANDLED;
 }
 
-CTRL_CMD_DEFINE_WO(net_ussd_notify, "ussd-notify-v1");
-static int set_net_ussd_notify(struct ctrl_cmd *cmd, void *data)
-{
-	struct gsm_subscriber_connection *conn;
-	struct gsm_network *net;
-	char *saveptr = NULL;
-	char *cic_str, *alert_str, *text_str;
-	int cic, alert;
-
-	/* Verify has done the test for us */
-	cic_str = strtok_r(cmd->value, ",", &saveptr);
-	alert_str = strtok_r(NULL, ",", &saveptr);
-	text_str = strtok_r(NULL, ",", &saveptr);
-
-	if (!cic_str || !alert_str || !text_str) {
-		cmd->reply = "Programming issue. How did this pass verify?";
-		return CTRL_CMD_ERROR;
-	}
-
-	cmd->reply = "No connection found";
-
-	cic = atoi(cic_str);
-	alert = atoi(alert_str);
-
-	net = cmd->node;
-	llist_for_each_entry(conn, &net->subscr_conns, entry) {
-		if (conn->user_plane.msc_assigned_cic != cic)
-			continue;
-
-		/*
-		 * This is a hack. My E71 does not like to immediately
-		 * receive a release complete on a TCH. So schedule a
-		 * release complete to clear any previous attempt. The
-		 * right thing would be to track invokeId and only send
-		 * the release complete when we get a returnResultLast
-		 * for this invoke id.
-		 */
-		bsc_send_ussd_release_complete(conn);
-		bsc_send_ussd_notify(conn, alert, text_str);
-		cmd->reply = "Found a connection";
-		break;
-	}
-
-	return CTRL_CMD_REPLY;
-}
-
-static int verify_net_ussd_notify(struct ctrl_cmd *cmd, const char *value, void *data)
-{
-	char *saveptr = NULL;
-	char *inp, *cic, *alert, *text;
-
-	OSMO_ASSERT(cmd);
-	inp = talloc_strdup(cmd, value);
-
-	cic = strtok_r(inp, ",", &saveptr);
-	alert = strtok_r(NULL, ",", &saveptr);
-	text = strtok_r(NULL, ",", &saveptr);
-
-	talloc_free(inp);
-	if (!cic || !alert || !text)
-		return 1;
-	return 0;
-}
-
 static int msc_signal_handler(unsigned int subsys, unsigned int signal,
 			void *handler_data, void *signal_data)
 {
@@ -762,6 +703,9 @@ int bsc_ctrl_cmds_install(struct gsm_network *net)
 	int rc;
 
 	rc = bsc_base_ctrl_cmds_install();
+	if (rc)
+		goto end;
+	rc = bsc_ho_ctrl_cmds_install(net);
 	if (rc)
 		goto end;
 	rc = ctrl_cmd_install(CTRL_NODE_BTS, &cmd_bts_loc);
@@ -789,9 +733,6 @@ int bsc_ctrl_cmds_install(struct gsm_network *net)
 	if (rc)
 		goto end;
 	rc = ctrl_cmd_install(CTRL_NODE_ROOT, &cmd_net_inform_msc);
-	if (rc)
-		goto end;
-	rc = ctrl_cmd_install(CTRL_NODE_ROOT, &cmd_net_ussd_notify);
 	if (rc)
 		goto end;
 	rc = osmo_signal_register_handler(SS_L_INPUT, &bts_connection_status_trap_cb, net);
