@@ -38,6 +38,7 @@
 #include <osmocom/bsc/assignment_fsm.h>
 #include <osmocom/bsc/handover_fsm.h>
 #include <osmocom/bsc/smscb.h>
+#include <osmocom/bsc/lb.h>
 
 #include <osmocom/ctrl/control_cmd.h>
 #include <osmocom/ctrl/control_if.h>
@@ -53,8 +54,10 @@
 #include <osmocom/vty/ports.h>
 #include <osmocom/vty/logging.h>
 #include <osmocom/vty/command.h>
+#include <osmocom/vty/cpu_sched_vty.h>
 
 #include <osmocom/mgcp_client/mgcp_client_endpoint_fsm.h>
+#include <osmocom/mgcp_client/mgcp_client_pool.h>
 
 #include <osmocom/abis/abis.h>
 #include <osmocom/bsc/abis_om2000.h>
@@ -63,8 +66,12 @@
 #include <osmocom/bsc/chan_alloc.h>
 #include <osmocom/bsc/e1_config.h>
 #include <osmocom/bsc/codec_pref.h>
+#include <osmocom/bsc/system_information.h>
+#include <osmocom/bsc/bts.h>
+#include <osmocom/bsc/bsc_stats.h>
 
 #include <osmocom/mgcp_client/mgcp_client.h>
+#include <osmocom/mgcp_client/mgcp_client_pool.h>
 
 #include <osmocom/sigtran/xua_msg.h>
 
@@ -80,16 +87,9 @@
 
 #include "../../bscconfig.h"
 
-struct gsm_network *bsc_gsmnet = 0;
 static const char *config_file = "osmo-bsc.cfg";
 static const char *rf_ctrl = NULL;
 static int daemonize = 0;
-static LLIST_HEAD(access_lists);
-
-struct llist_head *bsc_access_lists(void)
-{
-	return &access_lists;
-}
 
 static void print_usage()
 {
@@ -98,7 +98,7 @@ static void print_usage()
 
 static void print_help()
 {
-	printf("  Some useful help...\n");
+	printf("Some useful options:\n");
 	printf("  -h --help 			This text.\n");
 	printf("  -D --daemonize 		Fork the process into a background daemon.\n");
 	printf("  -d  --debug option 		--debug=DRLL:DMM:DRR:DRSL:DNM enable debugging.\n");
@@ -106,16 +106,44 @@ static void print_help()
 	printf("  -T --timestamp		Print a timestamp in the debug output.\n");
 	printf("  -V --version               	Print the version of OsmoBSC.\n");
 	printf("  -c --config-file filename	The config file to use.\n");
-	printf("  -l --local IP			The local address of the MGCP.\n");
 	printf("  -e --log-level number		Set a global loglevel.\n");
 	printf("  -r --rf-ctl NAME		A unix domain socket to listen for cmds.\n");
-	printf("  -t --testmode			A special mode to provoke failures at the MSC.\n");
+
+	printf("\nVTY reference generation:\n");
+	printf("     --vty-ref-mode MODE	VTY reference generation mode (e.g. 'expert').\n");
+	printf("     --vty-ref-xml		Generate the VTY reference XML output and exit.\n");
+}
+
+static void handle_long_options(const char *prog_name, const int long_option)
+{
+	static int vty_ref_mode = VTY_REF_GEN_MODE_DEFAULT;
+
+	switch (long_option) {
+	case 1:
+		vty_ref_mode = get_string_value(vty_ref_gen_mode_names, optarg);
+		if (vty_ref_mode < 0) {
+			fprintf(stderr, "%s: Unknown VTY reference generation "
+				"mode '%s'\n", prog_name, optarg);
+			exit(2);
+		}
+		break;
+	case 2:
+		fprintf(stderr, "Generating the VTY reference in mode '%s' (%s)\n",
+			get_value_string(vty_ref_gen_mode_names, vty_ref_mode),
+			get_value_string(vty_ref_gen_mode_desc, vty_ref_mode));
+		vty_dump_xml_ref_mode(stdout, (enum vty_ref_gen_mode) vty_ref_mode);
+		exit(0);
+	default:
+		fprintf(stderr, "%s: error parsing cmdline options\n", prog_name);
+		exit(2);
+	}
 }
 
 static void handle_options(int argc, char **argv)
 {
 	while (1) {
 		int option_index = 0, c;
+		static int long_option = 0;
 		static struct option long_options[] = {
 			{"help", 0, 0, 'h'},
 			{"debug", 1, 0, 'd'},
@@ -124,14 +152,14 @@ static void handle_options(int argc, char **argv)
 			{"disable-color", 0, 0, 's'},
 			{"timestamp", 0, 0, 'T'},
 			{"version", 0, 0, 'V' },
-			{"local", 1, 0, 'l'},
 			{"log-level", 1, 0, 'e'},
 			{"rf-ctl", 1, 0, 'r'},
-			{"testmode", 0, 0, 't'},
+			{"vty-ref-mode", 1, &long_option, 1},
+			{"vty-ref-xml", 0, &long_option, 2},
 			{0, 0, 0, 0}
 		};
 
-		c = getopt_long(argc, argv, "hd:DsTVc:e:r:t",
+		c = getopt_long(argc, argv, "hd:DsTVc:e:r:",
 				long_options, &option_index);
 		if (c == -1)
 			break;
@@ -141,6 +169,9 @@ static void handle_options(int argc, char **argv)
 			print_usage();
 			print_help();
 			exit(0);
+		case 0:
+			handle_long_options(argv[0], long_option);
+			break;
 		case 's':
 			log_set_use_color(osmo_stderr_target, 0);
 			break;
@@ -167,8 +198,9 @@ static void handle_options(int argc, char **argv)
 			rf_ctrl = optarg;
 			break;
 		default:
-			/* ignore */
-			break;
+			/* catch unknown options *as well as* missing arguments. */
+			fprintf(stderr, "Error in command line options. Exiting.\n");
+			exit(-1);
 		}
 	}
 
@@ -222,71 +254,70 @@ static int nm_sig_cb(unsigned int subsys, unsigned int signal,
 }
 
 /* Produce a MA as specified in 10.5.2.21 */
-static int generate_ma_for_ts(struct gsm_bts_trx_ts *ts)
+static void generate_ma_for_ts(struct gsm_bts_trx_ts *ts)
 {
 	/* we have three bitvecs: the per-timeslot ARFCNs, the cell chan ARFCNs
 	 * and the MA */
-	struct bitvec *cell_chan = &ts->trx->bts->si_common.cell_alloc;
-	struct bitvec *ts_arfcn = &ts->hopping.arfcns;
+	const size_t num_cell_arfcns = ts->trx->bts->si_common.cell_chan_num;
+	const struct bitvec *cell_chan = &ts->trx->bts->si_common.cell_alloc;
+	const struct bitvec *ts_arfcn = &ts->hopping.arfcns;
 	struct bitvec *ma = &ts->hopping.ma;
-	unsigned int num_cell_arfcns, bitnum, n_chan;
 	int i;
 
 	/* re-set the MA to all-zero */
-	ma->cur_bit = 0;
 	ts->hopping.ma_len = 0;
 	memset(ma->data, 0, ma->data_len);
 
 	if (!ts->hopping.enabled)
-		return 0;
-
-	/* count the number of ARFCNs in the cell channel allocation */
-	num_cell_arfcns = 0;
-	for (i = 0; i < 1024; i++) {
-		if (bitvec_get_bit_pos(cell_chan, i))
-			num_cell_arfcns++;
-	}
+		return;
 
 	/* pad it to octet-aligned number of bits */
-	ts->hopping.ma_len = num_cell_arfcns / 8;
-	if (num_cell_arfcns % 8)
-		ts->hopping.ma_len++;
+	ts->hopping.ma_len = OSMO_BYTES_FOR_BITS(num_cell_arfcns);
+	ma->cur_bit = (ts->hopping.ma_len * 8) - 1;
 
-	n_chan = 0;
-	for (i = 0; i < 1024; i++) {
+	for (i = 1; i < 1024; i++) {
 		if (!bitvec_get_bit_pos(cell_chan, i))
 			continue;
 		/* set the corresponding bit in the MA */
-		bitnum = (ts->hopping.ma_len * 8) - 1 - n_chan;
 		if (bitvec_get_bit_pos(ts_arfcn, i))
-			bitvec_set_bit_pos(ma, bitnum, 1);
+			bitvec_set_bit_pos(ma, ma->cur_bit, 1);
 		else
-			bitvec_set_bit_pos(ma, bitnum, 0);
-		n_chan++;
+			bitvec_set_bit_pos(ma, ma->cur_bit, 0);
+		ma->cur_bit--;
 	}
 
 	/* ARFCN 0 is special: It is coded last in the bitmask */
 	if (bitvec_get_bit_pos(cell_chan, 0)) {
-		n_chan++;
 		/* set the corresponding bit in the MA */
-		bitnum = (ts->hopping.ma_len * 8) - 1 - n_chan;
 		if (bitvec_get_bit_pos(ts_arfcn, 0))
-			bitvec_set_bit_pos(ma, bitnum, 1);
+			bitvec_set_bit_pos(ma, ma->cur_bit, 1);
 		else
-			bitvec_set_bit_pos(ma, bitnum, 0);
+			bitvec_set_bit_pos(ma, ma->cur_bit, 0);
 	}
+}
 
-	return 0;
+static void generate_ma_for_bts(struct gsm_bts *bts)
+{
+	struct gsm_bts_trx *trx;
+	unsigned int tn;
+
+	OSMO_ASSERT(bts->si_common.cell_chan_num > 0);
+	OSMO_ASSERT(bts->si_common.cell_chan_num <= 64);
+
+	llist_for_each_entry(trx, &bts->trx_list, list) {
+		for (tn = 0; tn < ARRAY_SIZE(trx->ts); tn++)
+			generate_ma_for_ts(&trx->ts[tn]);
+	}
 }
 
 static void bootstrap_rsl(struct gsm_bts_trx *trx)
 {
 	unsigned int i;
+	int rc;
 
-	LOGP(DRSL, LOGL_NOTICE, "bootstrapping RSL for BTS/TRX (%u/%u) "
+	LOG_TRX(trx, DRSL, LOGL_NOTICE, "bootstrapping RSL "
 		"on ARFCN %u using MCC-MNC %s LAC=%u CID=%u BSIC=%u\n",
-		trx->bts->nr, trx->nr, trx->arfcn,
-		osmo_plmn_name(&bsc_gsmnet->plmn),
+		trx->arfcn, osmo_plmn_name(&bsc_gsmnet->plmn),
 		trx->bts->location_area_code,
 		trx->bts->cell_identity, trx->bts->bsic);
 
@@ -299,10 +330,13 @@ static void bootstrap_rsl(struct gsm_bts_trx *trx)
 	 * This ensures that RACH control in system information is configured correctly.
 	 * TRX 0 should be usable and unlocked, otherwise starting ACC ramping is pointless.
 	 */
-	if (trx_is_usable(trx) && trx->mo.nm_state.administrative == NM_STATE_UNLOCKED)
+	if (trx_is_usable(trx))
 		acc_ramp_trigger(&trx->bts->acc_ramp);
 
-	gsm_bts_trx_set_system_infos(trx);
+	if (gsm_bts_trx_set_system_infos(trx) != 0) {
+		LOG_TRX(trx, DRSL, LOGL_ERROR, "Failed to generate System Information\n");
+		return;
+	}
 
 	if (trx->bts->type == GSM_BTS_TYPE_NOKIA_SITE) {
 		/* channel unspecific, power reduction in 2 dB steps */
@@ -310,163 +344,43 @@ static void bootstrap_rsl(struct gsm_bts_trx *trx)
 		rsl_nokia_si_end(trx);
 	}
 
+	if (trx->bts->model->power_ctrl_send_def_params != NULL) {
+		rc = trx->bts->model->power_ctrl_send_def_params(trx);
+		if (rc) {
+			LOG_TRX(trx, DRSL, LOGL_ERROR, "Failed to send default "
+				"MS/BS Power control parameters (rc=%d)\n", rc);
+			/* TODO: should we drop RSL connection here? */
+		}
+	}
+
 	for (i = 0; i < ARRAY_SIZE(trx->ts); i++) {
 		struct gsm_bts_trx_ts *ts = &trx->ts[i];
-		generate_ma_for_ts(ts);
 		OSMO_ASSERT(ts->fi);
 		osmo_fsm_inst_dispatch(ts->fi, TS_EV_RSL_READY, NULL);
 	}
 
-	/* Start CBCH transmit timer if CBCH is present */
-	if (trx->nr == 0 && gsm_bts_get_cbch(trx->bts))
-		bts_cbch_timer_schedule(trx->bts);
+	/* Drop all expired channel requests in the list */
+	abis_rsl_chan_rqd_queue_flush(trx->bts);
 }
 
-static void all_ts_dispatch_event(struct gsm_bts_trx *trx, uint32_t event)
+struct osmo_timer_list update_connection_stats_timer;
+
+/* Periodically call bsc_update_connection_stats() to keep stat items updated.
+ * It would be nicer to trigger this only when OML or RSL state is seen to flip. I tried hard to find all code paths
+ * that should call this and failed to get accurate results; this trivial timer covers all of them. */
+static void update_connection_stats_cb(void *data)
 {
-	int ts_i;
-	for (ts_i = 0; ts_i < ARRAY_SIZE(trx->ts); ts_i++) {
-		struct gsm_bts_trx_ts *ts = &trx->ts[ts_i];
-		if (ts->fi)
-			osmo_fsm_inst_dispatch(ts->fi, event, 0);
-	}
+	bsc_update_connection_stats(bsc_gsmnet);
+	osmo_timer_setup(&update_connection_stats_timer, update_connection_stats_cb, NULL);
+	osmo_timer_schedule(&update_connection_stats_timer, 1, 0);
 }
 
-/* Callback function to be called every time we receive a signal from INPUT */
-static int inp_sig_cb(unsigned int subsys, unsigned int signal,
-		      void *handler_data, void *signal_data)
+static void bootstrap_bts(struct gsm_bts *bts)
 {
-	struct input_signal_data *isd = signal_data;
-	struct gsm_bts_trx *trx = isd->trx;
-	/* N. B: we rely on attribute order when parsing response in abis_nm_rx_get_attr_resp() */
-	const uint8_t bts_attr[] = { NM_ATT_MANUF_ID, NM_ATT_SW_CONFIG, };
-	const uint8_t trx_attr[] = { NM_ATT_MANUF_STATE, NM_ATT_SW_CONFIG, };
-
-	/* we should not request more attributes than we're ready to handle */
-	OSMO_ASSERT(sizeof(bts_attr) < MAX_BTS_ATTR);
-	OSMO_ASSERT(sizeof(trx_attr) < MAX_BTS_ATTR);
-
-	if (subsys != SS_L_INPUT)
-		return -EINVAL;
-
-	LOGP(DLMI, LOGL_DEBUG, "%s(): Input signal '%s' received\n", __func__,
-		get_value_string(e1inp_signal_names, signal));
-	switch (signal) {
-	case S_L_INP_TEI_UP:
-		if (isd->link_type == E1INP_SIGN_OML) {
-			/* TODO: this is required for the Nokia BTS, hopping is configured
-			   during OML, other MA is not set.  */
-			struct gsm_bts_trx *cur_trx;
-			/* was static in system_information.c */
-			extern int generate_cell_chan_list(uint8_t *chan_list, struct gsm_bts *bts);
-			uint8_t ca[20];
-			/* has to be called before generate_ma_for_ts to
-			  set bts->si_common.cell_alloc */
-			generate_cell_chan_list(ca, trx->bts);
-
-			/* Request generic BTS-level attributes */
-			abis_nm_get_attr(trx->bts, NM_OC_BTS, 0xFF, 0xFF, 0xFF, bts_attr, sizeof(bts_attr));
-
-			llist_for_each_entry(cur_trx, &trx->bts->trx_list, list) {
-				int i;
-				/* Request TRX-level attributes */
-				abis_nm_get_attr(cur_trx->bts, NM_OC_BASEB_TRANSC, 0, cur_trx->nr, 0xFF,
-						 trx_attr, sizeof(trx_attr));
-				for (i = 0; i < ARRAY_SIZE(cur_trx->ts); i++)
-					generate_ma_for_ts(&cur_trx->ts[i]);
-			}
-		}
-		if (isd->link_type == E1INP_SIGN_RSL)
-			bootstrap_rsl(trx);
-		break;
-	case S_L_INP_TEI_DN:
-		LOGP(DLMI, LOGL_ERROR, "Lost some E1 TEI link: %d %p\n", isd->link_type, trx);
-
-		if (isd->link_type == E1INP_SIGN_OML) {
-			rate_ctr_inc(&trx->bts->bts_ctrs->ctr[BTS_CTR_BTS_OML_FAIL]);
-			all_ts_dispatch_event(trx, TS_EV_OML_DOWN);
-		} else if (isd->link_type == E1INP_SIGN_RSL) {
-			rate_ctr_inc(&trx->bts->bts_ctrs->ctr[BTS_CTR_BTS_RSL_FAIL]);
-			acc_ramp_abort(&trx->bts->acc_ramp);
-			all_ts_dispatch_event(trx, TS_EV_RSL_DOWN);
-			if (trx->nr == 0)
-				osmo_timer_del(&trx->bts->cbch_timer);
-		}
-
-		gsm_bts_mo_reset(trx->bts);
-
-		abis_nm_clear_queue(trx->bts);
-		break;
-	default:
-		break;
-	}
-
-	return 0;
-}
-
-static int bootstrap_bts(struct gsm_bts *bts)
-{
-	struct gsm_bts_trx *trx;
 	unsigned int n = 0;
 
-	if (!bts->model)
-		return -EFAULT;
-
-	if (bts->model->start && !bts->model->started) {
-		int ret = bts->model->start(bts->network);
-		if (ret < 0)
-			return ret;
-
-		bts->model->started = true;
-	}
-
-	/* FIXME: What about secondary TRX of a BTS?  What about a BTS that has TRX
-	 * in different bands? Why is 'band' a parameter of the BTS and not of the TRX? */
-	switch (bts->band) {
-	case GSM_BAND_1800:
-		if (bts->c0->arfcn < 512 || bts->c0->arfcn > 885) {
-			LOGP(DNM, LOGL_ERROR, "GSM1800 channel must be between 512-885.\n");
-			return -EINVAL;
-		}
-		break;
-	case GSM_BAND_1900:
-		if (bts->c0->arfcn < 512 || bts->c0->arfcn > 810) {
-			LOGP(DNM, LOGL_ERROR, "GSM1900 channel must be between 512-810.\n");
-			return -EINVAL;
-		}
-		break;
-	case GSM_BAND_900:
-		if ((bts->c0->arfcn > 124 && bts->c0->arfcn < 955) ||
-		    bts->c0->arfcn > 1023)  {
-			LOGP(DNM, LOGL_ERROR, "GSM900 channel must be between 0-124, 955-1023.\n");
-			return -EINVAL;
-		}
-		break;
-	case GSM_BAND_850:
-		if (bts->c0->arfcn < 128 || bts->c0->arfcn > 251) {
-			LOGP(DNM, LOGL_ERROR, "GSM850 channel must be between 128-251.\n");
-			return -EINVAL;
-		}
-		break;
-	default:
-		LOGP(DNM, LOGL_ERROR, "Unsupported frequency band.\n");
-		return -EINVAL;
-	}
-
-	/* Verify the physical channel mapping */
-	llist_for_each_entry(trx, &bts->trx_list, list) {
-		if (!trx_has_valid_pchan_config(trx)) {
-			LOGP(DNM, LOGL_ERROR, "TRX %u has invalid timeslot "
-					      "configuration\n", trx->nr);
-			return -EINVAL;
-		}
-	}
-
 	/* Control Channel Description is set from vty/config */
 
-	/* Indicate R99 MSC in SI3 */
-	bts->si_common.chan_desc.mscr = 1;
-
 	/* Determine the value of CCCH_CONF. Is TS0/C0 combined? */
 	if (bts->c0->ts[0].pchan_from_config != GSM_PCHAN_CCCH) {
 		bts->si_common.chan_desc.ccch_conf = RSL_BCCH_CCCH_CONF_1_C;
@@ -487,18 +401,75 @@ static int bootstrap_bts(struct gsm_bts *bts)
 		bts->si_common.chan_desc.ccch_conf = (n << 1);
 	}
 
-	bts->si_common.cell_options.pwrc = 0; /* PWRC not set */
-
-	bts->si_common.cell_sel_par.acs = 0;
-
-	bts->si_common.ncc_permitted = 0xff;
-
-	bts->chan_load_samples_idx = 0;
-
 	/* ACC ramping is initialized from vty/config */
 
 	/* Initialize the BTS state */
-	gsm_bts_mo_reset(bts);
+	gsm_bts_sm_mo_reset(bts->site_mgr);
+
+	/* Generate Mobile Allocation bit-masks for all timeslots.
+	 * This needs to be done here, because it's used for TS configuration. */
+	generate_ma_for_bts(bts);
+}
+
+/* Callback function to be called every time we receive a signal from INPUT */
+static int inp_sig_cb(unsigned int subsys, unsigned int signal,
+		      void *handler_data, void *signal_data)
+{
+	struct input_signal_data *isd = signal_data;
+	struct gsm_bts_trx *trx = isd->trx;
+	int rc;
+
+	if (subsys != SS_L_INPUT)
+		return -EINVAL;
+
+	LOGP(DLMI, LOGL_DEBUG, "%s(): Input signal '%s' received\n", __func__,
+		get_value_string(e1inp_signal_names, signal));
+	switch (signal) {
+	case S_L_INP_TEI_UP:
+		if (isd->link_type == E1INP_SIGN_OML) {
+			/* Check parameters and apply vty config dependent parameters */
+			rc = gsm_bts_check_cfg(trx->bts);
+			if (rc < 0) {
+				LOGP(DNM, LOGL_ERROR, "(bts=%u) Error in BTS configuration -- cannot bootstrap BTS\n",
+				     trx->bts->nr);
+				return rc;
+			}
+			bootstrap_bts(trx->bts);
+		}
+		if (isd->link_type == E1INP_SIGN_RSL) {
+			rc = gsm_bts_check_cfg(trx->bts);
+			if (rc < 0) {
+				LOGP(DNM, LOGL_ERROR, "(bts=%u) Error in BTS configuration -- cannot bootstrap RSL\n",
+				     trx->bts->nr);
+				return rc;
+			}
+			bootstrap_rsl(trx);
+		}
+		break;
+	case S_L_INP_TEI_DN:
+		LOG_TRX(trx, DLMI, LOGL_ERROR, "Lost E1 %s link\n", e1inp_signtype_name(isd->link_type));
+
+		if (isd->link_type == E1INP_SIGN_OML) {
+			rate_ctr_inc(rate_ctr_group_get_ctr(trx->bts->bts_ctrs, BTS_CTR_BTS_OML_FAIL));
+			/* ip.access BTS models have a single global A-bis/OML link for all
+			 * transceivers, so once it's lost we need to notify them all. */
+			if (is_ipaccess_bts(trx->bts))
+				gsm_bts_all_ts_dispatch(trx->bts, TS_EV_OML_DOWN, NULL);
+			else /* Other BTS models (e.g. Ericsson) have per-TRX OML links */
+				gsm_trx_all_ts_dispatch(trx, TS_EV_OML_DOWN, NULL);
+		} else if (isd->link_type == E1INP_SIGN_RSL) {
+			rate_ctr_inc(rate_ctr_group_get_ctr(trx->bts->bts_ctrs, BTS_CTR_BTS_RSL_FAIL));
+			acc_ramp_abort(&trx->bts->acc_ramp);
+			gsm_trx_all_ts_dispatch(trx, TS_EV_RSL_DOWN, NULL);
+		}
+
+		gsm_bts_sm_mo_reset(trx->bts->site_mgr);
+
+		abis_nm_clear_queue(trx->bts);
+		break;
+	default:
+		break;
+	}
 
 	return 0;
 }
@@ -524,11 +495,12 @@ static int bsc_network_configure(const char *config_file)
 	osmo_signal_register_handler(SS_L_INPUT, inp_sig_cb, NULL);
 
 	llist_for_each_entry(bts, &bsc_gsmnet->bts_list, list) {
-		rc = bootstrap_bts(bts);
+		rc = gsm_bts_check_cfg(bts);
 		if (rc < 0) {
-			LOGP(DNM, LOGL_FATAL, "Error bootstrapping BTS\n");
+			LOGP(DNM, LOGL_FATAL, "(bts=%u) cannot bootstrap BTS, invalid BTS configuration\n", bts->nr);
 			return rc;
 		}
+		bootstrap_bts(bts);
 		rc = e1_reconfig_bts(bts);
 		if (rc < 0) {
 			LOGP(DNM, LOGL_FATAL, "Error enabling E1 input driver\n");
@@ -555,6 +527,21 @@ static int bsc_vty_go_parent(struct vty *vty)
 			vty->index_sub = NULL;
 		}
 		break;
+	case POWER_CTRL_NODE:
+		vty->node = BTS_NODE;
+		{
+			const struct gsm_power_ctrl_params *cp = vty->index;
+			struct gsm_bts *bts;
+
+			if (cp->dir == GSM_PWR_CTRL_DIR_UL)
+				bts = container_of(cp, struct gsm_bts, ms_power_ctrl);
+			else
+				bts = container_of(cp, struct gsm_bts, bs_power_ctrl);
+
+			vty->index_sub = &bts->description;
+			vty->index = bts;
+		}
+		break;
 	case TRX_NODE:
 		vty->node = BTS_NODE;
 		{
@@ -570,7 +557,6 @@ static int bsc_vty_go_parent(struct vty *vty)
 			/* set vty->index correctly ! */
 			struct gsm_bts_trx_ts *ts = vty->index;
 			vty->index = ts->trx;
-			vty->index_sub = &ts->trx->description;
 		}
 		break;
 	case OML_NODE:
@@ -635,32 +621,55 @@ static struct vty_app_info vty_info = {
 	.version	= PACKAGE_VERSION,
 	.go_parent_cb	= bsc_vty_go_parent,
 	.is_config_node	= bsc_vty_is_config_node,
+	.usr_attr_desc	= {
+		[BSC_VTY_ATTR_RESTART_ABIS_OML_LINK] = \
+			"This command applies on A-bis OML link (re)establishment",
+		[BSC_VTY_ATTR_RESTART_ABIS_RSL_LINK] = \
+			"This command applies on A-bis RSL link (re)establishment",
+		[BSC_VTY_ATTR_NEW_LCHAN] = \
+			"This command applies for newly created lchans",
+		[BSC_VTY_ATTR_VENDOR_SPECIFIC] = \
+			"This command/parameter is BTS vendor specific",
+	},
+	.usr_attr_letters = {
+		[BSC_VTY_ATTR_RESTART_ABIS_OML_LINK]	= 'o',
+		[BSC_VTY_ATTR_RESTART_ABIS_RSL_LINK]	= 'r',
+		[BSC_VTY_ATTR_NEW_LCHAN]		= 'l',
+		[BSC_VTY_ATTR_VENDOR_SPECIFIC]		= 'v',
+	},
 };
 
 extern int bsc_shutdown_net(struct gsm_network *net);
-static void signal_handler(int signal)
+static void signal_handler(int signum)
 {
-	fprintf(stdout, "signal %u received\n", signal);
+	fprintf(stdout, "signal %u received\n", signum);
 
-	switch (signal) {
+	switch (signum) {
 	case SIGINT:
 	case SIGTERM:
+		/* If SIGTERM was already sent before, just terminate immediately. */
+		if (osmo_select_shutdown_requested())
+			exit(-1);
 		bsc_shutdown_net(bsc_gsmnet);
 		osmo_signal_dispatch(SS_L_GLOBAL, S_L_GLOBAL_SHUTDOWN, NULL);
-		sleep(3);
-		exit(0);
+		osmo_select_shutdown_request();
 		break;
 	case SIGABRT:
-		/* in case of abort, we want to obtain a talloc report
-		 * and then return to the caller, who will abort the process */
+		/* in case of abort, we want to obtain a talloc report and
+		 * then run default SIGABRT handler, who will generate coredump
+		 * and abort the process. abort() should do this for us after we
+		 * return, but program wouldn't exit if an external SIGABRT is
+		 * received.
+		 */
+		talloc_report(tall_vty_ctx, stderr);
+		talloc_report_full(tall_bsc_ctx, stderr);
+		signal(SIGABRT, SIG_DFL);
+		raise(SIGABRT);
+		break;
 	case SIGUSR1:
 		talloc_report(tall_vty_ctx, stderr);
 		talloc_report_full(tall_bsc_ctx, stderr);
 		break;
-	case SIGUSR2:
-		if (!bsc_gsmnet->bsc_data)
-			return;
-		break;
 	default:
 		break;
 	}
@@ -713,7 +722,7 @@ static const struct log_info_cat osmo_bsc_categories[] = {
 		.name = "DNM",
 		.description = "A-bis Network Management / O&M (NM/OML)",
 		.color = "\033[1;36m",
-		.enabled = 1, .loglevel = LOGL_INFO,
+		.enabled = 1, .loglevel = LOGL_NOTICE,
 	},
 	[DPAG]	= {
 		.name = "DPAG",
@@ -748,11 +757,6 @@ static const struct log_info_cat osmo_bsc_categories[] = {
 		.description = "Reference Counting",
 		.enabled = 0, .loglevel = LOGL_NOTICE,
 	},
-	[DNAT] = {
-		.name = "DNAT",
-		.description = "GSM 08.08 NAT/Multiplexer",
-		.enabled = 1, .loglevel = LOGL_NOTICE,
-	},
 	[DCTRL] = {
 		.name = "DCTRL",
 		.description = "Control interface",
@@ -761,12 +765,12 @@ static const struct log_info_cat osmo_bsc_categories[] = {
 	[DFILTER] = {
 		.name = "DFILTER",
 		.description = "BSC/NAT IMSI based filtering",
-		.enabled = 1, .loglevel = LOGL_DEBUG,
+		.enabled = 1, .loglevel = LOGL_NOTICE,
 	},
 	[DPCU] = {
 		.name = "DPCU",
 		.description = "PCU Interface",
-		.enabled = 1, .loglevel = LOGL_DEBUG,
+		.enabled = 1, .loglevel = LOGL_NOTICE,
 	},
 	[DLCLS] = {
 		.name = "DLCLS",
@@ -777,8 +781,23 @@ static const struct log_info_cat osmo_bsc_categories[] = {
 		.name = "DCBS",
 		.description = "Cell Broadcast System",
 		.enabled = 1, .loglevel = LOGL_NOTICE,
-	}
-
+	},
+	[DLCS] = {
+		.name = "DLCS",
+		.description = "Location Services",
+		.enabled = 1, .loglevel = LOGL_NOTICE,
+	},
+	[DRESET] = {
+		.name = "DRESET",
+		.description = "RESET/ACK on A and Lb interfaces",
+		.enabled = 1, .loglevel = LOGL_NOTICE,
+	},
+	[DLOOP] = {
+		.name = "DLOOP",
+		.description = "Control loops",
+		.color = "\033[0;34m",
+		.enabled = 1, .loglevel = LOGL_NOTICE,
+	},
 };
 
 static int filter_fn(const struct log_context *ctx, struct log_target *tar)
@@ -805,10 +824,44 @@ extern void *tall_fle_ctx;
 extern void *tall_tqe_ctx;
 extern void *tall_ctr_ctx;
 
+static int bsc_mgw_setup(void)
+{
+	struct mgcp_client *mgcp_client_single;
+	unsigned int pool_members_initalized;
+
+	/* Initalize MGW pool. This initalizes and connects all MGCP clients that are currently configured in
+	 * the pool. Adding additional MGCP clients to the pool is possible but the user has to configure and
+	 * (re)connect them manually from the VTY. */
+	pool_members_initalized = mgcp_client_pool_connect(bsc_gsmnet->mgw.mgw_pool);
+	if (pool_members_initalized) {
+		LOGP(DNM, LOGL_NOTICE,
+		     "MGW pool with %u pool members configured, (ignoring MGW configuration in VTY node 'msc').\n",
+		     pool_members_initalized);
+		return 0;
+	}
+
+	/* Initialize and connect a single MGCP client. This MGCP client will appear as the one and only pool
+	 * member if there is no MGW pool configured. */
+	LOGP(DNM, LOGL_NOTICE, "No MGW pool configured, using MGW configuration in VTY node 'msc'\n");
+        mgcp_client_single = mgcp_client_init(bsc_gsmnet, bsc_gsmnet->mgw.conf);
+	if (!mgcp_client_single) {
+		LOGP(DNM, LOGL_ERROR, "MGW (single) client initalization failed\n");
+		return -EINVAL;
+	}
+	if (mgcp_client_connect(mgcp_client_single)) {
+		LOGP(DNM, LOGL_ERROR, "MGW (single) connect failed at (%s:%u)\n",
+		     bsc_gsmnet->mgw.conf->remote_addr,
+		     bsc_gsmnet->mgw.conf->remote_port);
+		return -EINVAL;
+	}
+	mgcp_client_pool_register_single(bsc_gsmnet->mgw.mgw_pool, mgcp_client_single);
+
+	return 0;
+}
+
 int main(int argc, char **argv)
 {
 	struct bsc_msc_data *msc;
-	struct osmo_bsc_data *data;
 	int rc;
 
 	tall_bsc_ctx = talloc_named_const(NULL, 1, "osmo-bsc");
@@ -827,6 +880,7 @@ int main(int argc, char **argv)
 	rate_ctr_init(tall_bsc_ctx);
 
 	osmo_fsm_set_dealloc_ctx(OTC_SELECT);
+	osmo_fsm_log_timeouts(true);
 
 	/* Allocate global gsm_network struct */
 	rc = bsc_network_alloc();
@@ -836,6 +890,7 @@ int main(int argc, char **argv)
 	}
 
 	bsc_gsmnet->mgw.conf = talloc_zero(bsc_gsmnet, struct mgcp_client_conf);
+	bsc_gsmnet->mgw.mgw_pool = mgcp_client_pool_alloc(bsc_gsmnet);
 	mgcp_client_conf_init(bsc_gsmnet->mgw.conf);
 
 	bts_init();
@@ -846,10 +901,11 @@ int main(int argc, char **argv)
 	/* This needs to precede handle_options() */
 	vty_init(&vty_info);
 	bsc_vty_init(bsc_gsmnet);
-	bsc_msg_acc_lst_vty_init(tall_bsc_ctx, &access_lists, BSC_NODE);
 	ctrl_vty_init(tall_bsc_ctx);
+	osmo_cpu_sched_vty_init(tall_bsc_ctx);
 	logging_vty_add_deprecated_subsys(tall_bsc_ctx, "cc");
 	logging_vty_add_deprecated_subsys(tall_bsc_ctx, "mgcp");
+	logging_vty_add_deprecated_subsys(tall_bsc_ctx, "nat");
 
 	/* Initialize SS7 */
 	OSMO_ASSERT(osmo_ss7_init() == 0);
@@ -862,11 +918,14 @@ int main(int argc, char **argv)
 	/* seed the PRNG */
 	srand(time(NULL));
 
-	ts_fsm_init();
 	lchan_fsm_init();
 	bsc_subscr_conn_fsm_init();
 	assignment_fsm_init();
 	handover_fsm_init();
+	lb_init();
+	acc_ramp_global_init();
+	paging_global_init();
+	smscb_global_init();
 
 	/* Read the config */
 	rc = bsc_network_configure(config_file);
@@ -875,6 +934,11 @@ int main(int argc, char **argv)
 		exit(1);
 	}
 
+	if (neighbors_check_cfg()) {
+		fprintf(stderr, "Errors in neighbor configuration, check the DHO log. exiting.\n");
+		exit(1);
+	}
+
 	/* start control interface after reading config for
 	 * ctrl_vty_get_bind_addr() */
 	bsc_gsmnet->ctrl = bsc_controlif_setup(bsc_gsmnet,
@@ -891,37 +955,51 @@ int main(int argc, char **argv)
 		exit(1);
 	}
 
-	data = bsc_gsmnet->bsc_data;
-	if (rf_ctrl)
-		osmo_talloc_replace_string(data, &data->rf_ctrl_name, rf_ctrl);
-
-	data->rf_ctrl = osmo_bsc_rf_create(data->rf_ctrl_name, bsc_gsmnet);
-	if (!data->rf_ctrl) {
-		fprintf(stderr, "Failed to create the RF service.\n");
-		exit(1);
-	}
-
-	rc = check_codec_pref(&bsc_gsmnet->bsc_data->mscs);
-	if (rc < 0)
-		LOGP(DMSC, LOGL_ERROR, "Configuration contains mutually exclusive codec settings -- check configuration!\n");
-
-	llist_for_each_entry(msc, &bsc_gsmnet->bsc_data->mscs, entry) {
-		if (osmo_bsc_msc_init(msc) != 0) {
-			LOGP(DNAT, LOGL_ERROR, "Failed to start up. Exiting.\n");
+	if (bsc_gsmnet->neigh_ctrl.addr) {
+		bsc_gsmnet->neigh_ctrl.handle = neighbor_controlif_setup(bsc_gsmnet);
+		if (!bsc_gsmnet->neigh_ctrl.handle) {
+			fprintf(stderr, "Failed to bind Neighbor Resolution Service. Exiting.\n");
+			exit(1);
+		}
+		rc = neighbor_ctrl_cmds_install(bsc_gsmnet);
+		if (rc < 0) {
+			fprintf(stderr, "Failed to install Neighbor Resolution Service commands. Exiting.\n");
 			exit(1);
 		}
 	}
 
-	bsc_gsmnet->mgw.client = mgcp_client_init(bsc_gsmnet, bsc_gsmnet->mgw.conf);
+	if (rf_ctrl)
+		osmo_talloc_replace_string(bsc_gsmnet, &bsc_gsmnet->rf_ctrl_name, rf_ctrl);
 
-	if (mgcp_client_connect(bsc_gsmnet->mgw.client)) {
-		LOGP(DNM, LOGL_ERROR, "MGW connect failed at (%s:%u)\n",
-		     bsc_gsmnet->mgw.conf->remote_addr,
-		     bsc_gsmnet->mgw.conf->remote_port);
+	bsc_gsmnet->rf_ctrl = osmo_bsc_rf_create(bsc_gsmnet->rf_ctrl_name, bsc_gsmnet);
+	if (!bsc_gsmnet->rf_ctrl) {
+		fprintf(stderr, "Failed to create the RF service.\n");
 		exit(1);
 	}
 
-	if (osmo_bsc_sigtran_init(&bsc_gsmnet->bsc_data->mscs) != 0) {
+	rc = check_codec_pref(&bsc_gsmnet->mscs);
+	if (rc < 0) {
+		LOGP(DMSC, LOGL_ERROR, "Configuration contains mutually exclusive codec settings -- check"
+				       " configuration!\n");
+		if (!bsc_gsmnet->allow_unusable_timeslots) {
+			LOGP(DMSC, LOGL_ERROR, "You should really fix that! However, you can prevent OsmoBSC from"
+					       " stopping here by setting 'allow-unusable-timeslots' in the 'network'"
+					       " section of the config.\n");
+			exit(1);
+		}
+	}
+
+	llist_for_each_entry(msc, &bsc_gsmnet->mscs, entry) {
+		if (osmo_bsc_msc_init(msc) != 0) {
+			LOGP(DMSC, LOGL_ERROR, "Failed to start up. Exiting.\n");
+			exit(1);
+		}
+	}
+
+	if (bsc_mgw_setup() != 0)
+		exit(1);
+
+	if (osmo_bsc_sigtran_init(&bsc_gsmnet->mscs) != 0) {
 		LOGP(DNM, LOGL_ERROR, "Failed to initialize sigtran backhaul.\n");
 		exit(1);
 	}
@@ -929,6 +1007,7 @@ int main(int argc, char **argv)
 	handover_decision_1_init();
 	hodec2_init(bsc_gsmnet);
 	bsc_cbc_link_restart();
+	lb_start_or_stop();
 
 	signal(SIGINT, &signal_handler);
 	signal(SIGTERM, &signal_handler);
@@ -937,6 +1016,9 @@ int main(int argc, char **argv)
 	signal(SIGUSR2, &signal_handler);
 	osmo_init_ignore_signals();
 
+	update_connection_stats_cb(NULL);
+	chan_counts_sig_init();
+
 	if (daemonize) {
 		rc = osmo_daemonize();
 		if (rc < 0) {
@@ -945,7 +1027,7 @@ int main(int argc, char **argv)
 		}
 	}
 
-	while (1) {
+	while (!osmo_select_shutdown_done()) {
 		osmo_select_main_ctx(0);
 	}
 
