@@ -32,24 +32,18 @@
 #include <osmocom/bsc/osmo_bsc_grace.h>
 #include <osmocom/bsc/osmo_bsc_sigtran.h>
 #include <osmocom/bsc/a_reset.h>
-#include <osmocom/bsc/gsm_04_80.h>
 #include <osmocom/bsc/bsc_subscr_conn_fsm.h>
 #include <osmocom/bsc/gsm_data.h>
+#include <osmocom/bsc/bts.h>
+#include <osmocom/bsc/paging.h>
 #include <osmocom/mgcp_client/mgcp_common.h>
 
 /* A pointer to a list with all involved MSCs
  * (a copy of the pointer location submitted with osmo_bsc_sigtran_init() */
 static struct llist_head *msc_list;
 
-#define RESET_INTERVAL 1	/* sek */
-#define SCCP_MSG_MAXSIZE 1024
-#define CS7_POINTCODE_DEFAULT_OFFSET 2
-#define DEFAULT_ASP_REMOTE_IP "127.0.0.1"
-
-/* The SCCP stack will not assign connection IDs to us automatically, we
- * will do this ourselves using a counter variable, that counts one up
- * for every new connection */
-static uint32_t conn_id_counter;
+#define DEFAULT_ASP_LOCAL_IP "localhost"
+#define DEFAULT_ASP_REMOTE_IP "localhost"
 
 /* Helper function to Check if the given connection id is already assigned */
 static struct gsm_subscriber_connection *get_bsc_conn_by_conn_id(int conn_id)
@@ -65,22 +59,17 @@ static struct gsm_subscriber_connection *get_bsc_conn_by_conn_id(int conn_id)
 	return NULL;
 }
 
-/* Pick a free connection id */
-static int pick_free_conn_id(const struct bsc_msc_data *msc)
+struct gsm_subscriber_connection *bsc_conn_by_bsub(struct bsc_subscr *bsub)
 {
-	int conn_id = conn_id_counter;
-	int i;
+	struct gsm_subscriber_connection *conn;
+	if (!bsub)
+		return NULL;
 
-	for (i = 0; i < 0xFFFFFF; i++) {
-		conn_id++;
-		conn_id &= 0xFFFFFF;
-		if (get_bsc_conn_by_conn_id(conn_id) == false) {
-			conn_id_counter = conn_id;
-			return conn_id;
-		}
+	llist_for_each_entry(conn, &bsc_gsmnet->subscr_conns, entry) {
+		if (conn->bsub == bsub)
+			return conn;
 	}
-
-	return -1;
+	return NULL;
 }
 
 /* Patch regular BSSMAP RESET to add extra T to announce Osmux support (osmocom extension) */
@@ -92,19 +81,20 @@ static void _gsm0808_extend_announce_osmux(struct msgb *msg)
 }
 
 /* Send reset to MSC */
-static void osmo_bsc_sigtran_tx_reset(const struct bsc_msc_data *msc)
+void osmo_bsc_sigtran_tx_reset(const struct bsc_msc_data *msc)
 {
 	struct osmo_ss7_instance *ss7;
 	struct msgb *msg;
 
 	ss7 = osmo_ss7_instance_find(msc->a.cs7_instance);
 	OSMO_ASSERT(ss7);
-	LOGP(DMSC, LOGL_NOTICE, "Sending RESET to MSC: %s\n", osmo_sccp_addr_name(ss7, &msc->a.msc_addr));
+	LOGP(DRESET, LOGL_INFO, "Sending RESET to MSC: %s\n", osmo_sccp_addr_name(ss7, &msc->a.msc_addr));
 	msg = gsm0808_create_reset();
 
 	if (msc_is_aoip(msc) && msc->use_osmux != OSMUX_USAGE_OFF)
 		_gsm0808_extend_announce_osmux(msg);
 
+	rate_ctr_inc(rate_ctr_group_get_ctr(msc->msc_ctrs, MSC_CTR_BSSMAP_TX_UDT_RESET));
 	osmo_sccp_tx_unitdata_msg(msc->a.sccp_user, &msc->a.bsc_addr,
 				  &msc->a.msc_addr, msg);
 }
@@ -118,12 +108,13 @@ void osmo_bsc_sigtran_tx_reset_ack(const struct bsc_msc_data *msc)
 
 	ss7 = osmo_ss7_instance_find(msc->a.cs7_instance);
 	OSMO_ASSERT(ss7);
-	LOGP(DMSC, LOGL_NOTICE, "Sending RESET ACK to MSC: %s\n", osmo_sccp_addr_name(ss7, &msc->a.msc_addr));
+	LOGP(DRESET, LOGL_NOTICE, "Sending RESET ACK to MSC: %s\n", osmo_sccp_addr_name(ss7, &msc->a.msc_addr));
 	msg = gsm0808_create_reset_ack();
 
 	if (msc_is_aoip(msc) && msc->use_osmux != OSMUX_USAGE_OFF)
 		_gsm0808_extend_announce_osmux(msg);
 
+	rate_ctr_inc(rate_ctr_group_get_ctr(msc->msc_ctrs, MSC_CTR_BSSMAP_TX_UDT_RESET_ACK));
 	osmo_sccp_tx_unitdata_msg(msc->a.sccp_user, &msc->a.bsc_addr,
 				  &msc->a.msc_addr, msg);
 }
@@ -143,14 +134,14 @@ static struct bsc_msc_data *get_msc_by_addr(const struct osmo_sccp_addr *msc_add
 	return NULL;
 }
 
-/* Send data to MSC, use the connection id which MSC it is */
+/* Received data from MSC, use the connection id which MSC it is */
 static int handle_data_from_msc(struct gsm_subscriber_connection *conn, struct msgb *msg)
 {
 	msg->l3h = msgb_l2(msg);
 	return bsc_handle_dt(conn, msg, msgb_l2len(msg));
 }
 
-/* Sent unitdata to MSC, use the point code to determine which MSC it is */
+/* Received unitdata from MSC, use the point code to determine which MSC it is */
 static int handle_unitdata_from_msc(const struct osmo_sccp_addr *msc_addr, struct msgb *msg,
 				    const struct osmo_sccp_user *scu)
 {
@@ -215,7 +206,7 @@ refuse:
 	return rc;
 }
 
-/* Callback function, called by the SSCP stack when data arrives */
+/* Callback function, called by the SCCP stack when data arrives */
 static int sccp_sap_up(struct osmo_prim_hdr *oph, void *_scu)
 {
 	struct osmo_scu_prim *scu_prim = (struct osmo_scu_prim *)oph;
@@ -302,15 +293,15 @@ enum bsc_con osmo_bsc_sigtran_new_conn(struct gsm_subscriber_connection *conn, s
 
 	ss7 = osmo_ss7_instance_find(msc->a.cs7_instance);
 	OSMO_ASSERT(ss7);
-	LOGP(DMSC, LOGL_NOTICE, "Initializing resources for new SIGTRAN connection to MSC: %s...\n",
-	     osmo_sccp_addr_name(ss7, &msc->a.msc_addr));
+	LOGP(DMSC, LOGL_INFO, "Initializing resources for new SCCP connection to MSC %d: %s...\n",
+	     msc->nr, osmo_sccp_addr_name(ss7, &msc->a.msc_addr));
 
 	if (a_reset_conn_ready(msc) == false) {
-		LOGP(DMSC, LOGL_ERROR, "MSC is not connected. Dropping.\n");
+		LOGP(DMSC, LOGL_ERROR, "MSC %d is not connected. Dropping.\n", msc->nr);
 		return BSC_CON_REJECT_NO_LINK;
 	}
 
-	if (!bsc_grace_allow_new_connection(bts->network, bts)) {
+	if (bts && !bsc_grace_allow_new_connection(bts->network, bts)) {
 		LOGP(DMSC, LOGL_NOTICE, "BSC in grace period. No new connections.\n");
 		return BSC_CON_REJECT_RF_GRACE;
 	}
@@ -321,7 +312,8 @@ enum bsc_con osmo_bsc_sigtran_new_conn(struct gsm_subscriber_connection *conn, s
 }
 
 /* Open a new connection oriented sigtran connection */
-int osmo_bsc_sigtran_open_conn(struct gsm_subscriber_connection *conn, struct msgb *msg)
+/* Allow test to overwrite it */
+__attribute__((weak)) int osmo_bsc_sigtran_open_conn(struct gsm_subscriber_connection *conn, struct msgb *msg)
 {
 	struct osmo_ss7_instance *ss7;
 	struct bsc_msc_data *msc;
@@ -340,7 +332,7 @@ int osmo_bsc_sigtran_open_conn(struct gsm_subscriber_connection *conn, struct ms
 		return -EINVAL;
 	}
 
-	conn->sccp.conn_id = conn_id = pick_free_conn_id(msc);
+	conn->sccp.conn_id = conn_id = bsc_sccp_inst_next_conn_id(conn->sccp.msc->a.sccp);
 	if (conn->sccp.conn_id < 0) {
 		LOGP(DMSC, LOGL_ERROR, "Unable to allocate SCCP Connection ID\n");
 		return -1;
@@ -348,8 +340,8 @@ int osmo_bsc_sigtran_open_conn(struct gsm_subscriber_connection *conn, struct ms
 	LOGP(DMSC, LOGL_DEBUG, "Allocated new connection id: %d\n", conn->sccp.conn_id);
 	ss7 = osmo_ss7_instance_find(msc->a.cs7_instance);
 	OSMO_ASSERT(ss7);
-	LOGP(DMSC, LOGL_NOTICE, "Opening new SIGTRAN connection (id=%i) to MSC: %s\n", conn_id,
-	     osmo_sccp_addr_name(ss7, &msc->a.msc_addr));
+	LOGP(DMSC, LOGL_INFO, "Opening new SCCP connection (id=%i) to MSC %d: %s\n", conn_id,
+	     msc->nr, osmo_sccp_addr_name(ss7, &msc->a.msc_addr));
 
 	rc = osmo_sccp_tx_conn_req_msg(msc->a.sccp_user, conn_id, &msc->a.bsc_addr,
 				       &msc->a.msc_addr, msg);
@@ -383,20 +375,26 @@ int osmo_bsc_sigtran_send(struct gsm_subscriber_connection *conn, struct msgb *m
 	if (msg->len >= 3) {
 		switch (msg->data[0]) {
 		case BSSAP_MSG_BSS_MANAGEMENT:
+			rate_ctr_inc(rate_ctr_group_get_ctr(msc->msc_ctrs, MSC_CTR_BSSMAP_TX_BSS_MANAGEMENT));
 			LOGP(DMSC, LOGL_INFO, "Tx MSC: BSSMAP: %s\n",
 			     gsm0808_bssmap_name(msg->data[2]));
 			break;
 		case BSSAP_MSG_DTAP:
+			rate_ctr_inc(rate_ctr_group_get_ctr(msc->msc_ctrs, MSC_CTR_BSSMAP_TX_DTAP));
 			LOGP(DMSC, LOGL_INFO, "Tx MSC: DTAP\n");
 			break;
 		default:
+			rate_ctr_inc(rate_ctr_group_get_ctr(msc->msc_ctrs, MSC_CTR_BSSMAP_TX_UNKNOWN));
 			LOGP(DMSC, LOGL_ERROR, "Tx MSC: unknown message type: 0x%x\n",
 			     msg->data[0]);
 		}
-	} else
+	} else {
+		rate_ctr_inc(rate_ctr_group_get_ctr(msc->msc_ctrs, MSC_CTR_BSSMAP_TX_SHORT));
 		LOGP(DMSC, LOGL_ERROR, "Tx MSC: message too short: %u\n", msg->len);
+	}
 
 	if (a_reset_conn_ready(msc) == false) {
+		rate_ctr_inc(rate_ctr_group_get_ctr(msc->msc_ctrs, MSC_CTR_BSSMAP_TX_ERR_CONN_NOT_READY));
 		LOGP(DMSC, LOGL_ERROR, "MSC is not connected. Dropping.\n");
 		msgb_free(msg);
 		return -EINVAL;
@@ -410,43 +408,29 @@ int osmo_bsc_sigtran_send(struct gsm_subscriber_connection *conn, struct msgb *m
 	     conn_id, osmo_sccp_addr_name(ss7, &msc->a.msc_addr), osmo_hexdump(msg->data, msg->len));
 
 	rc = osmo_sccp_tx_data_msg(msc->a.sccp_user, conn_id, msg);
+	if (rc >= 0)
+		rate_ctr_inc(rate_ctr_group_get_ctr(msc->msc_ctrs, MSC_CTR_BSSMAP_TX_SUCCESS));
+	else
+		rate_ctr_inc(rate_ctr_group_get_ctr(msc->msc_ctrs, MSC_CTR_BSSMAP_TX_ERR_SEND));
 
 	return rc;
 }
 
-/* Send an USSD notification in case we loose the connection to the MSC */
-static void bsc_notify_msc_lost(struct gsm_subscriber_connection *conn)
-{
-	/* Check if sccp conn is still present */
-	if (!conn)
-		return;
-
-	/* check for config string */
-	if (!conn->sccp.msc->ussd_msc_lost_txt)
-		return;
-	if (conn->sccp.msc->ussd_msc_lost_txt[0] == '\0')
-		return;
-
-	/* send USSD notification */
-	bsc_send_ussd_notify(conn, 1, conn->sccp.msc->ussd_msc_lost_txt);
-	bsc_send_ussd_release_complete(conn);
-}
-
 /* Close all open sigtran connections and channels */
-void osmo_bsc_sigtran_reset(const struct bsc_msc_data *msc)
+void osmo_bsc_sigtran_reset(struct bsc_msc_data *msc)
 {
 	struct gsm_subscriber_connection *conn, *conn_temp;
 	OSMO_ASSERT(msc);
 
+	/* Drop all ongoing paging requests that this MSC has created on any BTS */
+	paging_flush_network(msc->network, msc);
+
 	/* Close all open connections */
 	llist_for_each_entry_safe(conn, conn_temp, &bsc_gsmnet->subscr_conns, entry) {
 
 		/* We only may close connections which actually belong to this
 		 * MSC. All other open connections are left untouched */
 		if (conn->sccp.msc == msc) {
-			/* Notify active connection users via USSD that the MSC is down */
-			bsc_notify_msc_lost(conn);
-
 			/* Take down all occopied RF channels */
 			/* Disconnect all Sigtran connections */
 			/* Delete subscriber connection */
@@ -455,18 +439,6 @@ void osmo_bsc_sigtran_reset(const struct bsc_msc_data *msc)
 	}
 }
 
-/* Callback function: Close all open connections */
-static void osmo_bsc_sigtran_reset_cb(const void *priv)
-{
-	struct bsc_msc_data *msc = (struct bsc_msc_data*) priv;
-
-	/* Shut down all ongoing traffic */
-	osmo_bsc_sigtran_reset(msc);
-
-	/* Send reset to MSC */
-	osmo_bsc_sigtran_tx_reset(msc);
-}
-
 /* Default point-code to be used as local address (BSC) */
 #define BSC_DEFAULT_PC "0.23.3"
 
@@ -478,66 +450,25 @@ static int asp_rx_unknown(struct osmo_ss7_asp *asp, int ppid_mux, struct msgb *m
 /* Initialize osmo sigtran backhaul */
 int osmo_bsc_sigtran_init(struct llist_head *mscs)
 {
-	bool free_attempt_used = false;
-	bool fail_on_next_invalid_cfg = false;
-
 	struct bsc_msc_data *msc;
-	char msc_name[32];
 	uint32_t default_pc;
+	struct osmo_ss7_instance *inst;
+	int create_instance_0_for_msc_nr = -1;
 
 	osmo_ss7_register_rx_unknown_cb(&asp_rx_unknown);
 
 	OSMO_ASSERT(mscs);
 	msc_list = mscs;
 
+	/* Guard against multiple MSCs with identical config */
 	llist_for_each_entry(msc, msc_list, entry) {
-		snprintf(msc_name, sizeof(msc_name), "msc-%u", msc->nr);
-		LOGP(DMSC, LOGL_NOTICE, "Initializing SCCP connection to MSC %s\n", msc_name);
+		struct bsc_msc_data *msc2;
 
-		/* Check if the VTY could determine a valid CS7 instance,
-		 * use safe default in case none is set */
-		if (msc->a.cs7_instance_valid == false) {
-			msc->a.cs7_instance = 0;
-			if (fail_on_next_invalid_cfg)
-				goto fail_auto_cofiguration;
-			free_attempt_used = true;
-		}
-		LOGP(DMSC, LOGL_NOTICE, "CS7 Instance identifier, A-Interface: %u\n", msc->a.cs7_instance);
+		/* An MSC with invalid cs7 instance defaults to cs7 instance 0 */
+		uint32_t msc_inst = (msc->a.cs7_instance_valid ? msc->a.cs7_instance : 0);
 
-		/* Pre-Check if there is an ss7 instance present */
-		if (osmo_ss7_instance_find(msc->a.cs7_instance) == NULL) {
-			if (fail_on_next_invalid_cfg)
-				goto fail_auto_cofiguration;
-			free_attempt_used = true;
-		}
-
-		/* SS7 Protocol stack */
-		default_pc = osmo_ss7_pointcode_parse(NULL, BSC_DEFAULT_PC);
-		msc->a.sccp =
-		    osmo_sccp_simple_client_on_ss7_id(msc, msc->a.cs7_instance, msc_name, default_pc,
-						      msc->a.asp_proto, 0, NULL, 0, DEFAULT_ASP_REMOTE_IP);
-		if (!msc->a.sccp)
-			return -EINVAL;
-
-		/* In SCCPlite, the MSC side of the MGW endpoint is configured by the MSC. Since we have
-		 * no way to figure out which CallID ('C:') the MSC will issue in its CRCX command, set
-		 * an X-Osmo-IGN flag telling osmo-mgw to ignore CallID mismatches for this endpoint.
-		 * If an explicit VTY command has already indicated whether or not to send X-Osmo-IGN, do
-		 * not overwrite that setting. */
-		if (msc_is_sccplite(msc) && !msc->x_osmo_ign_configured)
-			msc->x_osmo_ign |= MGCP_X_OSMO_IGN_CALLID;
-
-		/* If unset, use default local SCCP address */
-		if (!msc->a.bsc_addr.presence)
-			osmo_sccp_local_addr_by_instance(&msc->a.bsc_addr, msc->a.sccp,
-							 OSMO_SCCP_SSN_BSSAP);
-
-		if (!osmo_sccp_check_addr(&msc->a.bsc_addr, OSMO_SCCP_ADDR_T_SSN | OSMO_SCCP_ADDR_T_PC)) {
-			LOGP(DMSC, LOGL_ERROR,
-			     "(%s) A-interface: invalid local (BSC) SCCP address: %s\n",
-			     msc_name, osmo_sccp_inst_addr_name(msc->a.sccp, &msc->a.bsc_addr));
-			return -EINVAL;
-		}
+		if (!msc->a.cs7_instance_valid)
+			create_instance_0_for_msc_nr = msc->nr;
 
 		/* If unset, use default SCCP address for the MSC */
 		if (!msc->a.msc_addr.presence)
@@ -545,46 +476,147 @@ int osmo_bsc_sigtran_init(struct llist_head *mscs)
 						   osmo_ss7_pointcode_parse(NULL, MSC_DEFAULT_PC),
 						   OSMO_SCCP_SSN_BSSAP);
 
-		if (!osmo_sccp_check_addr(&msc->a.msc_addr, OSMO_SCCP_ADDR_T_SSN | OSMO_SCCP_ADDR_T_PC)) {
-			LOGP(DMSC, LOGL_ERROR,
-			     "(%s) A-interface: invalid remote (MSC) SCCP address: %s\n",
-			     msc_name, osmo_sccp_inst_addr_name(msc->a.sccp, &msc->a.msc_addr));
-			return -EINVAL;
+		/* (more optimally, we'd only iterate the remaining other mscs after this msc, but this happens only
+		 * during startup, so nevermind that complexity and rather check each pair twice. That also ensures to
+		 * compare all MSCs that have no explicit msc_addr set, see osmo_sccp_make_addr_pc_ssn() above.) */
+		llist_for_each_entry(msc2, msc_list, entry) {
+			uint32_t msc2_inst;
+
+			if (msc2 == msc)
+				continue;
+
+			msc2_inst = (msc2->a.cs7_instance_valid ? msc2->a.cs7_instance : 0);
+			if (msc_inst != msc2_inst)
+				continue;
+
+			if (osmo_sccp_addr_cmp(&msc->a.msc_addr, &msc2->a.msc_addr, OSMO_SCCP_ADDR_T_PC) == 0) {
+				LOGP(DMSC, LOGL_ERROR, "'msc %d' and 'msc %d' cannot use the same remote PC"
+				     " %s on the same cs7 instance %u\n",
+				     msc->nr, msc2->nr, osmo_sccp_addr_dump(&msc->a.msc_addr), msc_inst);
+				return -EINVAL;
+			}
+		}
+	}
+
+	if (create_instance_0_for_msc_nr >= 0 && !osmo_ss7_instance_find(0)) {
+		LOGP(DMSC, LOGL_NOTICE, "To auto-configure msc %d, creating cs7 instance 0 implicitly\n",
+		     create_instance_0_for_msc_nr);
+		OSMO_ASSERT(osmo_ss7_instance_find_or_create(tall_bsc_ctx, 0));
+	}
+
+	/* Set up exactly one SCCP user and one ASP+AS per cs7 instance.
+	 * Iterate cs7 instance indexes and see for each one whether an MSC is configured for it.
+	 * The 'msc' / 'msc-addr' command selects the cs7 instance used for an MSC.
+	 */
+	llist_for_each_entry(inst, &osmo_ss7_instances, list) {
+		char inst_name[32];
+		enum osmo_ss7_asp_protocol used_proto = OSMO_SS7_ASP_PROT_NONE;
+		int prev_msc_nr;
+
+		struct osmo_sccp_instance *sccp;
+
+		llist_for_each_entry(msc, msc_list, entry) {
+			/* An MSC with invalid cs7 instance id defaults to cs7 instance 0 */
+			if ((inst->cfg.id != msc->a.cs7_instance)
+			    && !(inst->cfg.id == 0 && !msc->a.cs7_instance_valid))
+				continue;
+
+			/* This msc runs on this cs7 inst. Check the asp_proto. */
+			if (used_proto != OSMO_SS7_ASP_PROT_NONE
+			    && used_proto != msc->a.asp_proto) {
+				LOGP(DMSC, LOGL_ERROR, "'msc %d' and 'msc %d' with differing ASP protocols"
+				     " %s and %s cannot use the same cs7 instance %u\n",
+				     prev_msc_nr, msc->nr,
+				     osmo_ss7_asp_protocol_name(used_proto),
+				     osmo_ss7_asp_protocol_name(msc->a.asp_proto),
+				     inst->cfg.id);
+				return -EINVAL;
+			}
+
+			used_proto = msc->a.asp_proto;
+			prev_msc_nr = msc->nr;
+			/* still run through the other MSCs to catch asp_proto mismatches */
 		}
 
-		LOGP(DMSC, LOGL_NOTICE, "(%s) A-interface: local (BSC) SCCP address: %s\n",
-		     msc_name, osmo_sccp_inst_addr_name(msc->a.sccp, &msc->a.bsc_addr));
-		LOGP(DMSC, LOGL_NOTICE, "(%s) A-interface: remote (MSC) SCCP address: %s\n",
-		     msc_name, osmo_sccp_inst_addr_name(msc->a.sccp, &msc->a.msc_addr));
+		if (used_proto == OSMO_SS7_ASP_PROT_NONE) {
+			/* This instance has no MSC associated with it */
+			LOGP(DMSC, LOGL_ERROR, "cs7 instance %u has no MSCs configured to run on it\n", inst->cfg.id);
+			continue;
+		}
 
-		/* Bind SCCP user */
-		msc->a.sccp_user = osmo_sccp_user_bind(msc->a.sccp, msc_name, sccp_sap_up, msc->a.bsc_addr.ssn);
-		if (!msc->a.sccp_user)
+		snprintf(inst_name, sizeof(inst_name), "A-%u-%s", inst->cfg.id, osmo_ss7_asp_protocol_name(used_proto));
+		LOGP(DMSC, LOGL_NOTICE, "Initializing SCCP connection for A/%s on cs7 instance %u\n",
+		     osmo_ss7_asp_protocol_name(used_proto), inst->cfg.id);
+
+		/* SS7 Protocol stack */
+		default_pc = osmo_ss7_pointcode_parse(NULL, BSC_DEFAULT_PC);
+		sccp = osmo_sccp_simple_client_on_ss7_id(tall_bsc_ctx, inst->cfg.id, inst_name,
+							 default_pc, used_proto,
+							 0, DEFAULT_ASP_LOCAL_IP,
+							 0, DEFAULT_ASP_REMOTE_IP);
+		if (!sccp)
 			return -EINVAL;
 
-		/* Start MSC-Reset procedure */
-		a_reset_alloc(msc, msc_name, osmo_bsc_sigtran_reset_cb);
+		/* Now that the SCCP client is set up, configure all MSCs on this cs7 instance to use it */
+		llist_for_each_entry(msc, msc_list, entry) {
+			char msc_name[32];
 
-		/* If we have detected that the SS7 configuration of the MSC we have just initialized
-		 * was incomplete or completely missing, we can not tolerate another incomplete
-		 * configuration. The reason for this is that we do only specify exactly one default
-		 * pointcode pair. We also specify localhost as default IP-Address. If we have wanted
-		 * to support multiple MSCs with automatic configuration we would be forced to invent
-		 * a complex ruleset how to allocate the pointcodes and respective IP-Addresses.
-		 * Furthermore, the situation where a single BSC is connected to multiple MSCs
-		 * is a very rare situation anyway. In this case we expect the user to experienced
-		 * enough to create a valid SS7/CS7 VTY configuration that does not lack any
-		 * components */
-		if (free_attempt_used)
-			fail_on_next_invalid_cfg = true;
+			/* Skip MSCs that don't run on this cs7 instance */
+			if ((inst->cfg.id != msc->a.cs7_instance)
+			    && !(inst->cfg.id == 0 && !msc->a.cs7_instance_valid))
+				continue;
+
+			snprintf(msc_name, sizeof(msc_name), "msc-%d", msc->nr);
+
+			msc->a.sccp = sccp;
+
+			/* In SCCPlite, the MSC side of the MGW endpoint is configured by the MSC. Since we have
+			 * no way to figure out which CallID ('C:') the MSC will issue in its CRCX command, set
+			 * an X-Osmo-IGN flag telling osmo-mgw to ignore CallID mismatches for this endpoint.
+			 * If an explicit VTY command has already indicated whether or not to send X-Osmo-IGN, do
+			 * not overwrite that setting. */
+			if (msc_is_sccplite(msc) && !msc->x_osmo_ign_configured)
+				msc->x_osmo_ign |= MGCP_X_OSMO_IGN_CALLID;
+
+			/* If unset, use default local SCCP address */
+			if (!msc->a.bsc_addr.presence)
+				osmo_sccp_local_addr_by_instance(&msc->a.bsc_addr, sccp,
+								 OSMO_SCCP_SSN_BSSAP);
+
+			if (!osmo_sccp_check_addr(&msc->a.bsc_addr, OSMO_SCCP_ADDR_T_SSN | OSMO_SCCP_ADDR_T_PC)) {
+				LOGP(DMSC, LOGL_ERROR,
+				     "%s %s: invalid local (BSC) SCCP address: %s\n",
+				     inst_name, msc_name, osmo_sccp_inst_addr_name(sccp, &msc->a.bsc_addr));
+				return -EINVAL;
+			}
+
+			if (!osmo_sccp_check_addr(&msc->a.msc_addr, OSMO_SCCP_ADDR_T_SSN | OSMO_SCCP_ADDR_T_PC)) {
+				LOGP(DMSC, LOGL_ERROR,
+				     "%s %s: invalid remote (MSC) SCCP address: %s\n",
+				     inst_name, msc_name, osmo_sccp_inst_addr_name(sccp, &msc->a.msc_addr));
+				return -EINVAL;
+			}
+
+			LOGP(DMSC, LOGL_NOTICE, "%s %s: local (BSC) SCCP address: %s\n",
+			     inst_name, msc_name, osmo_sccp_inst_addr_name(sccp, &msc->a.bsc_addr));
+			LOGP(DMSC, LOGL_NOTICE, "%s %s: remote (MSC) SCCP address: %s\n",
+			     inst_name, msc_name, osmo_sccp_inst_addr_name(sccp, &msc->a.msc_addr));
+
+			/* Bind SCCP user. Bind only one user per sccp_instance and bsc_addr. */
+			msc->a.sccp_user = osmo_sccp_user_find(sccp, msc->a.bsc_addr.ssn, msc->a.bsc_addr.pc);
+			LOGP(DMSC, LOGL_NOTICE, "%s %s: %s\n", inst_name, msc_name,
+			     msc->a.sccp_user ? "user already bound for this SCCP instance" : "binding SCCP user");
+			if (!msc->a.sccp_user)
+				msc->a.sccp_user = osmo_sccp_user_bind(sccp, msc_name, sccp_sap_up, msc->a.bsc_addr.ssn);
+			if (!msc->a.sccp_user)
+				return -EINVAL;
+
+			/* Start MSC-Reset procedure */
+			a_reset_alloc(msc, msc_name);
+		}
 	}
 
 	return 0;
-
-fail_auto_cofiguration:
-	LOGP(DMSC, LOGL_ERROR,
-	     "A-interface: More than one invalid/inclomplete configuration detected, unable to revover - check config file!\n");
-	return -EINVAL;
 }
 
 /* this function receives all messages received on an ASP for a PPID / StreamID that
