@@ -22,10 +22,12 @@
 #include <limits.h>
 
 #include <osmocom/core/stats.h>
+#include <osmocom/core/utils.h>
 #include <osmocom/core/select.h>
 #include <osmocom/core/msgb.h>
 #include <osmocom/core/talloc.h>
 #include <osmocom/core/byteswap.h>
+#include <osmocom/core/signal.h>
 
 #include <osmocom/gsm/cbsp.h>
 #include <osmocom/gsm/protocol/gsm_23_041.h>
@@ -37,10 +39,11 @@
 #include <osmocom/bsc/debug.h>
 #include <osmocom/bsc/gsm_data.h>
 #include <osmocom/bsc/smscb.h>
-#include <osmocom/bsc/vty.h>
 #include <osmocom/bsc/gsm_04_08_rr.h>
 #include <osmocom/bsc/lchan_fsm.h>
 #include <osmocom/bsc/abis_rsl.h>
+#include <osmocom/bsc/bts.h>
+#include <osmocom/bsc/signal.h>
 
 /*********************************************************************************
  * Helper Functions
@@ -57,8 +60,6 @@ static void llist_replace_head(struct llist_head *new, struct llist_head *old)
 	INIT_LLIST_HEAD(old);
 }
 
-#define ETWS_PRIM_NOTIF_SIZE	56
-
 /* Build a ETWS Primary Notification message as per TS 23.041 9.4.1.3 */
 static int gen_etws_primary_notification(uint8_t *out, uint16_t serial_nr, uint16_t msg_id,
 					 uint16_t warn_type, const uint8_t *sec_info)
@@ -70,13 +71,24 @@ static int gen_etws_primary_notification(uint8_t *out, uint16_t serial_nr, uint1
 	osmo_store16be(serial_nr, out);
 	etws->msg_id = osmo_htons(msg_id);
 	etws->warning_type = osmo_htons(warn_type);
-
-	if (sec_info)
-		memcpy(etws->data, sec_info, ETWS_PRIM_NOTIF_SIZE - sizeof(*etws));
+	memcpy(etws->data, sec_info, ETWS_PRIM_NOTIF_SIZE - sizeof(*etws));
 
 	return ETWS_PRIM_NOTIF_SIZE;
 }
 
+static void bts_cbch_init_state(struct bts_smscb_chan_state *cstate, struct gsm_bts *bts)
+{
+	cstate->bts = bts;
+	INIT_LLIST_HEAD(&cstate->messages);
+}
+
+void bts_cbch_init(struct gsm_bts *bts)
+{
+	bts_cbch_init_state(&bts->cbch_basic, bts);
+	bts_cbch_init_state(&bts->cbch_extended, bts);
+	osmo_timer_setup(&bts->cbch_timer, &bts_cbch_timer_cb, bts);
+}
+
 /*! Obtain SMSCB Channel State for given BTS (basic or extended CBCH) */
 struct bts_smscb_chan_state *bts_get_smscb_chan(struct gsm_bts *bts, bool extended)
 {
@@ -109,6 +121,8 @@ static void __bts_smscb_add(struct bts_smscb_chan_state *cstate, struct bts_smsc
 			return;
 		}
 	}
+	/* we didn't find any messages with longer period than us, insert us at tail */
+	llist_add_tail(&new->list, &cstate->messages);
 }
 
 /* stringify a SMSCB for logging */
@@ -241,6 +255,12 @@ static void append_bcast_compl(struct response_state *r_state, struct gsm_bts *b
 	llist_add_tail(&cent->list, &r_state->num_completed.list);
 }
 
+static bool etws_msg_id_matches(uint16_t a, uint16_t b)
+{
+	/* ETWS messages are identified by the twelve most significant bits of the Message ID */
+	return (a & 0xFFF0) == (b & 0xFFF0);
+}
+
 /*! Iterate over all BTSs, find matching ones, execute command on BTS, add result
  *  to succeeded/failed lists.
  *  \param[in] net GSM network in which we operate
@@ -365,7 +385,7 @@ static struct bts_smscb_message *bts_smscb_msg_from_wrepl(struct gsm_bts *bts,
 	smscb->input.dcs = wrepl->u.cbs.dcs;
 	smscb->num_pages = llist_count(&wrepl->u.cbs.msg_content);
 	if (smscb->num_pages > ARRAY_SIZE(smscb->page)) {
-		LOG_BTS(bts, DCBS, LOGL_ERROR, "SMSCB with too many pages (%u > %lu)\n",
+		LOG_BTS(bts, DCBS, LOGL_ERROR, "SMSCB with too many pages (%u > %zu)\n",
 			smscb->num_pages, ARRAY_SIZE(smscb->page));
 		talloc_free(smscb);
 		return NULL;
@@ -381,6 +401,9 @@ static struct bts_smscb_message *bts_smscb_msg_from_wrepl(struct gsm_bts *bts,
 		page = &smscb->page[i++];
 		msg_param = (struct gsm23041_msg_param_gsm *) &page->data[0];
 
+		/* ensure we don't overflow in the memcpy below */
+		osmo_static_assert(sizeof(*page) > sizeof(*msg_param) + sizeof(cont->data), smscb_space);
+
 		/* build 6 byte header according to TS 23.041 9.4.1.2 */
 		osmo_store16be(wrepl->new_serial_nr, &msg_param->serial_nr);
 		osmo_store16be(wrepl->msg_id, &msg_param->message_id);
@@ -390,7 +413,9 @@ static struct bts_smscb_message *bts_smscb_msg_from_wrepl(struct gsm_bts *bts,
 
 		OSMO_ASSERT(cont->user_len <= ARRAY_SIZE(cont->data));
 		OSMO_ASSERT(cont->user_len <= ARRAY_SIZE(page->data) - sizeof(*msg_param));
-		memcpy(&msg_param->content, cont->data, cont->user_len);
+		/* we must not use cont->user_len as length here, as it would truncate any
+		 * possible 7-bit padding at the end. Always copy the whole page */
+		memcpy(&msg_param->content, cont->data, sizeof(cont->data));
 		bytes_used = sizeof(*msg_param) + cont->user_len;
 		/* compute number of valid blocks in page */
 		page->num_blocks = bytes_used / 22;
@@ -451,6 +476,54 @@ int cbsp_tx_restart(struct bsc_cbc_link *cbc, bool is_emerg)
 	return cbsp_tx_decoded(cbc, cbsp);
 }
 
+/* transmit a CBSP RESTART-INDICATION message stating a cell is operative again */
+int cbsp_tx_restart_bts(struct bsc_cbc_link *cbc, bool is_emerg, struct gsm_bts *bts)
+{
+	struct osmo_cbsp_decoded *cbsp = osmo_cbsp_decoded_alloc(cbc, CBSP_MSGT_RESTART);
+	struct osmo_cbsp_cell_ent cell_ent;
+
+	if (is_emerg)
+		cbsp->u.restart.bcast_msg_type = 0x01;
+	cbsp->u.restart.recovery_ind = 0x00; /* message data available */
+	cbsp->u.restart.cell_list.id_discr = CELL_IDENT_LAC_AND_CI;
+
+	cell_ent = (struct osmo_cbsp_cell_ent){
+		.cell_id = {
+			.lac_and_ci = {
+				.lac = bts->location_area_code,
+				.ci = bts->cell_identity,
+			}
+		}
+	};
+	llist_add(&cell_ent.list, &cbsp->u.restart.cell_list.list);
+
+	return cbsp_tx_decoded(cbc, cbsp);
+}
+
+/* transmit a CBSP FAILURE-INDICATION message stating all message data was lost for one cell */
+int cbsp_tx_failure_bts(struct bsc_cbc_link *cbc, bool is_emerg, struct gsm_bts *bts)
+{
+	struct osmo_cbsp_decoded *cbsp = osmo_cbsp_decoded_alloc(cbc, CBSP_MSGT_FAILURE);
+	struct osmo_cbsp_fail_ent fail_ent;
+
+	if (is_emerg)
+		cbsp->u.failure.bcast_msg_type = 0x01;
+
+	fail_ent = (struct osmo_cbsp_fail_ent){
+		.id_discr = CELL_IDENT_LAC_AND_CI,
+		.cell_id = {
+			.lac_and_ci = {
+				.lac = bts->location_area_code,
+				.ci = bts->cell_identity,
+			}
+		},
+		.cause = OSMO_CBSP_CAUSE_CELL_BROADCAST_NOT_OPERATIONAL
+	};
+	llist_add(&fail_ent.list, &cbsp->u.failure.fail_list);
+
+	return cbsp_tx_decoded(cbc, cbsp);
+}
+
 /* transmit a CBSP KEEPALIVE COMPLETE to the CBC */
 static int tx_cbsp_keepalive_compl(struct bsc_cbc_link *cbc)
 {
@@ -462,25 +535,34 @@ static int tx_cbsp_keepalive_compl(struct bsc_cbc_link *cbc)
  * Per-BTS Processing of CBSP from CBC, called via cbsp_per_bts()
  *********************************************************************************/
 
+static void etws_pn_stop(struct gsm_bts *bts, bool timeout)
+{
+	if (osmo_bts_has_feature(&bts->features, BTS_FEAT_ETWS_PN)) {
+		LOG_BTS(bts, DCBS, LOGL_NOTICE, "ETWS PN broadcast via PCH disabled (cause=%s)\n",
+			timeout ? "timeout" : "request");
+		rsl_etws_pn_command(bts, RSL_CHAN_PCH_AGCH, NULL, 0);
+	}
+	bts->etws.active = false;
+	if (!timeout)
+		osmo_timer_del(&bts->etws.timer);
+}
+
 /* timer call-back once ETWS warning period has expired */
 static void etws_pn_cb(void *data)
 {
 	struct gsm_bts *bts = (struct gsm_bts *)data;
-	LOG_BTS(bts, DCBS, LOGL_NOTICE, "ETWS PN Timeout; disabling broadcast via PCH\n");
-	rsl_etws_pn_command(bts, RSL_CHAN_PCH_AGCH, NULL, 0);
+	etws_pn_stop(bts, true);
 }
 
-static void etws_primary_to_bts(struct gsm_bts *bts, const struct osmo_cbsp_write_replace *wrepl)
+
+/* the actual "execution" part: Send ETWS to all active lchan in the BTS and via PCH */
+static void bts_send_etws(struct gsm_bts *bts)
 {
-	uint8_t etws_primary[ETWS_PRIM_NOTIF_SIZE];
+	struct bts_etws_state *bes = &bts->etws;
 	struct gsm_bts_trx *trx;
 	unsigned int count = 0;
 	int i, j;
 
-	gen_etws_primary_notification(etws_primary, wrepl->new_serial_nr, wrepl->msg_id,
-				      wrepl->u.emergency.warning_type,
-				      wrepl->u.emergency.warning_sec_info);
-
 	/* iterate over all lchan in each TS in each TRX of this BTS */
 	llist_for_each_entry(trx, &bts->trx_list, list) {
 		for (i = 0; i < ARRAY_SIZE(trx->ts); i++) {
@@ -489,8 +571,8 @@ static void etws_primary_to_bts(struct gsm_bts *bts, const struct osmo_cbsp_writ
 				struct gsm_lchan *lchan = &ts->lchan[j];
 				if (!lchan_may_receive_data(lchan))
 					continue;
-				gsm48_send_rr_app_info(lchan, 0x1, 0x0, etws_primary,
-							sizeof(etws_primary));
+				gsm48_send_rr_app_info(lchan, 0x1, 0x0, bes->primary,
+							sizeof(bes->primary));
 				count++;
 			}
 		}
@@ -501,18 +583,64 @@ static void etws_primary_to_bts(struct gsm_bts *bts, const struct osmo_cbsp_writ
 
 	/* Notify BTS of primary ETWS notification via vendor-specific Abis message */
 	if (osmo_bts_has_feature(&bts->features, BTS_FEAT_ETWS_PN)) {
-		rsl_etws_pn_command(bts, RSL_CHAN_PCH_AGCH, etws_primary, sizeof(etws_primary));
+		rsl_etws_pn_command(bts, RSL_CHAN_PCH_AGCH, bes->primary, sizeof(bes->primary));
 		LOG_BTS(bts, DCBS, LOGL_NOTICE, "Sent ETWS Primary Notification via common channel\n");
-		if (wrepl->u.emergency.warning_period != 0xffffffff) {
-			osmo_timer_setup(&bts->etws_timer, etws_pn_cb, bts);
-			osmo_timer_schedule(&bts->etws_timer, wrepl->u.emergency.warning_period, 0);
-		} else
-			LOG_BTS(bts, DCBS, LOGL_NOTICE, "Unlimited ETWS PN broadcast, this breaks "
-				"normal network operation due to PCH blockage\n");
 	} else
 		LOG_BTS(bts, DCBS, LOGL_ERROR, "BTS doesn't support RSL command for ETWS PN\n");
 }
 
+static int etws_primary_to_bts(struct gsm_bts *bts, const struct osmo_cbsp_write_replace *wrepl)
+{
+	struct bts_etws_state *bes = &bts->etws;
+
+	if (bes->active) {
+		/* we were already broadcasting emergency before receiving this WRITE-REPLACE */
+
+		/* If only the New Serial Number IE, and not the Old Serial Number IE, is included in the
+		 * WRITE-REPLACE message, then the BSC shall interpret the message as a write request, i.e. a
+		 * broadcast request of a new emergency message without replacing an ongoing emergency message
+		 * broadcast. */
+		if (!wrepl->old_serial_nr) {
+			/* If a write request is received for a cell where an emergency message broadcast is
+			 * currently ongoing, the write request is considered as failed */
+			LOG_BTS(bts, DCBS, LOGL_NOTICE, "Rx CBSP WRITE rejected due to ongoing emergency "
+				"while no Old Serial Nr IE present in CBSP WRITE\n");
+			return -CBSP_CAUSE_BSC_CAPACITY_EXCEEDED;
+		}
+
+		if (!etws_msg_id_matches(*wrepl->old_serial_nr, bes->input.serial_nr)) {
+			LOG_BTS(bts, DCBS, LOGL_NOTICE, "Rx CBSP WRITE-REPLACE old_serial 0x%04x doesn't match "
+				"current serial 0x%04x. Is the CBC confused?\n",
+				*wrepl->old_serial_nr, bes->input.serial_nr);
+			/* we allow the WRITE-REPLACE to continue, TS 48.049 doesn't specify how to
+			 * handle situations like this */
+		}
+	}
+
+	/* copy over all the data to per-BTS private state */
+	bes->input.msg_id = wrepl->msg_id;
+	bes->input.serial_nr = wrepl->new_serial_nr;
+	bes->input.warn_type = wrepl->u.emergency.warning_type;
+	memcpy(bes->input.sec_info, wrepl->u.emergency.warning_sec_info, sizeof(bes->input.sec_info));
+
+	/* generate the encoded ETWS PN */
+	gen_etws_primary_notification(bes->primary, bes->input.serial_nr, bes->input.msg_id,
+				      bes->input.warn_type, bes->input.sec_info);
+
+	bes->active = true;
+
+	bts_send_etws(bts);
+
+	/* start the expiration timer, if any */
+	if (wrepl->u.emergency.warning_period != 0xffffffff) {
+		osmo_timer_schedule(&bts->etws.timer, wrepl->u.emergency.warning_period, 0);
+	} else
+		LOG_BTS(bts, DCBS, LOGL_NOTICE, "Unlimited ETWS PN broadcast, this breaks "
+			"normal network operation due to PCH blockage\n");
+
+	return 0;
+}
+
 /*! Try to execute a write-replace operation; roll-back if it fails.
  *  \param[in] chan_state BTS CBCH channel state
  *  \param[in] extended_cbch Basic (false) or Extended (true) CBCH
@@ -582,8 +710,7 @@ static int bts_rx_write_replace(struct gsm_bts *bts, const struct osmo_cbsp_deco
 	int rc;
 
 	if (!wrepl->is_cbs) {
-		etws_primary_to_bts(bts, wrepl);
-		return 0;
+		return etws_primary_to_bts(bts, wrepl);
 	}
 
 	/* check if cell has a CBCH at all */
@@ -629,21 +756,53 @@ static int bts_rx_kill(struct gsm_bts *bts, const struct osmo_cbsp_decoded *dec,
 			struct response_state *r_state, void *priv)
 {
 	const struct osmo_cbsp_kill *kill = &dec->u.kill;
-	struct bts_smscb_chan_state *chan_state;
-	struct bts_smscb_message *smscb;
-	bool extended = false;
 
-	if (kill->channel_ind && *kill->channel_ind == 0x01)
-		extended = true;
-	chan_state = bts_get_smscb_chan(bts, extended);
+	if (kill->channel_ind) {
+		/* KILL for CBS message */
+		struct bts_smscb_chan_state *chan_state;
+		struct bts_smscb_message *smscb;
+		bool extended = false;
 
-	/* Find message by msg_id + old_serial_nr */
-	smscb = bts_find_smscb(chan_state, kill->msg_id, kill->old_serial_nr);
-	if (!smscb)
-		return -CBSP_CAUSE_MSG_REF_NOT_IDENTIFIED;
+		if (*kill->channel_ind == 0x01)
+			extended = true;
 
-	/* Remove it */
-	bts_smscb_del(smscb, chan_state, "KILL");
+		chan_state = bts_get_smscb_chan(bts, extended);
+
+		/* Find message by msg_id + old_serial_nr */
+		smscb = bts_find_smscb(chan_state, kill->msg_id, kill->old_serial_nr);
+		if (!smscb)
+			return -CBSP_CAUSE_MSG_REF_NOT_IDENTIFIED;
+
+		append_bcast_compl(r_state, chan_state->bts, smscb);
+
+		/* Remove it */
+		bts_smscb_del(smscb, chan_state, "KILL");
+	} else {
+		/* KILL for Emergency */
+		struct bts_etws_state *bes = &bts->etws;
+
+		if (!bes->active) {
+			LOG_BTS(bts, DCBS, LOGL_NOTICE, "Rx CBSP KILL (Emerg) but no emergency "
+				"broadcast is currently active in this cell\n");
+			return -CBSP_CAUSE_MSG_REF_NOT_IDENTIFIED;
+		}
+
+		if (kill->msg_id != bes->input.msg_id) {
+			LOG_BTS(bts, DCBS, LOGL_NOTICE, "Rx CBSP KILL (Emerg) for msg_id 0x%04x, but "
+				"current emergency msg_id is 0x%04x\n", kill->msg_id, bes->input.msg_id);
+			return -CBSP_CAUSE_MSG_REF_NOT_IDENTIFIED;
+		}
+
+		if (!etws_msg_id_matches(kill->old_serial_nr, bes->input.serial_nr)) {
+			LOG_BTS(bts, DCBS, LOGL_NOTICE, "Rx CBSP KILL (Emerg) for old_serial_nr 0x%04x, but "
+				"current emergency serial_nr is 0x%04x\n",
+				kill->old_serial_nr, bes->input.serial_nr);
+			return -CBSP_CAUSE_MSG_REF_NOT_IDENTIFIED;
+		}
+
+		/* stop broadcasting the PN in this BTS */
+		etws_pn_stop(bts, false);
+	}
 	return 0;
 }
 
@@ -653,6 +812,8 @@ static int bts_rx_reset(struct gsm_bts *bts, const struct osmo_cbsp_decoded *dec
 	struct bts_smscb_chan_state *chan_state;
 	struct bts_smscb_message *smscb, *smscb2;
 
+	LOG_BTS(bts, DCBS, LOGL_NOTICE, "Rx CBSP RESET: clearing all state; disabling broadcast\n");
+
 	/* remove all SMSCB from CBCH BASIC this BTS */
 	chan_state = bts_get_smscb_chan(bts, false);
 	llist_for_each_entry_safe(smscb, smscb2, &chan_state->messages, list)
@@ -663,9 +824,36 @@ static int bts_rx_reset(struct gsm_bts *bts, const struct osmo_cbsp_decoded *dec
 	llist_for_each_entry_safe(smscb, smscb2, &chan_state->messages, list)
 		bts_smscb_del(smscb, chan_state, "RESET");
 
+	osmo_timer_del(&bts->etws.timer);
+
+	/* Make sure that broadcast is disabled */
+	rsl_etws_pn_command(bts, RSL_CHAN_PCH_AGCH, NULL, 0);
 	return 0;
 }
 
+static int bts_rx_status_query(struct gsm_bts *bts, const struct osmo_cbsp_decoded *dec,
+			       struct response_state *r_state, void *priv)
+{
+	const struct osmo_cbsp_msg_status_query *query = &dec->u.msg_status_query;
+	struct bts_smscb_chan_state *chan_state;
+	struct bts_smscb_message *smscb;
+	bool extended = false;
+
+	if (query->channel_ind == 0x01)
+		extended = true;
+	chan_state = bts_get_smscb_chan(bts, extended);
+
+	/* Find message by msg_id + old_serial_nr */
+	smscb = bts_find_smscb(chan_state, query->msg_id, query->old_serial_nr);
+	if (!smscb)
+		return -CBSP_CAUSE_MSG_REF_NOT_IDENTIFIED;
+
+	append_bcast_compl(r_state, chan_state->bts, smscb);
+
+	return 0;
+}
+
+
 /*********************************************************************************
  * Receive of CBSP from CBC
  *********************************************************************************/
@@ -751,11 +939,16 @@ static int cbsp_rx_kill(struct bsc_cbc_link *cbc, const struct osmo_cbsp_decoded
 		fail->channel_ind = kill->channel_ind;
 		llist_replace_head(&fail->fail_list, &r_state->fail);
 
-		fail->cell_list.id_discr = r_state->success.id_discr;
-		llist_replace_head(&fail->cell_list.list, &r_state->success.list);
-
-		fail->num_compl_list.id_discr = r_state->num_completed.id_discr;
-		llist_replace_head(&fail->num_compl_list.list, &r_state->num_completed.list);
+		/* if the KILL relates to CBS, the "Channel Indicator" IE is present */
+		if (kill->channel_ind) {
+			/* only if it was CBS */
+			fail->num_compl_list.id_discr = r_state->num_completed.id_discr;
+			llist_replace_head(&fail->num_compl_list.list, &r_state->num_completed.list);
+		} else {
+			/* only if it was emergency */
+			fail->cell_list.id_discr = r_state->success.id_discr;
+			llist_replace_head(&fail->cell_list.list, &r_state->success.list);
+		}
 	} else {
 		resp = osmo_cbsp_decoded_alloc(cbc, CBSP_MSGT_KILL_COMPL);
 		struct osmo_cbsp_kill_complete *compl = &resp->u.kill_compl;
@@ -763,11 +956,16 @@ static int cbsp_rx_kill(struct bsc_cbc_link *cbc, const struct osmo_cbsp_decoded
 		compl->old_serial_nr = kill->old_serial_nr;
 		compl->channel_ind = kill->channel_ind;
 
-		compl->cell_list.id_discr = r_state->success.id_discr;
-		llist_replace_head(&compl->cell_list.list, &r_state->success.list);
-
-		compl->num_compl_list.id_discr = r_state->num_completed.id_discr;
-		llist_replace_head(&compl->num_compl_list.list, &r_state->num_completed.list);
+		/* if the KILL relates to CBS, the "Channel Indicator" IE is present */
+		if (kill->channel_ind) {
+			/* only if it was CBS */
+			compl->num_compl_list.id_discr = r_state->num_completed.id_discr;
+			llist_replace_head(&compl->num_compl_list.list, &r_state->num_completed.list);
+		} else {
+			/* only if it was emergency */
+			compl->cell_list.id_discr = r_state->success.id_discr;
+			llist_replace_head(&compl->cell_list.list, &r_state->success.list);
+		}
 	}
 
 	cbsp_tx_decoded(cbc, resp);
@@ -809,6 +1007,48 @@ static int cbsp_rx_reset(struct bsc_cbc_link *cbc, const struct osmo_cbsp_decode
 	return rc;
 }
 
+static int cbsp_rx_status_query(struct bsc_cbc_link *cbc, const struct osmo_cbsp_decoded *dec)
+{
+	const struct osmo_cbsp_msg_status_query *query = &dec->u.msg_status_query;
+	struct gsm_network *net = cbc->net;
+	struct response_state *r_state = talloc_zero(cbc, struct response_state);
+	struct osmo_cbsp_decoded *resp;
+	int rc;
+
+	LOGP(DCBS, LOGL_DEBUG, "CBSP Rx MESSAGE STATUS QUERY\n");
+
+	rc = cbsp_per_bts(net, r_state, &dec->u.msg_status_query.cell_list, bts_rx_status_query, dec, NULL);
+	if (rc < 0) {
+		resp = osmo_cbsp_decoded_alloc(cbc, CBSP_MSGT_MSG_STATUS_QUERY_FAIL);
+		struct osmo_cbsp_msg_status_query_failure *fail = &resp->u.msg_status_query_fail;
+		fail->msg_id = query->msg_id;
+		fail->old_serial_nr = query->old_serial_nr;
+		fail->channel_ind = query->channel_ind;
+		llist_replace_head(&fail->fail_list, &r_state->fail);
+
+		fail->num_compl_list.id_discr = r_state->num_completed.id_discr;
+		llist_replace_head(&fail->num_compl_list.list, &r_state->num_completed.list);
+	} else {
+		resp = osmo_cbsp_decoded_alloc(cbc, CBSP_MSGT_MSG_STATUS_QUERY_COMPL);
+		struct osmo_cbsp_msg_status_query_complete *compl = &resp->u.msg_status_query_compl;
+		compl->msg_id = query->msg_id;
+		compl->old_serial_nr = query->old_serial_nr;
+		compl->channel_ind = query->channel_ind;
+
+		if (dec->u.msg_status_query.cell_list.id_discr == CELL_IDENT_BSS) {
+			/* replace the list of individual cell identities with CELL_IDENT_BSS */
+			compl->num_compl_list.id_discr = CELL_IDENT_BSS;
+			/* no need to free num_completed_list entries, hierarchical talloc works */
+		} else {
+			compl->num_compl_list.id_discr = r_state->num_completed.id_discr;
+			llist_replace_head(&compl->num_compl_list.list, &r_state->num_completed.list);
+		}
+	}
+	cbsp_tx_decoded(cbc, resp);
+	talloc_free(r_state);
+	return rc;
+}
+
 
 /*! process an incoming, already decoded CBSP message from the CBC.
  *  \param[in] cbc link to the CBC
@@ -831,8 +1071,10 @@ int cbsp_rx_decoded(struct bsc_cbc_link *cbc, const struct osmo_cbsp_decoded *de
 	case CBSP_MSGT_RESET:		/* stop broadcasting of all messages */
 		rc = cbsp_rx_reset(cbc, dec);
 		break;
-	case CBSP_MSGT_LOAD_QUERY:
 	case CBSP_MSGT_MSG_STATUS_QUERY:
+		rc = cbsp_rx_status_query(cbc, dec);
+		break;
+	case CBSP_MSGT_LOAD_QUERY:
 	case CBSP_MSGT_SET_DRX:
 		LOGP(DCBS, LOGL_ERROR, "Received Unimplemented CBSP Message Type %s",
 			get_value_string(cbsp_msg_type_names, dec->msg_type));
@@ -846,54 +1088,74 @@ int cbsp_rx_decoded(struct bsc_cbc_link *cbc, const struct osmo_cbsp_decoded *de
 	return rc;
 }
 
-/*********************************************************************************
- * VTY Interface (Introspection)
- *********************************************************************************/
-
-static void vty_dump_smscb_chan_state(struct vty *vty, const struct bts_smscb_chan_state *cs)
+/* initialize the ETWS state of a BTS */
+void bts_etws_init(struct gsm_bts *bts)
 {
-	const struct bts_smscb_message *sm;
-
-	vty_out(vty, "%s CBCH:%s", cs == &cs->bts->cbch_basic ? "BASIC" : "EXTENDED", VTY_NEWLINE);
-
-	vty_out(vty, " MsgId | SerNo | Pg |      Category | Perd | #Tx  | #Req | DCS%s", VTY_NEWLINE);
-	vty_out(vty, "-------|-------|----|---------------|------|------|------|----%s", VTY_NEWLINE);
-	llist_for_each_entry(sm, &cs->messages, list) {
-		vty_out(vty, "  %04x |  %04x | %2u | %13s | %4u | %4u | %4u | %02x%s",
-			sm->input.msg_id, sm->input.serial_nr, sm->num_pages,
-			get_value_string(cbsp_category_names, sm->input.category),
-			sm->input.rep_period, sm->bcast_count, sm->input.num_bcast_req,
-			sm->input.dcs, VTY_NEWLINE);
-	}
-	vty_out(vty, "%s", VTY_NEWLINE);
+	bts->etws.active = false;
+	osmo_timer_setup(&bts->etws.timer, etws_pn_cb, bts);
 }
 
-DEFUN(bts_show_cbs, bts_show_cbs_cmd,
-	"show bts <0-255> smscb [(basic|extended)]",
-	SHOW_STR "Display information about a BTS\n" "BTS number\n"
-	"SMS Cell Broadcast State\n"
-	"Show only information related to CBCH BASIC\n"
-	"Show only information related to CBCH EXTENDED\n")
+/* BSC is bootstrapping a BTS; install any currently active ETWS PN */
+static void bts_etws_bootstrap(struct gsm_bts *bts)
 {
-	struct gsm_network *net = gsmnet_from_vty(vty);
-	int bts_nr = atoi(argv[0]);
+	if (bts->etws.active)
+		bts_send_etws(bts);
+}
+
+/* Callback function to be called every time we receive a signal from NM */
+static int nm_sig_cb(unsigned int subsys, unsigned int signal,
+		     void *handler_data, void *signal_data)
+{
+	struct nm_running_chg_signal_data *nsd;
 	struct gsm_bts *bts;
+	struct gsm_bts_trx *trx;
 
-	if (bts_nr >= net->num_bts) {
-		vty_out(vty, "%% can't find BTS '%s'%s", argv[0], VTY_NEWLINE);
-		return CMD_WARNING;
+	if (signal != S_NM_RUNNING_CHG)
+		return 0;
+
+	nsd = signal_data;
+	bts = nsd->bts;
+
+	switch (nsd->obj_class) {
+	case NM_OC_RADIO_CARRIER:
+		trx = (struct gsm_bts_trx *)nsd->obj;
+		break;
+	case NM_OC_BASEB_TRANSC:
+		trx = gsm_bts_bb_trx_get_trx((struct gsm_bts_bb_trx *)nsd->obj);
+		break;
+	default:
+		return 0;
 	}
-	bts = gsm_bts_num(net, bts_nr);
 
-	if (argc < 2 || !strcmp(argv[1], "basic"))
-		vty_dump_smscb_chan_state(vty, &bts->cbch_basic);
-	if (argc < 2 || !strcmp(argv[1], "extended"))
-		vty_dump_smscb_chan_state(vty, &bts->cbch_extended);
+	struct gsm_lchan *cbch = gsm_bts_get_cbch(bts);
+	if (!cbch)
+		return 0;
+	/* We only care about state changes of TRX holding the CBCH. */
+	if (trx != cbch->ts->trx)
+		return 0;
 
-	return CMD_SUCCESS;
+	if (nsd->running) {
+		if (trx_is_usable(trx)) {
+			LOG_BTS(bts, DCBS, LOGL_INFO, "BTS becomes available for CBCH\n");
+			/* Start CBCH transmit timer if CBCH is present */
+			bts_cbch_timer_schedule(trx->bts);
+			/* Start ETWS/PWS Primary Notification, if active */
+			bts_etws_bootstrap(trx->bts);
+			cbsp_tx_restart_bts(bts->network->cbc, false, bts);
+		}
+	} else {
+		if (osmo_timer_pending(&bts->cbch_timer)) {
+			/* If timer is ongoing it means CBCH was available */
+			LOG_BTS(bts, DCBS, LOGL_INFO, "BTS becomes unavailable for CBCH\n");
+			osmo_timer_del(&bts->cbch_timer);
+			cbsp_tx_failure_bts(bts->network->cbc, false, bts);
+		} /* else: CBCH was already unavailable before */
+	}
+	return 0;
 }
 
-void smscb_vty_init(void)
+/* To be called once at startup of the process: */
+void smscb_global_init(void)
 {
-	install_element_ve(&bts_show_cbs_cmd);
+	osmo_signal_register_handler(SS_NM, nm_sig_cb, NULL);
 }
