@@ -31,16 +31,17 @@
 #include <osmocom/core/utils.h>
 #include <osmocom/gsm/sysinfo.h>
 #include <osmocom/gsm/gsm48_ie.h>
+#include <osmocom/gsm/gsm48_rest_octets.h>
 #include <osmocom/gsm/gsm48.h>
+#include <osmocom/gsm/gsm48_arfcn_range_encode.h>
 
 #include <osmocom/bsc/debug.h>
 #include <osmocom/bsc/gsm_data.h>
 #include <osmocom/bsc/abis_rsl.h>
-#include <osmocom/bsc/rest_octets.h>
-#include <osmocom/bsc/arfcn_range_encode.h>
 #include <osmocom/bsc/gsm_04_08_rr.h>
-#include <osmocom/bsc/acc_ramp.h>
+#include <osmocom/bsc/acc.h>
 #include <osmocom/bsc/neighbor_ident.h>
+#include <osmocom/bsc/bts.h>
 
 struct gsm0808_cell_id_list2;
 
@@ -168,7 +169,12 @@ static int make_si2quaters(struct gsm_bts *bts, bool counting)
 			si2q->header.system_information = GSM48_MT_RR_SYSINFO_2quater;
 		}
 
-		rc = rest_octets_si2quater(si2q->rest_octets, bts);
+		rc = osmo_gsm48_rest_octets_si2quater_encode(si2q->rest_octets, bts->si2q_index,
+							     bts->si2q_count, bts->si_common.data.uarfcn_list,
+							     &bts->u_offset, bts->si_common.uarfcn_length,
+							     bts->si_common.data.scramble_list,
+							     &bts->si_common.si2quater_neigh_list,
+							     &bts->e_offset);
 		if (rc < 0)
 			return rc;
 
@@ -398,7 +404,7 @@ static int freq_list_bmrel_set_arfcn(uint8_t *chan_list, unsigned int arfcn)
 
 /* generate a variable bitmap */
 static inline int enc_freq_lst_var_bitmap(uint8_t *chan_list,
-				struct bitvec *bv, const struct gsm_bts *bts,
+				const struct bitvec *bv, const struct gsm_bts *bts,
 				bool bis, bool ter, int min, bool pgsm)
 {
 	int i;
@@ -425,7 +431,7 @@ static inline int enc_freq_lst_var_bitmap(uint8_t *chan_list,
 	return 0;
 }
 
-int range_encode(enum gsm48_range r, int *arfcns, int arfcns_used, int *w,
+int range_encode(enum osmo_gsm48_range r, int *arfcns, int arfcns_used, int *w,
 		 int f0, uint8_t *chan_list)
 {
 	/*
@@ -434,22 +440,22 @@ int range_encode(enum gsm48_range r, int *arfcns, int arfcns_used, int *w,
 	 */
 	int rc, f0_included;
 
-	range_enc_filter_arfcns(arfcns, arfcns_used, f0, &f0_included);
+	osmo_gsm48_range_enc_filter_arfcns(arfcns, arfcns_used, f0, &f0_included);
 
-	rc = range_enc_arfcns(r, arfcns, arfcns_used, w, 0);
+	rc = osmo_gsm48_range_enc_arfcns(r, arfcns, arfcns_used, w, 0);
 	if (rc < 0)
 		return rc;
 
 	/* Select the range and the amount of bits needed */
 	switch (r) {
-	case ARFCN_RANGE_128:
-		return range_enc_range128(chan_list, f0, w);
-	case ARFCN_RANGE_256:
-		return range_enc_range256(chan_list, f0, w);
-	case ARFCN_RANGE_512:
-		return range_enc_range512(chan_list, f0, w);
-	case ARFCN_RANGE_1024:
-		return range_enc_range1024(chan_list, f0, f0_included, w);
+	case OSMO_GSM48_ARFCN_RANGE_128:
+		return osmo_gsm48_range_enc_128(chan_list, f0, w);
+	case OSMO_GSM48_ARFCN_RANGE_256:
+		return osmo_gsm48_range_enc_256(chan_list, f0, w);
+	case OSMO_GSM48_ARFCN_RANGE_512:
+		return osmo_gsm48_range_enc_512(chan_list, f0, w);
+	case OSMO_GSM48_ARFCN_RANGE_1024:
+		return osmo_gsm48_range_enc_1024(chan_list, f0, f0_included, w);
 	default:
 		return -ERANGE;
 	};
@@ -459,11 +465,11 @@ int range_encode(enum gsm48_range r, int *arfcns, int arfcns_used, int *w,
 
 /* generate a frequency list with the range 512 format */
 static inline int enc_freq_lst_range(uint8_t *chan_list,
-				struct bitvec *bv, const struct gsm_bts *bts,
+				const struct bitvec *bv, const struct gsm_bts *bts,
 				bool bis, bool ter, bool pgsm)
 {
-	int arfcns[RANGE_ENC_MAX_ARFCNS];
-	int w[RANGE_ENC_MAX_ARFCNS];
+	int arfcns[OSMO_GSM48_RANGE_ENC_MAX_ARFCNS];
+	int w[OSMO_GSM48_RANGE_ENC_MAX_ARFCNS];
 	int arfcns_used = 0;
 	int i, range, f0;
 
@@ -482,8 +488,8 @@ static inline int enc_freq_lst_range(uint8_t *chan_list,
 	/*
 	 * Check if the given list of ARFCNs can be encoded.
 	 */
-	range = range_enc_determine_range(arfcns, arfcns_used, &f0);
-	if (range == ARFCN_RANGE_INVALID)
+	range = osmo_gsm48_range_enc_determine_range(arfcns, arfcns_used, &f0);
+	if (range == OSMO_GSM48_ARFCN_RANGE_INVALID)
 		return -2;
 
 	memset(w, 0, sizeof(w));
@@ -491,27 +497,36 @@ static inline int enc_freq_lst_range(uint8_t *chan_list,
 }
 
 /* generate a cell channel list as per Section 10.5.2.1b of 04.08 */
-static int bitvec2freq_list(uint8_t *chan_list, struct bitvec *bv,
+static int bitvec2freq_list(uint8_t *chan_list, const struct bitvec *bv,
 			    const struct gsm_bts *bts, bool bis, bool ter)
 {
 	int i, rc, min = -1, max = -1, arfcns = 0;
 	bool pgsm = false;
 	memset(chan_list, 0, 16);
 
-	if (bts->band == GSM_BAND_900
-	 && bts->c0->arfcn >= 1 && bts->c0->arfcn <= 124)
+	/* According to 3GPP TS 44.018, section 10.5.2.1b.2, only ARFCN values
+	 * in range 1..124 can be encoded using the 'bit map 0' format. */
+	if (bts->band == GSM_BAND_900)
 		pgsm = true;
+	/* Check presence of E-GSM ARFCN 0 */
+	if (pgsm && bitvec_get_bit_pos(bv, 0) == ONE)
+		pgsm = false;
+	/* Check presence of E-GSM ARFCNs 975..1023 */
+	for (i = 975; pgsm && i <= 1023; i++) {
+		if (bitvec_get_bit_pos(bv, i) == ONE)
+			pgsm = false;
+	}
+
 	/* P-GSM-only handsets only support 'bit map 0 format' */
 	if (!bis && !ter && pgsm) {
 		chan_list[0] = 0;
 
-		for (i = 0; i < bv->data_len*8; i++) {
-			if (i >= 1 && i <= 124
-			 && bitvec_get_bit_pos(bv, i)) {
-				rc = freq_list_bm0_set_arfcn(chan_list, i);
-				if (rc < 0)
-					return rc;
-			}
+		for (i = 1; i <= 124; i++) {
+			if (!bitvec_get_bit_pos(bv, i))
+				continue;
+			rc = freq_list_bm0_set_arfcn(chan_list, i);
+			if (rc < 0)
+				return rc;
 		}
 		return 0;
 	}
@@ -547,7 +562,7 @@ static int bitvec2freq_list(uint8_t *chan_list, struct bitvec *bv,
 			max = i;
 	}
 
-	if (max == -1) {
+	if (arfcns == 0) {
 		/* Empty set, use 'bit map 0 format' */
 		chan_list[0] = 0;
 		return 0;
@@ -568,51 +583,63 @@ static int bitvec2freq_list(uint8_t *chan_list, struct bitvec *bv,
 	return -EINVAL;
 }
 
-/* generate a cell channel list as per Section 10.5.2.1b of 04.08 */
-/* static*/ int generate_cell_chan_list(uint8_t *chan_list, struct gsm_bts *bts)
+/* (Re)generate Cell Allocation (basically a bit-vector of all cell channels) */
+int generate_cell_chan_alloc(struct gsm_bts *bts)
 {
-	struct gsm_bts_trx *trx;
-	struct bitvec *bv = &bts->si_common.cell_alloc;
+	const struct gsm_bts_trx *trx;
+	unsigned int chan, chan_num;
+	unsigned int tn;
 
-	/* Zero-initialize the bit-vector */
-	memset(bv->data, 0, bv->data_len);
+	/* Temporary Cell Allocation bit-vector */
+	uint8_t ca[1024 / 8] = { 0 };
+	struct bitvec bv = {
+		.data_len = sizeof(ca),
+		.data = &ca[0],
+	};
 
-	/* first we generate a bitvec of all TRX ARFCN's in our BTS */
+	/* Calculate a bit-mask of all assigned ARFCNs */
 	llist_for_each_entry(trx, &bts->trx_list, list) {
-		unsigned int i, j;
 		/* Always add the TRX's ARFCN */
-		bitvec_set_bit_pos(bv, trx->arfcn, 1);
-		for (i = 0; i < ARRAY_SIZE(trx->ts); i++) {
-			struct gsm_bts_trx_ts *ts = &trx->ts[i];
+		bitvec_set_bit_pos(&bv, trx->arfcn, 1);
+		for (tn = 0; tn < ARRAY_SIZE(trx->ts); tn++) {
+			const struct gsm_bts_trx_ts *ts = &trx->ts[tn];
 			/* Add any ARFCNs present in hopping channels */
-			for (j = 0; j < 1024; j++) {
-				if (bitvec_get_bit_pos(&ts->hopping.arfcns, j))
-					bitvec_set_bit_pos(bv, j, 1);
+			for (chan = 0; chan < sizeof(ca) * 8; chan++) {
+				if (bitvec_get_bit_pos(&ts->hopping.arfcns, chan) == ONE)
+					bitvec_set_bit_pos(&bv, chan, ONE);
 			}
 		}
 	}
 
-	/* then we generate a GSM 04.08 frequency list from the bitvec */
-	return bitvec2freq_list(chan_list, bv, bts, false, false);
+	/* Calculate the overall number of assigned ARFCNs */
+	for (chan_num = 0, chan = 0; chan < sizeof(ca) * 8; chan++) {
+		if (bitvec_get_bit_pos(&bv, chan) == ONE)
+			chan_num++;
+	}
+
+	/* The Mobile Allocation IE may contain up to 64 bits, so here we
+	 * cannot allow more than 64 channels in Cell Allocation. */
+	if (chan_num > 64) {
+		LOGP(DRR, LOGL_ERROR, "Failed to (re)generate Cell Allocation: "
+		     "number of channels (%u) exceeds the maximum number of "
+		     "ARFCNs in Mobile Allocation (64)\n", chan_num);
+		return -E2BIG;
+	}
+
+	/* Commit the new Channel Allocation */
+	memcpy(&bts->si_common.data.cell_alloc[0], ca, sizeof(ca));
+	bts->si_common.cell_chan_num = chan_num;
+
+	return 0;
 }
 
-struct generate_bcch_chan_list__ni_iter_data {
-	struct gsm_bts *bts;
-	struct bitvec *bv;
-};
-
-static bool generate_bcch_chan_list__ni_iter_cb(const struct neighbor_ident_key *key,
-						const struct gsm0808_cell_id_list2 *val,
-						void *cb_data)
+/* generate a cell channel list as per Section 10.5.2.1b of 04.08 */
+int generate_cell_chan_list(uint8_t *chan_list, struct gsm_bts *bts)
 {
-	struct generate_bcch_chan_list__ni_iter_data *data = cb_data;
+	const struct bitvec *bv = &bts->si_common.cell_alloc;
 
-	if (key->from_bts != NEIGHBOR_IDENT_KEY_ANY_BTS
-	    && key->from_bts != data->bts->nr)
-		return true;
-
-	bitvec_set_bit_pos(data->bv, key->arfcn, 1);
-	return true;
+	/* generate a Frequency List from the Cell Allocation */
+	return bitvec2freq_list(chan_list, bv, bts, false, false);
 }
 
 /*! generate a cell channel list as per Section 10.5.2.22 of 04.08
@@ -640,7 +667,7 @@ static int generate_bcch_chan_list(uint8_t *chan_list, struct gsm_bts *bts,
 		/* Zero-initialize the bit-vector */
 		memset(bv->data, 0, bv->data_len);
 
-		if (llist_empty(&bts->local_neighbors)) {
+		if (llist_empty(&bts->neighbors)) {
 			/* There are no explicit neighbors, assume all BTS are. */
 			llist_for_each_entry(cur_bts, &bts->network->bts_list, list) {
 				if (cur_bts == bts)
@@ -649,21 +676,23 @@ static int generate_bcch_chan_list(uint8_t *chan_list, struct gsm_bts *bts,
 			}
 		} else {
 			/* Only add explicit neighbor cells */
-			struct gsm_bts_ref *neigh;
-			llist_for_each_entry(neigh, &bts->local_neighbors, entry) {
-				bitvec_set_bit_pos(bv, neigh->bts->c0->arfcn, 1);
+			struct neighbor *n;
+			llist_for_each_entry(n, &bts->neighbors, entry) {
+				if (n->type == NEIGHBOR_TYPE_CELL_ID && n->cell_id.ab_present) {
+					bitvec_set_bit_pos(bv, n->cell_id.ab.arfcn, 1);
+				} else {
+					struct gsm_bts *neigh_bts;
+					if (resolve_local_neighbor(&neigh_bts, bts, n)) {
+						LOGP(DHO, LOGL_ERROR,
+						     "Neither local nor remote neighbor: BTS %u -> %s\n",
+						     bts->nr, neighbor_to_str_c(OTC_SELECT, n));
+						continue;
+					}
+					if (neigh_bts->c0)
+						bitvec_set_bit_pos(bv, neigh_bts->c0->arfcn, 1);
+				}
 			}
 		}
-
-		/* Also add neighboring BSS cells' ARFCNs */
-		{
-			struct generate_bcch_chan_list__ni_iter_data data = {
-				.bv = bv,
-				.bts = bts,
-			};
-			neighbor_ident_iter(bts->network->neighbor_bss_cells,
-					    generate_bcch_chan_list__ni_iter_cb, &data);
-		}
 	}
 
 	/* then we generate a GSM 04.08 frequency list from the bitvec */
@@ -689,7 +718,7 @@ static int list_arfcn(uint8_t *chan_list, uint8_t mask, char *text)
 	struct gsm_sysinfo_freq freq[1024];
 
 	memset(freq, 0, sizeof(freq));
-	gsm48_decode_freq_list(freq, chan_list, 16, 0xce, 1);
+	gsm48_decode_freq_list(freq, chan_list, 16, mask, 1);
 	for (i = 0; i < 1024; i++) {
 		if (freq[i].mask) {
 			if (!n)
@@ -722,14 +751,13 @@ static int generate_si1(enum osmo_sysinfo_type t, struct gsm_bts *bts)
 	list_arfcn(si1->cell_channel_description, 0xce, "Serving cell:");
 
 	si1->rach_control = bts->si_common.rach_control;
-	if (acc_ramp_is_enabled(&bts->acc_ramp))
-		acc_ramp_apply(&si1->rach_control, &bts->acc_ramp);
+	acc_mgr_apply_acc(&bts->acc_mgr, &si1->rach_control);
 
 	/*
 	 * SI1 Rest Octets (10.5.2.32), contains NCH position and band
 	 * indicator but that is not in the 04.08.
 	 */
-	rc = rest_octets_si1(si1->rest_octets, NULL, is_dcs_net(bts));
+	rc = osmo_gsm48_rest_octets_si1_encode(si1->rest_octets, NULL, is_dcs_net(bts));
 
 	return sizeof(*si1) + rc;
 }
@@ -754,12 +782,25 @@ static int generate_si2(enum osmo_sysinfo_type t, struct gsm_bts *bts)
 
 	si2->ncc_permitted = bts->si_common.ncc_permitted;
 	si2->rach_control = bts->si_common.rach_control;
-	if (acc_ramp_is_enabled(&bts->acc_ramp))
-		acc_ramp_apply(&si2->rach_control, &bts->acc_ramp);
+	acc_mgr_apply_acc(&bts->acc_mgr, &si2->rach_control);
 
 	return sizeof(*si2);
 }
 
+/* Generate SI2bis Rest Octests 3GPP TS 44.018 Table 10.5.2.33.1 */
+static int rest_octets_si2bis(uint8_t *data)
+{
+	struct bitvec bv;
+
+	memset(&bv, 0, sizeof(bv));
+	bv.data = data;
+	bv.data_len = 1;
+
+	bitvec_spare_padding(&bv, (bv.data_len * 8) - 1);
+
+	return bv.data_len;
+}
+
 static int generate_si2bis(enum osmo_sysinfo_type t, struct gsm_bts *bts)
 {
 	int rc;
@@ -789,8 +830,7 @@ static int generate_si2bis(enum osmo_sysinfo_type t, struct gsm_bts *bts)
 		bts->si_valid &= ~(1 << SYSINFO_TYPE_2bis);
 
 	si2b->rach_control = bts->si_common.rach_control;
-	if (acc_ramp_is_enabled(&bts->acc_ramp))
-		acc_ramp_apply(&si2b->rach_control, &bts->acc_ramp);
+	acc_mgr_apply_acc(&bts->acc_mgr, &si2b->rach_control);
 
 	/* SI2bis Rest Octets as per 3GPP TS 44.018 §10.5.2.33 */
 	rc = rest_octets_si2bis(si2b->rest_octets);
@@ -798,6 +838,24 @@ static int generate_si2bis(enum osmo_sysinfo_type t, struct gsm_bts *bts)
 	return sizeof(*si2b) + rc;
 }
 
+
+/* Generate SI2ter Rest Octests 3GPP TS 44.018 Table 10.5.2.33a.1 */
+static int rest_octets_si2ter(uint8_t *data)
+{
+	struct bitvec bv;
+
+	memset(&bv, 0, sizeof(bv));
+	bv.data = data;
+	bv.data_len = 4;
+
+	/* No SI2ter_MP_CHANGE_MARK */
+	bitvec_set_bit(&bv, L);
+
+	bitvec_spare_padding(&bv, (bv.data_len * 8) - 1);
+
+	return bv.data_len;
+}
+
 static int generate_si2ter(enum osmo_sysinfo_type t, struct gsm_bts *bts)
 {
 	int rc;
@@ -867,7 +925,7 @@ static int generate_si2quater(enum osmo_sysinfo_type t, struct gsm_bts *bts)
 	return sizeof(*si2q) + rc;
 }
 
-static struct gsm48_si_ro_info si_info = {
+static struct osmo_gsm48_si_ro_info si_info = {
 	.selection_params = {
 		.present = 0,
 	},
@@ -905,14 +963,17 @@ static int generate_si3(enum osmo_sysinfo_type t, struct gsm_bts *bts)
 	si3->header.skip_indicator = 0;
 	si3->header.system_information = GSM48_MT_RR_SYSINFO_3;
 
+	/* The value in bts->si_common.chan_desc may get out of sync with the actual value
+	 * in net->T_defs (e.g. after changing it via the VTY), so we need to sync it here. */
+	bts->si_common.chan_desc.t3212 = osmo_tdef_get(bts->network->T_defs, 3212, OSMO_TDEF_CUSTOM, 0);
+
 	si3->cell_identity = htons(bts->cell_identity);
 	gsm48_generate_lai2(&si3->lai, bts_lai(bts));
 	si3->control_channel_desc = bts->si_common.chan_desc;
 	si3->cell_options = bts->si_common.cell_options;
 	si3->cell_sel_par = bts->si_common.cell_sel_par;
 	si3->rach_control = bts->si_common.rach_control;
-	if (acc_ramp_is_enabled(&bts->acc_ramp))
-		acc_ramp_apply(&si3->rach_control, &bts->acc_ramp);
+	acc_mgr_apply_acc(&bts->acc_mgr, &si3->rach_control);
 
 	/* allow/disallow DTXu */
 	gsm48_set_dtx(&si3->cell_options, bts->dtxu, bts->dtxu, true);
@@ -937,7 +998,7 @@ static int generate_si3(enum osmo_sysinfo_type t, struct gsm_bts *bts)
 		CBQ, CELL_RESELECT_OFFSET, TEMPORARY_OFFSET, PENALTY_TIME
 		Power Offset, 2ter Indicator, Early Classmark Sending,
 		Scheduling if and WHERE, GPRS Indicator, SI13 position */
-	rc = rest_octets_si3(si3->rest_octets, &si_info);
+	rc = osmo_gsm48_rest_octets_si3_encode(si3->rest_octets, &si_info);
 
 	return sizeof(*si3) + rc;
 }
@@ -947,7 +1008,7 @@ static int generate_si4(enum osmo_sysinfo_type t, struct gsm_bts *bts)
 	int rc;
 	struct gsm48_system_information_type_4 *si4 = (struct gsm48_system_information_type_4 *) GSM_BTS_SI(bts, t);
 	struct gsm_lchan *cbch_lchan;
-	uint8_t *restoct = si4->data;
+	uint8_t *tail = si4->data;
 
 	/* length of all IEs present except SI4 rest octets and l2_plen */
 	int l2_plen = sizeof(*si4) - 1;
@@ -961,19 +1022,33 @@ static int generate_si4(enum osmo_sysinfo_type t, struct gsm_bts *bts)
 	gsm48_generate_lai2(&si4->lai, bts_lai(bts));
 	si4->cell_sel_par = bts->si_common.cell_sel_par;
 	si4->rach_control = bts->si_common.rach_control;
-	if (acc_ramp_is_enabled(&bts->acc_ramp))
-		acc_ramp_apply(&si4->rach_control, &bts->acc_ramp);
+	acc_mgr_apply_acc(&bts->acc_mgr, &si4->rach_control);
 
 	/* Optional: CBCH Channel Description + CBCH Mobile Allocation */
 	cbch_lchan = gsm_bts_get_cbch(bts);
 	if (cbch_lchan) {
+		const struct gsm_bts_trx_ts *ts = cbch_lchan->ts;
 		struct gsm48_chan_desc cd;
-		gsm48_lchan2chan_desc_as_configured(&cd, cbch_lchan);
-		tv_fixed_put(si4->data, GSM48_IE_CBCH_CHAN_DESC, 3,
-			     (uint8_t *) &cd);
-		l2_plen += 3 + 1;
-		restoct += 3 + 1;
-		/* we don't use hopping and thus don't need a CBCH MA */
+
+		/* 10.5.2.5 (TV) CBCH Channel Description IE.
+		 * CBCH is never in VAMOS mode, so just pass allow_osmo_cbits == false. */
+		if (gsm48_lchan_and_pchan2chan_desc(&cd, cbch_lchan, cbch_lchan->ts->pchan_from_config,
+						    gsm_ts_tsc(cbch_lchan->ts), false))
+			return -EINVAL;
+		tail = tv_fixed_put(tail, GSM48_IE_CBCH_CHAN_DESC,
+				    sizeof(cd), (uint8_t *) &cd);
+		l2_plen += 1 + sizeof(cd);
+
+		/* 10.5.2.21 (TLV) CBCH Mobile Allocation IE */
+		if (ts->hopping.enabled) {
+			/* Prevent potential buffer overflow */
+			if (ts->hopping.ma_len > 2)
+				return -ENOMEM;
+			tail = tlv_put(tail, GSM48_IE_CBCH_MOB_AL,
+				       ts->hopping.ma_len,
+				       ts->hopping.ma_data);
+			l2_plen += 2 + ts->hopping.ma_len;
+		}
 	}
 
 	si4->header.l2_plen = GSM48_LEN2PLEN(l2_plen);
@@ -981,7 +1056,7 @@ static int generate_si4(enum osmo_sysinfo_type t, struct gsm_bts *bts)
 	/* SI4 Rest Octets (10.5.2.35), containing
 		Optional Power offset, GPRS Indicator,
 		Cell Identity, LSA ID, Selection Parameter */
-	rc = rest_octets_si4(restoct, &si_info, (uint8_t *)GSM_BTS_SI(bts, t) + GSM_MACBLOCK_LEN - restoct);
+	rc = osmo_gsm48_rest_octets_si4_encode(tail, &si_info, (uint8_t *)GSM_BTS_SI(bts, t) + GSM_MACBLOCK_LEN - tail);
 
 	return l2_plen + 1 + rc;
 }
@@ -995,14 +1070,9 @@ static int generate_si5(enum osmo_sysinfo_type t, struct gsm_bts *bts)
 	memset(output, GSM_MACBLOCK_PADDING, GSM_MACBLOCK_LEN);
 
 	/* ip.access nanoBTS needs l2_plen!! */
-	switch (bts->type) {
-	case GSM_BTS_TYPE_NANOBTS:
-	case GSM_BTS_TYPE_OSMOBTS:
+	if (is_ipaccess_bts(bts)) {
 		*output++ = GSM48_LEN2PLEN(l2_plen);
 		l2_plen++;
-		break;
-	default:
-		break;
 	}
 
 	si5 = (struct gsm48_system_information_type_5 *) output;
@@ -1031,14 +1101,9 @@ static int generate_si5bis(enum osmo_sysinfo_type t, struct gsm_bts *bts)
 	memset(output, GSM_MACBLOCK_PADDING, GSM_MACBLOCK_LEN);
 
 	/* ip.access nanoBTS needs l2_plen!! */
-	switch (bts->type) {
-	case GSM_BTS_TYPE_NANOBTS:
-	case GSM_BTS_TYPE_OSMOBTS:
+	if (is_ipaccess_bts(bts)) {
 		*output++ = GSM48_LEN2PLEN(l2_plen);
 		l2_plen++;
-		break;
-	default:
-		break;
 	}
 
 	si5b = (struct gsm48_system_information_type_5bis *) output;
@@ -1075,14 +1140,9 @@ static int generate_si5ter(enum osmo_sysinfo_type t, struct gsm_bts *bts)
 	memset(output, GSM_MACBLOCK_PADDING, GSM_MACBLOCK_LEN);
 
 	/* ip.access nanoBTS needs l2_plen!! */
-	switch (bts->type) {
-	case GSM_BTS_TYPE_NANOBTS:
-	case GSM_BTS_TYPE_OSMOBTS:
+	if (is_ipaccess_bts(bts)) {
 		*output++ = GSM48_LEN2PLEN(l2_plen);
 		l2_plen++;
-		break;
-	default:
-		break;
 	}
 
 	si5t = (struct gsm48_system_information_type_5ter *) output;
@@ -1106,21 +1166,18 @@ static int generate_si5ter(enum osmo_sysinfo_type t, struct gsm_bts *bts)
 static int generate_si6(enum osmo_sysinfo_type t, struct gsm_bts *bts)
 {
 	struct gsm48_system_information_type_6 *si6;
+	struct osmo_gsm48_si6_ro_info si6_ro_info;
 	uint8_t *output = GSM_BTS_SI(bts, t);
 	int l2_plen = 11;
 	int rc;
 
 	memset(output, GSM_MACBLOCK_PADDING, GSM_MACBLOCK_LEN);
+	memset(&si6_ro_info, 0, sizeof(si6_ro_info));
 
 	/* ip.access nanoBTS needs l2_plen!! */
-	switch (bts->type) {
-	case GSM_BTS_TYPE_NANOBTS:
-	case GSM_BTS_TYPE_OSMOBTS:
+	if (is_ipaccess_bts(bts)) {
 		*output++ = GSM48_LEN2PLEN(l2_plen);
 		l2_plen++;
-		break;
-	default:
-		break;
 	}
 
 	si6 = (struct gsm48_system_information_type_6 *) output;
@@ -1137,50 +1194,17 @@ static int generate_si6(enum osmo_sysinfo_type t, struct gsm_bts *bts)
 	gsm48_set_dtx(&si6->cell_options, bts->dtxu, bts->dtxu, false);
 
 	/* SI6 Rest Octets: 10.5.2.35a: PCH / NCH info, VBS/VGCS options */
-	rc = rest_octets_si6(si6->rest_octets, is_dcs_net(bts));
+	si6_ro_info.band_indicator_1900 = !is_dcs_net(bts);
+	rc = osmo_gsm48_rest_octets_si6_encode(si6->rest_octets, &si6_ro_info);
 
 	return l2_plen + rc;
 }
 
-static struct gsm48_si13_info si13_default = {
-	.cell_opts = {
-		.nmo 		= GPRS_NMO_II,
-		.t3168		= 2000,
-		.t3192		= 1500,
-		.drx_timer_max	= 3,
-		.bs_cv_max	= 15,
-		.ctrl_ack_type_use_block = true,
-		.ext_info_present = 0,
-		.supports_egprs_11bit_rach = 0,
-		.ext_info = {
-			/* The values below are just guesses ! */
-			.egprs_supported = 0,
-			.use_egprs_p_ch_req = 1,
-			.bep_period = 5,
-			.pfc_supported = 0,
-			.dtm_supported = 0,
-			.bss_paging_coordination = 0,
-		},
-	},
-	.pwr_ctrl_pars = {
-		.alpha		= 0,	/* a = 0.0 */
-		.t_avg_w	= 16,
-		.t_avg_t	= 16,
-		.pc_meas_chan	= 0, 	/* downling measured on CCCH */
-		.n_avg_i	= 8,
-	},
-	.bcch_change_mark	= 1,
-	.si_change_field	= 0,
-	.rac		= 0,	/* needs to be patched */
-	.spgc_ccch_sup 	= 0,
-	.net_ctrl_ord	= 0,
-	.prio_acc_thr	= 6,
-};
-
 static int generate_si13(enum osmo_sysinfo_type t, struct gsm_bts *bts)
 {
 	struct gsm48_system_information_type_13 *si13 =
 		(struct gsm48_system_information_type_13 *) GSM_BTS_SI(bts, t);
+	struct osmo_gsm48_si13_info si13_info;
 	int ret;
 
 	memset(si13, GSM_MACBLOCK_PADDING, GSM_MACBLOCK_LEN);
@@ -1189,18 +1213,66 @@ static int generate_si13(enum osmo_sysinfo_type t, struct gsm_bts *bts)
 	si13->header.skip_indicator = 0;
 	si13->header.system_information = GSM48_MT_RR_SYSINFO_13;
 
-	si13_default.rac = bts->gprs.rac;
-	si13_default.net_ctrl_ord = bts->gprs.net_ctrl_ord;
+	si13_info = (struct osmo_gsm48_si13_info){
+		.cell_opts = {
+			.nmo		= GPRS_NMO_II,
+			.t3168		= 2000,
+			.t3192		= 1500,
+			.drx_timer_max	= 3,
+			.bs_cv_max	= 15,
+			.ctrl_ack_type_use_block = bts->gprs.ctrl_ack_type_use_block,
+			.ext_info_present = true,
+			.ext_info = {
+				.egprs_supported = 0,		/* overridden below */
+				.use_egprs_p_ch_req = 0,	/* overridden below */
+				.bep_period = 5,
+				.pfc_supported = 0,
+				.dtm_supported = 0,
+				.bss_paging_coordination = 0,	/* overridden below */
+				.ccn_active = false,		/* overridden below */
+			},
+		},
+		.pwr_ctrl_pars = {
+			.alpha		= bts->gprs.pwr_ctrl.alpha,	/* a = 0.0 */
+			.t_avg_w	= 16,
+			.t_avg_t	= 16,
+			.pc_meas_chan	= 0,	/* downling measured on CCCH */
+			.n_avg_i	= 8,
+		},
+		.bcch_change_mark	= bts->bcch_change_mark, /* Information about the other SIs */
+		.si_change_field	= 0,
+		.rac		= bts->gprs.rac,
+		.spgc_ccch_sup	= 0,
+		.net_ctrl_ord	= bts->gprs.net_ctrl_ord,
+		.prio_acc_thr	= 6,
+	};
 
-	si13_default.cell_opts.ctrl_ack_type_use_block =
-		bts->gprs.ctrl_ack_type_use_block;
+	switch (bts->gprs.mode) {
+	case BTS_GPRS_EGPRS:
+		si13_info.cell_opts.ext_info.egprs_supported = 1;
+		/* Whether EGPRS capable MSs shall use EGPRS PACKET CHANNEL REQUEST */
+		if (bts->gprs.egprs_pkt_chan_request)
+			si13_info.cell_opts.ext_info.use_egprs_p_ch_req = 1;
+		else
+			si13_info.cell_opts.ext_info.use_egprs_p_ch_req = 0;
+		break;
+	case BTS_GPRS_GPRS:
+	case BTS_GPRS_NONE:
+		si13_info.cell_opts.ext_info.egprs_supported = 0;
+		si13_info.cell_opts.ext_info.use_egprs_p_ch_req = 0;
+		break;
+	}
 
-	/* Information about the other SIs */
-	si13_default.bcch_change_mark = bts->bcch_change_mark;
-	si13_default.cell_opts.supports_egprs_11bit_rach =
-					bts->gprs.supports_egprs_11bit_rach;
+	if (osmo_bts_has_feature(&bts->features, BTS_FEAT_PAGING_COORDINATION))
+		si13_info.cell_opts.ext_info.bss_paging_coordination = 1;
+	else
+		si13_info.cell_opts.ext_info.bss_paging_coordination = 0;
 
-	ret = rest_octets_si13(si13->rest_octets, &si13_default);
+	si13_info.cell_opts.ext_info.ccn_active = bts->gprs.ccn.forced_vty ?
+						  bts->gprs.ccn.active :
+						  osmo_bts_has_feature(&bts->features, BTS_FEAT_CCN);
+
+	ret = osmo_gsm48_rest_octets_si13_encode(si13->rest_octets, &si13_info);
 	if (ret < 0)
 		return ret;
 
@@ -1234,9 +1306,6 @@ int gsm_generate_si(struct gsm_bts *bts, enum osmo_sysinfo_type si_type)
 
 	switch (bts->gprs.mode) {
 	case BTS_GPRS_EGPRS:
-		si13_default.cell_opts.ext_info_present = 1;
-		si13_default.cell_opts.ext_info.egprs_supported = 1;
-		/* fallthrough */
 	case BTS_GPRS_GPRS:
 		si_info.gprs_ind.present = 1;
 		break;
@@ -1247,7 +1316,7 @@ int gsm_generate_si(struct gsm_bts *bts, enum osmo_sysinfo_type si_type)
 
 	memcpy(&si_info.selection_params,
 	       &bts->si_common.cell_ro_sel_par,
-	       sizeof(struct gsm48_si_selection_params));
+	       sizeof(struct osmo_gsm48_si_selection_params));
 
 	gen_si = gen_si_fn[si_type];
 	if (!gen_si) {
