@@ -33,9 +33,7 @@
 
 static void clock_debug(char* str)
 {
-	struct timespec ts;
 	struct timeval tv;
-	osmo_clock_gettime(CLOCK_MONOTONIC, &ts);
 	osmo_gettimeofday(&tv, NULL);
 	fprintf(stderr, "sys={%lu.%06lu}: %s\n",
 		tv.tv_sec, tv.tv_usec, str);
@@ -60,8 +58,6 @@ static inline struct gsm_bts *_bts_init(struct gsm_network *net, const char *msg
 #define bts_del(bts) _bts_del(bts, __func__)
 static inline void _bts_del(struct gsm_bts *bts, const char *msg)
 {
-	osmo_stat_item_group_free(bts->bts_statg);
-	rate_ctr_group_free(bts->bts_ctrs);
 	if (osmo_timer_pending(&bts->acc_mgr.rotate_timer))
 		osmo_timer_del(&bts->acc_mgr.rotate_timer);
 	if (osmo_timer_pending(&bts->acc_ramp.step_timer))
@@ -492,13 +488,14 @@ int main(int argc, char **argv)
 	osmo_gettimeofday_override = true;
 	osmo_gettimeofday_override_time = (struct timeval) {0, 0};
 
-	tall_bsc_ctx = talloc_named_const(NULL, 0, "gsm0408_test");
+	tall_bsc_ctx = talloc_named_const(NULL, 0, "acc_test");
 	osmo_init_logging2(tall_bsc_ctx, &log_info);
 	log_set_log_level(osmo_stderr_target, LOGL_INFO);
 	log_set_print_category_hex(osmo_stderr_target, 0);
 	log_set_print_category(osmo_stderr_target, 0);
 	log_set_print_filename2(osmo_stderr_target, LOG_FILENAME_NONE);
 	log_set_use_color(osmo_stderr_target, 0);
+	log_parse_category_mask(osmo_stderr_target, "DRSL,1:");
 	osmo_fsm_log_addr(false);
 
 	net = gsm_network_init(tall_bsc_ctx);
