@@ -20,9 +20,9 @@
  */
 
 #include <osmocom/bsc/gsm_data.h>
+#include <osmocom/bsc/bts.h>
 #include <osmocom/bsc/osmo_bsc.h>
 #include <osmocom/bsc/bsc_msc_data.h>
-#include <osmocom/bsc/gsm_04_80.h>
 #include <osmocom/core/application.h>
 #include <osmocom/bsc/codec_pref.h>
 
